/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	BNLoader.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Junho/2007 (2007-06-09)
 * @brief 	Pacote com a funo para ler e interpertar o arquivo com a definio de Rede bayesiana (BN: \e Bayesian \e Network).
 *
 * @todo
 * 	-# Checar eventual memory leak
 */

#ifndef BNLOADER_H
#define BNLOADER_H

/* Includes */
#include <string.h>

/* Bibliotecas especficas */
#include "BNMain.h"
#include "singlelinkedlist.h"
#include "hash.h"

/* Defines */

#define STATE_NOT_DEFINED 0				//!< Constante para estado no definido.
#define STATE_FINISHED 1				//!< Constante para indicar que o parser do arquivo foi concludo. 
#define STATE_BIF 2						//!< Constante para indicar que iniciou o parser da definio da Rede Bayesiana.
#define STATE_NETWORK 3					//!< Constante para indicar que iniciou o parser da Rede Bayesiana.
#define STATE_VARIABLE 4				//!< Constante para indicar que o parser esta analisando uma TAG que define uma varivel.
#define STATE_PROBABILITY 5				//!< Constante para indicar que o parser esta analisando uma TAG que define um potencial (ou probabilidade).
#define STATE_QUERY 6					//!< Constante para indicar que o parser esta analisando o contedo de uma TAG 'QUERY'.
#define STATE_EVIDENCES 7				//!< Constante para indicar que o parser esta analisando o contedo de uma TAG 'EVIDENCES'.
#define STATE_COMMENT 8					//!< Constante para indicar que o parser esta percorrendo um comentrio.

#define TAG_BIF "BIF"					//!< TAG que inicia o contedo XML para a definio de uma Rede Bayesiana.
#define TAG_NETWORK "NETWORK"			//!< TAG que inicia a definio de uma Rede Bayesiana propriamente dita.
#define TAG_NAME "NAME"					//!< TAG para nome de diversas estruturas de dados.
#define TAG_VARIABLE "VARIABLE"			//!< TAG para definio de uma varivel.
#define TAG_VALUE "VALUE"				//!< TAG para definio dos valores dos estados das variveis (alternativa para TAG_OUTCOME).
#define TAG_OUTCOME "OUTCOME"			//!< TAG para definio dos valores dos estados das variveis (alternativa para TAG_VALUE).
#define TAG_TYPE "TYPE"					//!< TAG que indica o tipo da varivel.
#define TAG_PROBABILITY "PROBABILITY"	//!< TAG para definio dos potenciais (alternativa para TAG_DEFINITION).
#define TAG_DEFINITION "DEFINITION"		//!< TAG para definio dos potenciais (alternativa para TAG_PROBABILITY).
#define TAG_FOR "FOR"					//!< TAG que indica a varivel "condicionada" nos potenciais.
#define TAG_GIVEN "GIVEN"				//!< TAG que indica as variveis "condicionadoras" nos potenciais.
#define TAG_TABLE "TABLE"				//!< TAG para a definio da tabela com a distribuio de probabilidades. 
#define TAG_QUERY "QUERY"				//!< TAG para a definio das variveis questionadas \e Query \e Variables.
#define TAG_EVIDENCES "EVIDENCES"		//!< TAG para definir umam varivel observada e o seu respectivo estado observado.

#define TYPE_DISCRETE "DISCRETE"		//!< Constante para identificar varivel discreta.
#define TYPE_CONTINUOUS "CONTINUOUS"	//!< Constante para identificar varivel contnua.

#define MAX_LINE_LENGTH 8192			//!< Tamanho mximo da linha que pode ser lida.
#define MAX_TAG_LENGTH 1024				//!< Tamanho mximo da TAG utilizada para identificar as partes do arquivo.
#define MAX_TAGVALUE_LENGTH 8192		//!< Tamanho mximo do contedo de uma TAG.

#define MAX_POTENTIAL_VALUES 4096		//!< Nmero mximo de valores que pode ser lido para um potencial.

/*!
 * @brief Carrega a definio de uma Rede Bayesiana a partir de um arquivo com formato predefinido
 * @param[in] filename Nome do arquivo que define a Rede Bayesiana.
 * @return Ponteiro para a Rede Bayesiana construda a partir do arquivo.
 */
BayesNet* bnload(char* filename);

/*!
 * @brief Carrega de uma arquivo com formato predefinido as variveis questionadas (\e Query \e Variables)
 * sobre as quais se deseja fazer a inferncia.
 * @param[out] nxq Ponteiro para um inteiro com o nmero de elementos do \e array de \e ids das \e Query \e Variables.
 * @param[in] filename Nome do arquivo no qual esto definidas as \e Query \e Variables.
 * @param[in] bayesnet Rede Bayesiana sobre a qual ser feita a inferncia.
 * @return \e Array com os \e ids das \e Query \e Variables.
 */

int* bnloadquery(int* nxq, char* filename, BayesNet* bayesnet);

/*!
 * @brief Carrega de uma arquivo com formato predefinido as variveis observadas (\e Evidence) e os respectios estados observados (\e findings)
 * sobre as quais se deseja fazer a inferncia.
 * @param[out] nxe Ponteiro para um inteiro com o nmero de elementos do \e array de \e ids das variveis observadas. 
 * @param[out] findings Potenciais que indicam os estados observados das variveis.
 * @param[in] filename Nome do arquivo no qual esto definidas as variveis observadas e os seus respectivos estados.
 * @param[in] bayesnet Rede Bayesiana sobre a qual ser feita a inferncia.
 * @return \e Array com os \e ids das variveis observadas.
 */

int* bnloadevidences(int* nxe, Potential** findings, char* filename, BayesNet* bayesnet);

#endif
