/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	potential.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Abril/2007
 * @brief 	Implementao da estrutura de dados e operaes com variveis e potenciais.
 *
 * Implementao de uma biblioteca genrica que define as estruturas de uma varivel de 
 * um potencial e as principais operaes com potenciais.
 */

#ifndef DBGOUT
#define DBGOUT stdout
#endif

#ifndef POTENTIAL_H
#define POTENTIAL_H

/* Includes */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

/* Defines  e constantes */

// Tipos de Variveis
#define VARIABLE_TYPE_DISCRETE 0	//!< Tipo de varivel discreta. Esta biblioteca suporta apenas variveis DISCRETAS.
#define VARIABLE_TYPE_CONTINUOUS 1	//!< Tipo de varivel continua. Esta biblioteca suporta apenas variveis DISCRETAS.

// Tamanhos mximos
#define VARIABLE_NAME_LENGTH 50			//!< Tamanho mximo do nome da varivel.
#define VARIABLE_STATE_NAME_LENGTH 50	//!< Tamanho mximo do nome que identifica o estado de uma varivel.

/* Definio dos tipos de dados */

/*! Estrutura de dados de uma Varivel */
typedef struct struct_Variable {
	int id;			//!< \e id que identifica a varivel. Ateno: No h verificao da unicidade do \e id da varivel.
	int type;		//!< Tipo da varivel. Tipos vlidos: VARIABLE_TYPE_DISCRETE e VARIABLE_TYPE_CONTINUOUS. Esta biblioteca suporta apenas variveis DISCRETAS.
	char* name;		//!< Nome da varivel.
	int nstates;	//!< Nmero de estados possveis para a varivel.
	char** states;	//!< Nome dos estados possveis para a varivel.
} Variable;

/*! Estrutura de dados de um Potencial */
typedef struct struct_Potential {
	int id;					//!< \e id que identifica o potencial. Ateno: No h verificao da unicidade do \e id do potencial
	int nvars;				//!< Nmero de variveis totais do potencial.
	Variable** vars;		//!< Lista de todas as variveis do potencial.
	int nvalues;			//!< Nmero total de combinaes de estados possveis das variveis.
	double* values;			//!< Valores da tabela de potencial.
} Potential;

/* Funes */

/*!
 * @brief Constri uma varivel.
 *
 * O \e id deve ser nico e um nmero maior ou igual a zero. No h verificao da unicidade do \e id da varivel, cabendo esta responsabilidade.
 * @param[in] id \e id da varivel a ser criada. O \e id deve ser nico e um nmero maior ou igual a zero, mas no h checagem da unicidade do \e id;
 * @param[in] name Nome a ser atribudo  varivel.
 * @param[in] type Tipo da varivel a ser criada. Tipos vlidos: VARIABLE_TYPE_DISCRETE e VARIABLE_TYPE_CONTINUOUS. Esta biblioteca suporta apenas variveis DISCRETAS.
 * @param[in] nstates Nmero total de estados possveis para a varivel.
 * @param[in] states Nomes dos estados possveis para a varivel.
 * @return Varivel cronstruda a partir dos dados de entrada.
 */

Variable* buildVariable(int id, char* name, int type, int nstates, char** states);

/*!
 * @brief Dump do contedo da varivel.
 * @param[in] var Varivel da qual se deseja fazer o dump.
 */

void dumpVariable(Variable* var);

/*!
 * @brief Destri uma varivel e desaloca os recursos que estavam alocados para a mesma.
 * @param[out] refVariable Referncia para a varivel a ser destruda.
 */

void destroyVariable(Variable** refVariable);

/*!
 * @brief Constri um potencial.
 *
 * O \e id deve ser nico e um nmero maior ou igual a zero. No h verificao da unicidade do potencial, cabendo esta responsabilidade
 * ao desenvolvedor.  importante observar a perfeita correspondncia entre os valores e a ordem com que as variveis esto dispostas
 * o array \e vars.
 * @param[in] id \e id do potencial. O \e id deve ser nico e um nmero maior ou igual a zero. No h verificao da unicidade do potencial.
 * @param[in] nvars Nmero de variveis do potencial.
 * @param[in] vars Variveis do potencial.
 * @param[in] values Valores do potencial.
 * @return Potencial criado a partir dos dados de entrada.
 */
Potential* buildPotential (int id, int nvars, Variable** vars, double* values);

/*!
 * @brief Dump do contedo do potencial.
 * @param[in] pot Potencial da qual se deseja fazer o dump.
 */

void dumpPotential(Potential* pot);

/*!
 * @brief Destri um potencial e desaloca os recursos que estavam alocados para o mesmo.
 * @param[out] refPotential Referncia para o potencial a ser destrudo.
 */

void destroyPotential(Potential** refPotential);

/*!
 * @brief Executa o produto entre dois potenciais.
 *
 * Esta funo executa a operao de multiplicao SEM permutar, ou seja, sem alterar os potenciais originais
 * as operaes so feitas apenas com o mapeamento dos ndices.
 * @param[in] pot_1 Primeiro potencial.
 * @param[in] pot_2 Segundo potencial.
 * @return Potencial resultante do produto de potenciais.
 */

Potential* multPotentials (Potential* pot_1, Potential* pot_2);

/*!
 * @brief Marginaliza o potencial para uma dada varivel.
 *
 * Esta funo executa a operao de marginalizao SEM permutar, ou seja, sem alterar os potenciais originais
 * as operaes so feitas apenas com o mapeamento dos ndices.
 * @param[in] variable Varivel a ser marginalizada do potencial.
 * @param[in] pot Potencial envolvido na operaoo de marginalizao.
 * @return Potencial resultante da marginalizao da varivel.
 */

Potential* marginalizePotential (Variable* variable, Potential* pot);

/*!
 * @brief Permuta as variveis e os valores de um potencial de acordo com um vetor de referncia.
 *
 * A permutao altera a ordem dos valores e das variveis no potencial. Esta operao pode ser utilizada para
 * compatibilizar a ordem das variveis de um potencial para executar operaes como marginalizao e produto.
 * Observe no entanto que  mais eficiente ao invs de permutar os potenciais fazer as operaes sem a operao
 * de permutao. Veja as funes: \e multPotentials e \e marginalizePotential.
 * @param[in] from_to Array de referncia com os \e ids dos potenciais.
 * @param[out] refPotential Referncia para o potencial a ser permutado.
 */
void permutePotential (int* from_to, Potential** refPotential);

/*!
 * @brief Compatibiliza a ordem das variveis em dois potenciais para executar operaes de produto.
 *
 * A compatiliza atravs da permutao a ordem dos valores e das variveis de dois potenciais. Esta operao pode ser 
 * utilizada para compatibilizar a ordem das variveis de um potencial para executar operaes de produto.
 * Observe no entanto que  mais eficiente ao invs de permutar os potenciais fazer as operaes sem a operao
 * de permutao. Veja as funes: \e multPotentials e \e marginalizePotential.
 * @param[in] first Varivel que dever ser a primeira  nos potenciais aps a sincronizao.
 * @param[out] refPotential_1 Referncia para o primeiro potencial envolvido na sincronizao.
 * @param[out] refPotential_2 Referncia para o segundo potencial envolvido na sincronizao.
 */

void syncPotentials (Variable* first, Potential** refPotential_1, Potential** refPotential_2);

/*!
 * @brief Executa o produto entre dois potenciais permutando os potenciais.
 *
 * Esta funo executa a operao de multiplicao PERMUTANDO a ordem do dos potenciais, ou seja, alterando os potenciais
 * originais. Esta funco faz parte de um conjunto de funes que foram utilizadas para desenvolvimento para checar os valores.
 * Observe no entanto que  mais eficiente ao invs de permutar os potenciais fazer as operaes sem a operao
 * de permutao. Veja as funes: \e multPotentials e \e marginalizePotential.
 * @param[in] first Varivel que deve ser a primeira aps a sincronizao dos vetores, para acelera futuras a operaes do produto.
 * @param[out] refPotential_1 Referncia para o primeiro potencial envolvido no produto.
 * @param[out] refPotential_2 Referncia para o segundo potencial envolvido no produto.
 * @return Potencial resultante do produto de potenciais.
 */

Potential* permuteMultPotentials (Variable* first, Potential** refPotential_1, Potential** refPotential_2);

/*!
 * @brief Marginaliza um varivel de um potencial permutando o potencial para excutar a operao.
 *
 * Esta funo executa a operao de marginalizao PERMUTANDO a ordem do do potencial, ou seja, alterando o potencial original.
 * Esta funco faz parte de um conjunto de funes que foram utilizadas para desenvolvimento para checar os valores.
 * Observe no entanto que  mais eficiente ao invs de permutar os potenciais fazer as operaes sem a operao
 * de permutao. Veja as funes: \e multPotentials e \e marginalizePotential.
 * @param[in] variable Varivel que deve ser marginalizada.
 * @param[out] refPotential Referncia para o potencial envolvido na marginalizao.
 * @return Potencial resultante da marginalizao da varivel.
 */


Potential* permuteMarginalizePotential (Variable* variable, Potential** refPotential);

/*!
 * @brief Funo auxiliar para criar um conjunto de variveis a ser utilizado para testes.
 *
 * @param[in] nvars Nmerode variveis.
 * @param[in] nstates Array com o nmero de estados das variveis.
 * @return Array de variveis criado com base nos dados de entrada.
 */

Variable** fakeVariables (int nvars, int* nstates);

/*!
 * @brief Funo auxiliar que cria um potencial \e fake para testes.
 * @param[in] id \e id do potencial a ser criado.
 * @param[in] first Primeira varivel que ser parte do potencial.
 * @param[in] offset Nmero de variveis a serem incorporadas no potencial.
 * @param[in] vars Array de variveis que sero utilizadas para criar o potencial
 * @return Potencial criado.
 */

Potential* fakePotential (int id, int first, int offset, Variable** vars);

#endif
