/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	BNMain.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Julho/2007 (2007-06-07)
 * @brief 	Main Package com as definies das estruturas de Redes Bayesianas (BN: \e Bayesian \e Networks).
 *
 * @todo
 * 	-# Na funo buildBayesNet ao invs de copiar (duplicar) os dados utilizados como 
 * 	parmetro de entrada, faz-se a atribuio direta para as estruturas de dados correspondentes
 * 	na BayesNet. Eventualmente a melhor forma seria copiar o contedo dos parmetros da memria
 * 	para as estruturas internas da BayesNet e desta forma isolar.Verificar se  utilizada esta funo.
 *
 */

#ifndef BNMAIN_H
#define BNMAIN_H

/* Bibliotecas Especficas */
#include "graph.h"
#include "potential.h"

/* Defines */

/* Definies das estruturas e tipos de dados */

/*! Estrutura que armazena uma Rede Bayesiana (BN: \e Bayesian \e Networks). Armazena os vrtices, os arcos e as
 * distribuies de probabilidades os potenciais (potential.h) associado a cada uma das variveis.
 */
typedef struct struct_BayesNet {
	struct struct_Graph* graph;		//!< Grafo que define a topologia da Rede.
	Variable** variables;			//!< Variveis associadas aos vrtices do grafo. O \e id da varivel  igual ao \e id do vrtice correspondente no grafo.
	Potential** potentials;			//!< Potenciais associados s vriaveis.
} BayesNet;

/* Funes */

/*!
 * @brief Cria uma nova Rede Bayesiana com um grafo direcionado vazio sem nenhuma varivel e sem nenhum potencial.
 * @return Um ponteiro para uma estrutura de dados para armazenar uma Rede Bayesiana.
 */

BayesNet* newBayesNet();

/*!
 * @brief Cria uma nova Rede Bayesiana a partir dos parmetros de entrada.
 * @param[in] graph Grafo que descreve a topologia da rede.
 * @param[in] variables \e Array de variveis que iro compor ao BN.
 * @param[in] potentials \e Array de potenciais que caracterizam a BN.
 * @return Um ponteiro para uma estrutura de dados para armazenar uma Rede Bayesiana.
 */

BayesNet* buildBayesNet(Graph* graph, Variable** variables, Potential** potentials);

/*!
 * @brief Dump para o \e stdout do contedo da BN.
 * @param[in] bayesnet Ponteiro para a BN da qual se deseja fazer o dump.
 */

void dumpBayesNet(BayesNet* bayesnet);

/*!
 * @brief Destri a BN e libera os recursos alocados. Ateno: Destri e libera inclusive os recursos que estavam
 * alocados para armazenar os potenciais e as variveis.
 * @param[out] refBayesNet Referncia para a BN que se deseja destruir.
 */
void destroyBayesNet(BayesNet** refBayesNet);

/*!
 * @brief Tamanho (nmero de vrtices=nmero de variveis) que compe o grafo da BN.
 * @param[in] bayesnet Ponteiro para a BN da qual se deseja saber o nmero de vrtices.
 * @return Nmero de vrtices (=nmero de variveis) da BN.
 */
int bnsize(BayesNet* bayesnet);

/*!
 * @brief Retorna o \e array de potenciais da BN.
 * @param[in] bayesnet Ponteiro para a BN da qual se deseja o \e array de potenciais.
 * @return \e Array de potenciais da BN.
 */

Potential** bnpotentials(BayesNet* bayesnet);

/*!
 * @brief Retorna o \e array de variveis da BN.
 * @param[in] bayesnet Ponteiro para a BN da qual se deseja o \e array de variveis.
 * @return \e Array de variveis da BN.
 */

Variable** bnvariables(BayesNet* bayesnet);

#endif
