/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	graph.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Fevereiro/2007
 * @brief 	Implementao de uma biblioteca para trabalhar com grafos
 *
 * Implementao de uma biblioteca genrica de grafos direcionais e no direcionais. Grafos direcionais so compostos por vrtices (\e nodes)
 * e arcos (\e arcs). Grafos no direcionais so formados por vrtices e links (\e links).
 * O desenvolvedor deve invocar explicitamente apenas as funes para criar/destruir/inserir o prprio grafo e os vrtices
 * as funes que lidam com as estruturas internas como a lista ligada, os \e nodes da mesma e a indexao so gerenciados pela prpria
 * biblioteca e no devem ser invocados diretamente.
 *
 * @todo
 * 	-# Melhorar a atribuio de \e ids para os vrtices.
 *	-# Garantir a unicidade dos \e ids dos vrtices. na vereso atual no h checagem da unicidade.
 *	-# Verificar que os \e ids so necessriamente positivos
 */

#ifndef GRAPH_H
#define GRAPH_H

/* Includes */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

/* Defines */

#define NOT_DIRECTED_GRAPH 0	//!< Constante para indicar um grafo no direcionado.
#define DIRECTED_GRAPH 1		//!< Constante para indicar um grafo direcionado.

#define ARCS "arcs"				//!< Palavra-reservada do arquivo de definio para grafos para indicar um grafo DIRECIONADO.
#define LINKS "links"			//!< Palavra-reservada do arquivo de definio para grafos para indicar um grafo NO DIRECIONADO.

/* Constantes para definir a forma de ordenao */

#define IN_DEGREE 0		//!< Ordenao pelo grau de entrada.
#define OUT_DEGREE 1	//!< Ordenao pelo grau de sada.
#define LINK_DEGREE 2	//!< Ordenao pelo nmero de links.
#define VERTEX_ID 3		//!< Ordenao pelo \e id do vrtice.

#define ASCENDANT 0		//!< Ordenao ascendente.
#define DESCENDANT 1	//!< Ordenao descendente.

#define VERTEX_NAME_MAX_LENGTH 100 //!< Tamanho mximo do nome que identifica o vrtice.

/* Definio dos tipos de dados */

/*! Tipos de dados que podero ser armazenados nos campos de dados do vrtice. */
typedef union union_Util {
	struct struct_Graph* G; 			//!< Ponteiro para um grafo.
	struct struct_Vertex* V;			//!< Ponteiro para um vrtice.
	struct struct_VertexLinkedList* L;	//!< Ponteiro para um lista de vrtices.
	char* S;							//!< Ponteiro para string.
	float* FP; 							//!< Ponteiro para float.
	int* IP;							//!< Ponteiro para inteiros.
	char C;								//!< Caracter.
	float F;							//!< Nmero \e float.
	int I;								//!< Nmero inteiro.
	void* data;							//!< Ponteiro \e genrico
} util;

/*! Vrtice (\e node) de um grafo */
typedef struct struct_Vertex {
	int id;										//!< \e id nico de um vrtice no grafo. No precisa ser sequencial e no h garantia que seja.  desejvel que seja em funo da forma como  feita a indexao. O valor \e id = -1  reservado para vrtices novos.
	char* name;									//!< Nome do \e node. O tamanho mximo  definido por VERTEX_NAME_MAX_LENGTH
	struct struct_Graph* graph;					//!< Ponteiro para o grado ao qual o \e node pertence.
	struct struct_VertexLinkedList* parents; 	//!< Lista de \e nodes \e pais do \e node.
	struct struct_VertexLinkedList* children; 	//!< Lista de \e nodes \e filhos do \e node.
	util user[4]; 								//!< utility field para ser utilizado pelo usurio.
	util lib[4]; 								//!< utility field para ser utilizado pela biblioteca.
} Vertex;

/*! Grafo genrico */
typedef struct struct_Graph {
	int type; 									//!< Indica o tipo do grafo: \e NOT_DIRECTED (no direcionado), \e DIRECTED (direcionado).
	char* name;									//!< Nome do grafo. No obrigatrio.
	int* order;									//!< \e Array que mapea um vrtice  sua ordem considerando a seqncia de \e ids do grafo.
	struct struct_Vertex** index;				//!< Index de vrtices para acelerar a busca pelo \e id. Mapea o \e ao seu \e id. Tempo de busca constante.
	struct struct_VertexLinkedList* vertices;	//!< Lista Ligada Simples de vrtices do grafo. No  garantia de ordenao pelo \e id.
	util user[4]; 								//!< \e Utility \e field para ser utilizado pelo usurio. No so inicializados cabendo ao desenvolvedor alocar/desalocar. Recomenda-se sempre inicializ-lo antes de utiliz-lo.
	util lib[4];								//!< \e Utility \e field para ser utilizado internamente por bibliotecas. No so inicializados cabendo ao desenvolvedor alocar/desalocar. Recomenda-se sempre inicializ-lo antes de utiliz-lo.
} Graph;

/*! Node da lista ligada simples de vrtices de um grafo. */
typedef struct struct_VertexNode {
	struct struct_Vertex* vertex;	//!< Vrtice do grafo.
	struct struct_VertexNode* next;	//!< Ponteiro para o prximo elemento da lista.
} VertexNode;

/*! Lista Ligada Simples (\e Single \e Linked \e List) de vrtices de um grafo.*/
typedef struct struct_VertexLinkedList {
	int length;							//!< Tamanho da lista (=nmero de vrtices da lista).
	struct struct_VertexNode* first;	//!< Ponteiro para o primeiro vrtice da lista.
	struct struct_VertexNode* last;		//!< Ponteiro para o ltimo vrtice da lista.
} VertexLinkedList;


/*!
 * @brief Cria um novo grafo, sem nome e sem vrtices.
 *
 * Para desalocar a memria alocada utilizar a funo destroyGraph.
 * @param[in] type Tipo de grafo no direcionado (\e NOT_DIRECTED) ou direcionado (\e DIRECTED).
 * @return Ponteiro para um grafo.
 */

Graph* newGraph (int type);

/*!
 * @brief Destri o grafo e desaloca a memria alocada.
 * @param[out] refGraph Referncia para o grafo que ser destrudo.
 */
void destroyGraph (Graph** refGraph);

/*!
 * @brief Cria um novo vrtice, sem nome, com as listas vazias e com \e id = -1.
 *
 * O \e id do vrtice pode ser alterado automaticamente ao inserir o vrtice no grafo na funo addVertex.
 * Para desalocar a memria alocada utilizar a funo destroyVertex.
 * @return Ponteiro para um vtice.
 */

Vertex* newVertex ();

/*!
 * @brief Constri um novo vrtice a partir da lista de argumentos.
 *
 * O \e id do vrtice deve ser nico e no h garantia de que ser seqncial. Em funo da forma como  feita a indexao, para
 * para melhor performance e reduzir a memria necessria para indexar  desejvel que os \e ids sejam seqenciais e contnuos.
 * Para desalocar a memria alocada utilizar a funo destroyVertex.
 * @param[in] id \e id do vrtice.
 * @param[in] name Nome do vrtice.
 * @return Vrtice construido com os argumento fornecidos.
 */

Vertex* buildVertex (int id, char* name);

/*!
 * @brief Destroy o vrtice e desaloca a memria.
 *
 * \e NO destri (desaloca a memria) do contedo do vrtice. Isto deve ser feito explicitamente no cdigo.
 * @param[out] refVertex Referncia para o vrtice que ser destrudo.
 */

void destroyVertex (Vertex** refVertex);

/*!
 * @brief Adiciona um vrtice ao grafo.
 *
 * Se o vrtice estiver com o \e id = -1 acrescentar, \e sem verificar duplicidade, o \e id como a posio do vrtice na lista.
 * @param[out] refGraph	Referncia para o grafo no qual o vrtice ser adicionado.
 * @param[out] refVertex Referncia para o vrtice a ser adicionado.
 * @return \e id do vrtice inserido.
 */

int addVertex(Graph** refGraph, Vertex** refVertex);

/*!
 * @brief Retorna o nmero de vrtices do grafo.
 * @param[in] graph Grafo do qual se deseja saber o nmero de ns.
 * @return Nmero de vrtices do grafo.
 */

int sizeGraph(Graph* graph);

/*!
 * @brief Retorna o nmero de vrtices do grafo.
 * @param[in] graph Grafo do qual se deseja saber o nmero de ns.
 * @return Nmero de vrtices do grafo.
 */

int graphsize(Graph* graph);

/*!
 * @brief Indexa a lista de vrtices pelo \e id do vrtice para acelerar a busca.
 *
 * Aps a indexao o tempo de busca do vrtice  constante. A indexao  feita automaticamente internamente pela bilioteca sempre que necessrio.
 * @param[out] refGraph Referncia para o grafo que ser indexado.
 */

void createGraphIndex(Graph** refGraph);

/*!
 * @brief Retorna o maior ID entre os vrtices, ou seja, o ltimo elemento da lista de vrtices.
 * @param[out] refGraph Referncia para o grafo do qual se deseja a informao.
 * @return Maior \e id da lista de vrtices.
 */

Vertex* getVertexMaxId(Graph** refGraph);

/*!
 * @brief Retorna o menor ID entre os vrtices, ou seja, o primeiro elemento da lista de vrtices.
 * @param[out] refGraph Referncia para o grafo do qual se deseja a informao.
 * @return Maior ID da lista de vrtices.
 */

Vertex* getVertexMinId(Graph** refGraph);

/*!
 * @brief Desaloca e destri o index de um grafo.
 * @param[out] refGraph Referncia para o grafo do qual ser desalocado o index.
 */

void destroyGraphIndex(Graph** refGraph);

/*!
 * @brief Retorna a ordem do vrtice de acordo com o \e id do mesmo.
 * @param[out] refGraph Referncia para o grafo do qual se deseja obter a ordem do vrtice.
 * @param[in] vertex Vertex que se deseja saber a posio.
 * @return Posio do vrtice.
 */

int getVertexOrder(Graph** refGraph, Vertex* vertex);

/*!
 * @brief Retorna o n-simo vrtice do grafo com base na seqncia de \e ids.
 * @param[out] refGraph Referncia para o grafo do qual se deseja obter o vrtice.
 * @param[in] n Nmero do elemento que se deseja.
 * @return n-simo vtice ou null se no houver vrtice.
 */

Vertex* getVertex(Graph** refGraph, int n);

/*!
 * @brief Retorna o vrtice identificado pelo \e id.
 * @param[out] refGraph Referncia para o grafo do qual se deseja obter o vrtice.
 * @param[in] id \e id do elemento que se deseja.
 * @return vrtice ou \e NULL se no houver vrtice.
 */

Vertex* getVertexById(Graph** refGraph, int id);

/*!
 * @brief Define um arco, necessariamente direcional, entre dois vrtices.
 * @param[out] refGraph  Referncia Grafo no qual ser adicionado o arco.
 * @param[out] refParent Referncia para o vrtice que  a origem do arco.
 * @param[out] refChild Referncia para o vrtice que  o destino do arco.
 */

void addArc(Graph** refGraph, Vertex** refParent, Vertex** refChild);

/*!
 * @brief Define um arco, necessariamente direcional, entre dois vrtices.
 * @param[out] refGraph  Referncia Grafo no qual ser adicionado o arco.
 * @param[in] idParent \e id do vrtice que  a origem do arco.
 * @param[in] idChild \e id do vrtice que  o destino do arco.
 */

void addArcByIds(Graph** refGraph, int idParent, int idChild);

/*!
 * @brief Define um Link, necessariamente NO direcional, entre dois vrtices.
 *
 * Um \e Link ser formado por dois arcos direcionais com direes opostas.
 * @param[out] refGraph  Referncia Grafo no qual ser adicionado o arco.
 * @param[out] refVertexA Referncia para o vrtice na extremidade do link.
 * @param[out] refVertexB Referncia para o vrtice na extremidade do link.
 */

void addLink(Graph** refGraph, Vertex** refVertexA, Vertex** refVertexB);

/*!
 * @brief Define um Link, necessariamente NO direcional, entre dois vrtices.
 *
 * Um \e Link ser formado por dois arcos direcionais com direes opostas.
 * @param[out] refGraph  Referncia Grafo no qual ser adicionado o arco.
 * @param[in] idVertexA Id do vrtice na extremidade do link.
 * @param[in] idVertexB Id do vrtice na extremidade do link.
 */

void addLinkByIds(Graph** refGraph, int idVertexA, int idVertexB);

/*!
 * @brief Grau de sada do vrtice, isto , nmero de arcos cuja origem  o vrtice.
 * @param[in] vertex Vrtice do qual se deseja saber o grau de sada.
 * @return Grau de sada.
 */

int getVertexOutDegree(Vertex* vertex);

/*!
 * @brief Grau de entrada do vrtice, isto , nmero de arcos cujo destino  o vrtice.
 * @param[in] vertex Vrtice do qual se deseja saber o grau de entrada.
 * @return Grau de entrada.
 */

int getVertexInDegree(Vertex* vertex);

/*!
 * @brief O nmero ligaes no vrtice.
 *
 * Se o vrtice est em um grafo direcionado o nmero de ligaes  o resultado da soma do arcos que entram e saem (parents + children).
 * Se o grafo for no direcionado o nmero de ligaes  o nmero de \e links.
 * @param[in] vertex Vrtice do qual se deseja saber o grau de entrada.
 * @return Grau total o vrtices
 */

int getVertexDegree(Vertex* vertex);

/*!
 * @brief Informa se o grafo  direcionado ou no, ou seja, se as ligaes entre os vrtices so direcionadas.
 * @param[in] graph Grafo do qual se deseja a informao.
 * @return NOT_DIRECTED_GRAPH se no for direcionado; (DIRECTED_GRAPH se for direcionado.
 */

int isDirected(Graph *graph);

/*!
 * @brief Interpreta arquivo com formato predefinido e constri um grafo.
 * @param[in] filename Nome do arquivo com a definio do grafo.
 * @return Ponteiro para o grafo construdo a partir dos dados do arquivo.
 */

Graph* loadGraph(char* filename);

/*!
 * @brief Constri um grafo a partir de um vetor de vrtices e links.
 * @param[in] type Tipo do grafo (Direcionado ou No direcionado).
 * @param[in] vertices \e Array de vrtices.
 * @param[in] nVertices Nmero de vrtices.
 * @param[in] links \e Matriz n x 2 com as ligaes (\e Arcs ou \e Links). Ordem: [parent, child].
 * @param[in] nLinks Nmero de ligaes.
 * @return Ponteiro para o grafo construdo a partir dos dados do arquivo.
 */

Graph* buildGraph(int type, char** vertices, int nVertices, int** links, int nLinks);

/*!
 * @brief Dump de um grafo para o \e stdout.
 * @param[in] graph Grafo do qual se deseja fazer o dump.
 */

void dumpGraph(Graph* graph);

/*!
 * @brief Obtem a lista ligada de vrtices do grafo.
 * @param[in] graph Grafo do qual se deseja a lista de vrtices.
 * @return Lista ligada (VertexLinkedList) com os vrtices.
 */

VertexLinkedList* getVertices(Graph* graph);

/*!
 * @brief Obtem a lista ligada de vrtices dos \e parents do vrtice.
 * @param[in] vertex Vertex do qual se deseja a lista de vrtices parents.
 * @return Lista ligada (VertexLinkedList) com os parents do vrtice.
 */

VertexLinkedList* getVertexParents(Vertex* vertex);

/*!
 * @brief Obtem a lista ligada de vrtices dos children do vrtice.
 * @param[in] vertex Vertex do qual se deseja a lista de vrtices children.
 * @return Lista ligada (VertexLinkedList) com os children do vrtice.
 */

VertexLinkedList* getVertexChildren(Vertex* vertex);

// VertexLinkedList: Funes que lidam com a lista de vrtices

/*!
 * @brief Cria uma nova lista ligada simples.
 * @return Ponteiro para uma lista ligada simples com nodes para armazenar vrtices.
 */

VertexLinkedList* newVertexLinkedList ();

/*!
 * @brief Cria uma nova lista ordenada com base em uma lista dada.
 * @param[in] list Lista de vrtices a ser ordenada.
 * @param[in] field Campo utilizado na ordenao (LINK_DEGREE/IN_DEGREE/OUT_DEGREE).
 * @param[in] direction direo da ordenao (ASCENDANT/DESCENDANT).
 * @return Ponteiro para uma lista ligada ordenada
 */

VertexLinkedList* newSortedVertexLinkedList (VertexLinkedList* list, int field, int direction);

/*!
 * @brief Destroy uma lista ligada simples desaloca os \e nodes, mas sem desalocar o vrtice armazenado no \e node.
 *
 * ATENO: Considerando que o Vertex pertence ao grafo e no  lista de vrtice,
 * que  apenas uma estrutura auxiliar, a destruio a lista NO desaloca os,
 * vrtices, os mesmos sero desalocados quando o grafo for destrudo.
 * @param[out] refList Referncia  lista que deve ser desalocada
 */

void destroyVertexLinkedList (VertexLinkedList** refList);

/*!
 * @brief Cria um novo node da lista ligada de vrtices.
 * @return Ponteiro para um novo node da lista ligada de vrtices.
*/

VertexNode* newVertexNode ();

/*!
 * @brief Destri um node da lista de vrtices, mas no desaloca o vrtice armazenado no \e node.
 *
 * ATENO: Considerando que o Vertex pertence ao grafo e no  lista de vrtice,
 * que  apenas uma estrutura auxiliar, a destruio a lista NO desaloca os
 * vrtices, os mesmos sero desalocados quando o grafo for destrudo.
 * @param[in] refNode Node da lista que deve ser destrudo
 */

void destroyVertexNode(VertexNode** refNode);

/*!
 * @brief Retorna o tamanho da lista.
 * @return Tamanho da lista
 */

int length(VertexLinkedList* list);

/*!
 * @brief Acrescenta um VertexNode no incio da lista ligada. 
 * @param[out] refList	Lista na qual deve ser acrescentado o \e node.
 * @param[in] vertex	VertexNode que dever ser acrescentado.
 */

void push (VertexLinkedList** refList, Vertex* vertex);

/*!
 * @brief Acrescenta um VertexNode no final da lista ligada. 
 * @param[out] refList	Lista na qual deve ser acrescentado o \e node.
 * @param[in] vertex	VertexNode que dever ser acrescentado.
*/

void appendVertexNode(VertexLinkedList** refList, Vertex* vertex);

/*!
 * @brief Adiona um vertex no final da lista de maneira ordenada.
 * @param[out] refList Referncia a lista de vrtices.
 * @param[in] vertex VertexNode que dever ser acrescentado.
 * @param[in] field Campo utilizado na ordenao (LINK_DEGREE/IN_DEGREE/OUT_DEGREE).
 * @param[in] direction direo da ordenao (ASCENDANT/DESCENDANT).
 */

void appendVertexNodeSorted(VertexLinkedList** refList, Vertex* vertex, int field, int direction);

/*!
 * @brief Remove o primeiro elemento da lista.
 * @param[out] refList Referncia a lista de vrtices.
 * @return Retorna o contedo que estava armazenado no VertexNode.
 */

Vertex* removeFirst(VertexLinkedList** refList);

/*!
 * @brief Remove o ltimo elemento da lista.
 * @param[out] refList Referncia a lista de vrtices.
 * @return Retorna o contedo que estava armazenado no VertexNode.
 */

Vertex* removeLast(VertexLinkedList** refList);

/*!
 * @brief Ordena uma lista de vrtices por um campo definido na ordem definida.
 * @param[out] refList Lista de vrtices a ser ordenada
 * @param[in] field Campo utilizado na ordenao (LINK_DEGREE/IN_DEGREE/OUT_DEGREE)
 * @param[in] direction direo da ordenao (ASCENDANT/DESCENDANT)
 */

void sortVertexLinkedList(VertexLinkedList** refList,int field, int direction);

/*!
 * @brief Dump de uma lista de vrtices.
 * @param[in] list Lista da qual se deseja o dump. 
 */

void dumpVertexLinkedList(VertexLinkedList* list);

#endif
