/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	hash.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Abril/2007
 * @brief 	Implementao de uma estrutura para criar hashs que mapeiam strings-nmeros_reais, ou seja, pares chave-valor.
 *
 * Implementao de uma biblioteca genrica para trabalhar com hashs que mapeiam palavras (strings) a nmeros reais.
 */

#ifndef HASH_H
#define HASH_H

/* Defines */

#define HASH_MAX 100 //!< Tamanho mximo da chave do hash

#define get16bits(d) (*((const uint16_t *) (d))) //!< Macro para fazer cast de valores utilizada na rotina que gera o hash

/* Definio dos tipos de dados */

/*! Estrutura de dados utilizada para armazenar o contedo do hash */
typedef struct struct_Hash {
	int length;						//!< Tamanho (nmero de elementos) do hash.
	struct struct_HashNode** nodes;	//!< N utilizado para armazenar o contedo do hash.
} Hash;

/*! HashNode de uma lista ligada */
typedef struct struct_HashNode {
	char* key;						//!< Chave que identifica o elemento no hash.
	double value;					//!< Valor numrico armazenado no elemento do hash.
	struct struct_HashNode* next;	//!< Ponteiro para o prximo \e node da lista na estrutura de armazenamento do hash.
} HashNode;


/*!
 * @brief Funo de espalhamento que gera o indicador da posio a partir da chave do hash.
 * @param[in] key Chave que identifica o elemento do hash.
 * @return Valor correspondente  posio no array de armazenamento de dados.
 */
int hashIndex (char* key);

/*!
 * @brief Cria um novo hash vazio.
 * @return Ponteiro para um hash criado.
 */

Hash* newHash ();

/*!
 * @brief Destri um hash e desaloda todo contedo alocado.
 * @param[out] refHash Referncia para o hash a ser destrudo.
 */

void destroyHash (Hash** refHash);

/*!
 * @brief Retorna o tamanho, nmero de elementos ou pares chave-valor, do hash.
 * @return Tamanho do hash.
 */

int hlength(Hash* Hash);

/*!
 * @brief Informa se o hash possui uma determinada chave armazenada.
 * @param[in] hash Hash a ser consultado.
 * @param[in] key Chave que se deseja saber se est armazenada no hash.
 * @return '1': se houver um elemento correspondente  chave procurada. '0': caso contrrio.
 */

int hcontains (Hash* hash, char* key);


/*!
 * @brief Retorna do hash o valor armazenado correspondente  chave.
 * @param[in] hash Hash a ser consultado.
 * @param[in] key Chave do qual se deseja o valor armazenado.
 * @return Valor correspondente  chave armazenada no hash ou '\\0' caso o hash no tenha a chave procurada.
 */

double hget (Hash* hash, char* key);

/*!
 * @brief Adiciona um par chave-valor em um hash.
 * @param[out] refHash Referncia para o hash no qual o par chave-valor ser armazenado
 * @param[in] key Chave do par chave-valor que ser armazenado.
 * @param[in] value Valor do par chave-valor que ser armazenado.
 * @return Nmero de elementos armazenados no hash.
 */

int hput (Hash** refHash,char* key, double value);

/*!
 * @brief Cria um novo node da lista ligada utilizada para armazenar os dados do hash.
 * @return Ponteiro para um novo node da lista ligada.
 */

HashNode* newHashNode ();

/*!
 * @brief Constri um novo node da lista ligada utilizada para armazenar os dados do hash.
 * @param[in] key Chave do par chave-valor que ser armazenado.
 * @param[in] value Valor do par chave-valor que ser armazenado.
 * @return Ponteiro para um novo node da lista ligada.
 */

HashNode* buildHashNode (char* key, double value);

/*!
 * @brief Destri um HashNode e desaloda todo contedo alocado.
 * @param[out] refNode Referncia para o node a ser destrudo.
 */

void destroyHashNode(HashNode** refNode);

/*!
 * @brief Dump do contedo armazenado no hash
 * @param[in] hash Hash do qual ser feito o dump.
 */

void dumpHash(Hash* hash);

#endif
