/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	threadtree.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Abril/2007
 * @brief 	\e ThreadTree (rvore de Threads) para executar processamento em paralelo.
 *
 * Implementao genrica de uma rvore de threads que executa funes genricas em paralelo. O vrtice pai  executado 
 * apenas aps os vrtices filhos terem sido executados, ou seja, o pai faz um \e join dos filhos. A raiz  o 
 * ltimo vrtice a ser eliminado.  responsabilidade do programador garantir que o cdigo ser \e threadsafe. A biblioteca
 * de threadtree no faz nenhuma checagem consistncia das estruturas de dados compartilhadas durante a execuo do processo.
 * O desenvolvedor deve invocar explicitamente apenas as funes para criar/destruir/inserir a prpria \e ThreadTree e os \e ThreadVertex
 * as funes que lidam com as estruturas internas como a lista ligada, os \e nodes da mesma e a indexao so gerenciados pela prpria
 * biblioteca e no devem ser invocados diretamente.
 */

#ifndef THREADTREE_H
#define THREADTREE_H

/* Includes */
#include <pthread.h>

/* Defines */

// Status possveis 
#define THREAD_READY 0		//!< \e Thread est pronta para ser executada. Para executar uma thread ela NECESSARIAMENTE deve estar no estado \e READY.
#define THREAD_RUNNING 1	//!< \e Thread est em execuo.
#define THREAD_FINISHED 2	//!< \e Thread foi concluda. Para execut-la novamente  necessria coloc-la no estado READY.

/* Definio dos tipos de dados */

/*! Vrtice de um grafo que contm ponteiros para a funo que ir executar bem como as para estruturas de dados necessrias */
typedef struct struct_ThreadVertex {
	int id;												//!< \e id nico de um vrtice no grafo. Atribudo automaticamente para o vrtice ao ser adicionado a uma \e threadtree. No deve ser manipulado pelo desenvolvedor.
	int status;											//!< Estado corrente da thread.
	pthread_t tid;										//!< \e id interno da thread utilizado pela biblioteca \e pthread.h
	void *data;											//!< Tipo genrico de dados apra armazenar dados que sero utilizados no processamento.
	void *task;											//!< Funo genrica a ser executada pela thread.
	struct struct_ThreadVertex* parent;					//!< \e parent do vrtice, como  uma rvore tem apenas um 'pai'.
	struct struct_ThreadVertexLinkedList* children;		//!< \e children do vrtices, lista de flihos dos vrtices.
} ThreadVertex;

/*! rvore de Threads */
typedef struct struct_ThreadTree {
	struct struct_ThreadVertex* root;					//!< Vrtice que ser a raiz da rvore
	struct struct_ThreadVertex** index;					//!< Array de vrtices utilizado para indexar os vrtices da \e threadtree
	struct struct_ThreadVertexLinkedList* vertices;		//!< Lista ligadas dos vrtices da \e threadtrer
} ThreadTree;

/*! ThreadVertexNode de uma lista ligada utilizado para armazenar o \e ThreadVertex em uma lista ligada */
typedef struct struct_ThreadVertexNode {
	struct struct_ThreadVertex* vertex;					//!< \e ThreadVertex armazenado no \e node da lista ligada.
	struct struct_ThreadVertexNode* next;				//!< Ponteiro para o prximo \e node da lista.
} ThreadVertexNode;

/*! Single Linked List de \e nodes contendo \e ThreadVertex */
typedef struct struct_ThreadVertexLinkedList {
	int length;											//!< Tamanho atual da lista, ou seja, nmero de nodes da lista ligada.
	struct struct_ThreadVertexNode* first;				//!< Ponteiro para o primeiro \e node da lista.
	struct struct_ThreadVertexNode* last;				//!< Ponteiro para o ltimo \e node da lista.
} ThreadVertexLinkedList;


/* Funes */

/*!
 * @brief Cria nova rvore sem vrtices.
 * @return Ponteiro para a rvore de threads criada.
 */

ThreadTree* buildThreadTree();

/*!
 * @brief Destri a rvore e desaloca a memria.
 *
 * Para liberar os recursos alocados para a \e threadtree basta invocar a funo \e destroyThreadTree. As demais funes
 * para desalocar o index, a lista ligada de vrtices, os repectivos \e nodes e os prprios vrtices sero invocadas no momento da
 * destruio da rvore.
 * @param[out] refThreadTree Referncia para a rvore que ser destruda.
 */
void destroyThreadTree(ThreadTree** refThreadTree);

/*!
 * @brief Cria um novo vrtice para rvore, com \e id=-1, com as listas vazias
 * @param[in] task Ponteiro para a funo a ser executada.
 * @param[in] data Dados para a funo que ser executada.
 * @return Ponteiro para o vrtice recm criado.
 */

ThreadVertex* buildThreadVertex (void* task, void* data);

/*!
 * @brief Destri o vrtice e desaloca a memria.
 *
 * Para liberar os recursos alocados para a \e threadtree basta invocar a funo \e destroyThreadTree. As demais funes
 * para desalocar o index, a lista ligada de vrtices, os repectivos \e nodes e os prprios vrtices sero invocadas no momento da
 * destruio da rvore.
 * @param[out] refThreadVertex Referncia para o vrtice que ser destrudo.
 */

void destroyThreadVertex (ThreadVertex** refThreadVertex);

/*!
 * @brief Adiciona um vrtice (\e ThreadVertex)  rvore de \e threads e atribui ao vrtice o \e id sequencial.
 * @param[out] refThreadTree Referncia para a rvore no qual o vrtice ser adicionado.
 * @param[out] refThreadVertex Referncia para o vrtice a ser adicionado.
 * @return Id sequencial para do vrtice adicionado.
 */

int addThreadVertex(ThreadTree** refThreadTree, ThreadVertex** refThreadVertex);

/*!
 * @brief Retorna o nmero de vrtices da \e threadtree.
 * @param[out] threadtree \e ThreadTree do qual se deseja saber o nmero de ns.
 * @return nmero de vrtices da \e ThreadTree.
 */

int sizeThreadTree(ThreadTree* threadtree);

/*!
 * @brief Indexa a lista de vrtices para acelerar a busca. Mtodo invocado internamente, No deve ser invocado explicitamente.
 * @param[out] refThreadTree Referncia para a rvore que ser indexado.
 */

void createThreadTreeIndex(ThreadTree** refThreadTree);

/*!
 * @brief Desaloca e destri o index da rvore de \e threads.
 *
 * Para liberar os recursos alocados para a \e threadtree basta invocar a funo \e destroyThreadTree. As demais funes
 * para desalocar o index, a lista ligada de vrtices, os repectivos \e nodes e os prprios vrtices sero invocadas no momento da
 * destruio da rvore.
 * @param[out] refThreadTree Referncia para a rvore que ser des-indexada.
 */

void destroyThreadTreeIndex(ThreadTree** refThreadTree);

/*!
 * @brief Retorna o n-simo vrtice da rvore. Os \e ids atrbuidos automaticamente aos vrtices ao adicion-los  rvore so sequenciais.
 * @param[out] refThreadTree Referncia para a rvore do qual se deseja obter o vrtice.
 * @param[in] n Nmero do elemento que se deseja.
 * @return \e n-simo vtice ou \e null se no houver vrtice.
 */

ThreadVertex* getThreadVertex(ThreadTree** refThreadTree, int n);

/*!
 * @brief Define um arco, necessariamente direcional, entre dois vrtices. O arco  direcional pois estamos lidando com uma rvore.
 * @param[out] refThreadTree Referncia  rvore na qual ser adicionado o arco.
 * @param[in] refParent Referncia para o vrtice que  a origem do arco.
 * @param[in] refChild Referncia para o vrtice que  o destino do arco.
 * @return Nmeros de filhos do \e parent, que corresponde ao nmero de arcos que tem origem no vrtice \e parent.
 */

int addTreeArc(ThreadTree** refThreadTree, ThreadVertex** refParent, ThreadVertex** refChild);

/*!
 * @brief Define um arco, necessariamente direcional, entre dois vrtices. O arco  direcional pois estamos lidando com uma rvore.
 * @param[out] refThreadTree Referncia  rvore na qual ser adicionado o arco.
 * @param[in] idParent Id do vrtice que  a origem do arco.
 * @param[in] idChild Id do vrtice que  o destino do arco.
 * @return Nmeros de filhos do \e parent, que corresponde ao nmero de arcos que tem origem no vrtice \e parent.
 */

int addTreeArcByIds(ThreadTree** refThreadTree, int idParent, int idChild);

/*!
 * @brief Dump no \e stdout do contedo da rvore de \e threads.
 * @param[in] threadtree \e ThreadTree da qual se deseja fazer o dump.
 */

void dumpThreadTree (ThreadTree* threadtree);

// LINKED LIST: Funes que lidam com a lista de vrtices

/*!
 * @brief Cria uma nova lista ligada simples para armazenar os vrtices do grafo utilizado para representar \e threadtree.
 * @return Ponteiro para uma lista ligada.
 */

ThreadVertexLinkedList* buildThreadVertexLinkedList ();

/*!
 * @brief Destroy uma lista ligada simples 
 *
 * Para liberar os recursos alocados para a \e threadtree basta invocar a funo \e destroyThreadTree. As demais funes
 * para desalocar o index, a lista ligada de vrtices, os repectivos \e nodes e os prprios vrtices sero invocadas no momento da
 * destruio da rvore.
 * ATENO: Considerando que o ThreadVertex pertence ao grafo e no  lista de vrtice, que  apenas uma estrutura auxiliar, a
 * destruio a lista NO desaloca os, vrtices, os mesmos sero desalocados quando a rvore for destrudo.
 * @param[out] refList Referncia  lista que deve ser desalocada
 */

void destroyThreadVertexLinkedList (ThreadVertexLinkedList** refList);

/*!
 * @brief Cria um novo \e node da lista ligada de vrtices que ir armazenar o \e ThreadVertex passado como parmetro.
 * @param[in] vertex \e ThreadVertex que ser armazenado no \e node a ser criado.
 * @param[in] next Ponteiro para o prximo \e node da lista ligada.
 * @return Ponteiro para um novo node da lista ligada de vrtices
 */

ThreadVertexNode* buildThreadVertexNode (ThreadVertex* vertex, ThreadVertexNode* next);

/*!
 * @brief Destri um node da lista de vrtices.
 * ATENO: Considerando que o ThreadVertex pertence  \e ThreadTree e no  lista de vrtice,
 * que  apenas uma estrutura auxiliar, a destruio a lista NO desaloca os, vrtices, os mesmos
 * sero desalocados quando a rvore for destrudo.
 * @param[out] refNode Node da lista que deve ser destrudo.
 */

void destroyThreadVertexNode(ThreadVertexNode** refNode);

/*!
 * @brief Acrescenta um \e ThreadVertexNode no final da lista ligada. 
 * @param[out] refList Lista na qual deve ser acrescentado o \e node.
 * @param[in] vertex \e ThreadVertexNode que deve ser acrescentado.
 * @return Posio na lista do node inserido.
 */

int appendThreadVertexNode(ThreadVertexLinkedList** refList, ThreadVertex* vertex);

/*!
 * @brief Dump de uma lista de vrtices.
 * @param[in] list Lista da qual se deseja o dump 
 */

void dumpThreadVertexLinkedList(ThreadVertexLinkedList* list);

// THREADS

/*!
 * @brief Executa uma \e thread, ou seja, a funo que foi atribuda ao \e ThreadVertex com os dados armazenados no mesmo.
 * A invocao desta funo no boqueia a execuo do programa (ou da \e thread que a invocou) que continua a ser executado.
 * 
 * ATENO: Para ser executado o \e ThreadVertex deve estar com o estado \e READY. Ao iniciar a execuo o \e ThreadVertex
 * passa para o estado \e RUNNING e ao conclu-la passa para o estado \e FINISHED.
 * @param[out] refVertex Referncia ao vertex da thread que deve ser executada.
 */

int run(ThreadVertex** refVertex);

/*!
 * @brief Define um estado para a \e thread.
 * 
 * ATENO: Para ser executado o \e ThreadVertex deve estar com o estado \e READY. Ao iniciar a execuo o \e ThreadVertex
 * passa para o estado \e RUNNING e ao conclu-la passa para o estado \e FINISHED.
 * @param[out] refVertex Referncia ao vertex da \e thread.
 * @param[in] status Estado a ser definido para o \e ThreadVertex. Estados possveis: READY, RUNNING E FINISHED.
 */

void setStatus (ThreadVertex** refVertex, int status);

/*!
 * @brief Retorna o estado atual da \e thread armazenada no vrtice.
 * @param[in] vertex Vtice que contm a \e thread.
 * @return Estado corrente da \e ThreadVertex. Estados possveis: READY, RUNNING E FINISHED.
 */

int getStatus (ThreadVertex* vertex);

/*!
 * @brief \e Join de um \e ThreadVertex, ou seja, da \e thread que esta estrutura de dados representa internamente.
 *
 * A invocao desta funo bloqueia a execuo do programa, ou da \e thread que a invocou, at que a \e thread aguardada seja concluda.
 * Esta funo  utilizada internamente pelas funes \e joinDescendants e \e joinThreadTree.
 * @param[out] refVertex Referncia ao vrtice da \e thread que deve ser aguardada.
 */

void join (ThreadVertex** refVertex);

/*!
 * @brief \e Join de todos os descendentes (filhos) de um \e ThreadVertex.
 *
 * A invocao desta funo bloqueia a execuo do programa, ou da \e thread que a invocou, at que a todos os descendentes 
 * tenham sido concludos.
 * @param[out] refVertex Referncia ao vrtice do qual a execuo dos descendentes deve ser aguardada.
 */

void joinDescendants (ThreadVertex** refVertex);

/*!
 * @brief \e Join da rvore como um todo.
 *
 * A invocao desta funo bloqueia a execuo do programa, ou da \e thread que a invocou, at que a execuo de todos os
 * vrtices tenha sido concluda, ou seja,  equivalente a aguardar a execuo dos descendentes da raiz da rvore.
 * @param[out] refThreadTree Referncia para a rvore de threads.
 */

void joinThreadTree (ThreadTree** refThreadTree);

/*!
 * @brief Inicia a execuo de uma rvore de threads em um processo em paralelo, ou seja, a invocao desta funo
 * no boqueia a execuo do program (ou da \e thread que a invocoua) que continua a ser executado. A aplicao deve
 * necessariamente definir as threads como THREAD_READY
 * @param[out] refThreadTree Referncia  rvore que deve ser executada
 */

void execute(ThreadTree** refThreadTree);


/*!
 * @brief RE-Inicia a execuo de uma rvore de threads em um processo em paralelo, ou seja, a invocao desta funo
 * no boqueia a execuo do program (ou da \e thread que a invocoua) que continua a ser executado. Igual execute(), mas
 * automaticamente define o status das threads como THREAD_READY.
 * @param[out] refThreadTree Referncia  rvore que deve ser executada
 */

void restart(ThreadTree** refThreadTree);

/*!
 * @brief Define o \e root da rvore pelo \e id do \e ThreadVertex.
 *
 * Caso no seja atribudo explicitamente pelo desenvolvedor, a biblioteca, se necessrio, tenta descobrir automaticamente
 * a raiz, ou seja, o \e node com \e parent = null. Na quase totalidade dos casos no h necessidade e no  recomendavel
 * que o desenvolvedor defina explicitamente o raiz, deixando esta tarefa para a biblioteca.
 * @param[out] refThreadTree Referncia  rvore que deve ser executada
 * @param[in] rootId Id do node root
 */

void setRootById(ThreadTree** refThreadTree, int rootId);

/*!
 * @brief Define o root da rvore informando diretamente o \e ThreadVertex. 
 *
 * Caso no seja atribudo explicitamente pelo desenvolvedor, a biblioteca, se necessrio, tenta descobrir automaticamente
 * a raiz, ou seja, o \e node com \e parent = null. Na quase totalidade dos casos no h necessidade e no  recomendavel
 * que o desenvolvedor defina explicitamente o raiz, deixando esta tarefa para a biblioteca.
 * @param[out] refThreadTree Referncia  rvore que deve ser executada.
 * @param[in] root node root.
 */

void setRoot(ThreadTree** refThreadTree, ThreadVertex* root);

#endif
