/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*
 * @file	BNVarElimination.c
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Junho/2007 (2007-06-28)
 *
 * Para documentao ver BNVarElimination.h
 */

/* Arquivo com o cabealho das funes */
#include "BNVarElimination.h"

/* Funes */

Elimination* buildElimination(int id, Variable* variable,int marginalize) {

	Elimination* elimination = malloc(sizeof(Elimination));
	elimination->id = id;
	elimination->marginalize = marginalize;
	elimination->nconst = 0;
	elimination->variable = variable;
	elimination->potentials = newSingleLinkedList();
	elimination->result = NULL;

	// Iniciliza o mutex
	pthread_mutex_init(&(elimination->mutex), NULL);

	return elimination;
}

void destroyElimination(Elimination** refElimination) {

	Elimination* elimination = *refElimination;
	elimination->variable= NULL;

	destroySingleLinkedList(&elimination->potentials);

	if (elimination->result!=NULL) destroyPotential(&elimination->result);

	// Destroi o multex
	pthread_mutex_destroy(&elimination->mutex);

	free(elimination);
	*refElimination = NULL;
}

void dumpElimination(Elimination* elimination) {

	SingleLinkedListNode* current;

	fprintf(DBGOUT,"Elimination: %d\n",elimination->id);

	if (elimination->variable!=NULL) {
		dumpVariable(elimination->variable);
	} else {
		fprintf(DBGOUT,"Variable = NULL\n");
	}

	fprintf(DBGOUT,"Potenciais:\n");
	current = elimination->potentials->first;
	while (current!=NULL) {
		dumpPotential((Potential*)current->data);
		current = current->next;
	}
}

// Mtodo que constroi a rvore de eliminao de threads
ThreadTree* buildBNThreadTree(int nxq, int* xq, int nxe, int* xe, Potential** findings,BayesNet* bayesnet, int* elmorder, Graph* moral, Graph* elmtree) {
	
	ThreadTree* threadtree = NULL;		// rvore de Threads
	ThreadVertex* thread = NULL;		// Varivel auxiliar que instancia um vrtive da rvore de threads

	VertexNode* child;				// N de uma lista ligada simples
	VertexNode* current = NULL;		// Varivel auxiliar para percorrer o Moral Graph

	Potential* result = NULL;		// Potencial resultante da sequencia de operaes
	Potential* potential = NULL;	// Varivel potencial auxiliar

	int varelmorder[bnsize(bayesnet)];	// Mapa o id da varivel para a ordem de eliminao
	int used[bnsize(bayesnet)];			// Vetor de flag para indicar se um potencial foi utilizado

	Elimination** eliminations = NULL;

	int nvars, ct_used;;
	int i,j,id;;

	// Nmero de variveis envolvidas, pode no ser toda a rede
	nvars = graphsize(elmtree);

	// Monta a estrutura que ser utilizada para passar dados para as threads
	eliminations = malloc(nvars*sizeof(Elimination*));

	// Constri a rvore de thread
	threadtree = buildThreadTree();

	// Cria-se as threads vrtices
	for (i=0;i<nvars;i++) {

		// Inicializa o mapa
		varelmorder[elmorder[i]] = i;

		// Inicializa a flag
		used[elmorder[i]] = 0;

		// Monta as eliminaes e popula varelmorder
		eliminations[i] = (Elimination*) buildElimination(i,bayesnet->variables[elmorder[i]],1);

		// Constri uma thread da rvore de thread
		thread = buildThreadVertex(taskVariableElimination,(void*)(eliminations));

		// Atribui o vrtice  rvore de threads
		addThreadVertex(&threadtree,&thread);
	}

	// Adiciona os arcos de dependncia
	for (i=0;i<nvars;i++) {
		child = (getVertexChildren(getVertex(&elmtree,i)))->first;
		while (child!=NULL) {
			addTreeArcByIds(&threadtree,i,child->vertex->id);
			child = child->next;
		}
	}

	//---
	
	// Tomando Xq, se no NULL, marca as operaes que no devem ser marginalizadas
	if (xq!=NULL) { // Apenas por segurana
		for (i=0;i<nxq;i++) eliminations[varelmorder[xq[i]]]->marginalize = 0;
	}
	
	for (i=0, ct_used=0;i<nvars && ct_used<nvars;i++) {
	
		// Atribui os potenciais da operao de eliminao
		// 1. Verifica o n da varivel "da vez"
		if (!used[elmorder[i]]) {
			id = elmorder[i];
			if (findings && findings[id]) {
				// Marginaliza a distribuio de probabilidade se necessrio
				potential =(bayesnet->potentials[id]->nvars>1)?marginalizeNotRequired(moral,bayesnet->potentials[id]):NULL;
				// Produto do finding pela distribuio de probabilidade
				sllappend(&eliminations[i]->potentials,multPotentials((potential?potential:bayesnet->potentials[id]),findings[id]));
				// Libera o espao alocado (se necessrio)
				if (potential) destroyPotential(&potential);
			} else {
				sllappend(&eliminations[i]->potentials,bayesnet->potentials[id]);
			}
			used[id] = 1;
			ct_used++;
		}
	
		// 2. Verifica os filhos da varivel "da vez"
		current = getVertexChildren(getVertexById(&bayesnet->graph,elmorder[i]))->first;
		while (current!=NULL) {
			if (getVertexById(&moral,current->vertex->id) && !used[current->vertex->id]) {
				id = current->vertex->id;
				if (findings && findings[id]) {
					// Marginaliza a distribuio de probabilidade se necessrio
					potential =(bayesnet->potentials[id]->nvars>1)?marginalizeNotRequired(moral,bayesnet->potentials[id]):NULL;
					// Produto do finding pela distribuio de probabilidade
					sllappend(&eliminations[i]->potentials,multPotentials((potential?potential:bayesnet->potentials[id]),findings[id]));
					// Libera o espao alocado (se necessrio)
					if (potential) destroyPotential(&potential);
				} else {
					sllappend(&eliminations[i]->potentials,bayesnet->potentials[id]);
				}
				used[id] = 1;
				ct_used++;
			}
			current = current->next;
		}
	
		// Fixa o nmero de potenciais fixos na elimiantion
		eliminations[i]->nconst = slllength(eliminations[i]->potentials);
	}	

	//---

	// Tem que atribuir NULL para liberar a memria pois eu fiz um malloc explicito para ela
	eliminations = NULL;

	return threadtree;
}


// Destri a rvore de threads criada
void destroyBNThreadTree(ThreadTree** refThreadTree) {
	
	ThreadTree* threadtree = *refThreadTree;
	Elimination** eliminations;	// Estrutura auxiliar para passar dados para as threads
	int i;

	// Utiliza o primeiro vrtice para liberar o contedo da ThreadData
	// que da forma como foi implementado para eliminao de variveis
	// em uma rede Bayesiana  comum a todos os ThreadVertex
	
	if (sizeThreadTree(threadtree)>0) {
		eliminations  = (Elimination**) getThreadVertex(&threadtree,0)->data;
	
		// Elimina as operaes executadas
		if (eliminations) {
			for (i=0;i<sizeThreadTree(threadtree);i++) destroyElimination(&eliminations[i]);
		}

		free(eliminations);	
		eliminations = NULL;
	}

	// Destri a rvore de threads
	destroyThreadTree(&threadtree);
	*refThreadTree = NULL;
}

// Potential* doBNVariableElimination(Potential** findings, ThreadTree* threadtree) {
Potential* doBNVariableElimination(ThreadTree* threadtree) {

	ThreadVertex* thread = NULL;	// Varivel auxiliar que instancia um vrtive da rvore de threads

	Potential* result = NULL;		// Potencial resultante da sequencia de operaes

	Elimination** eliminations = NULL;	// Array com operes de eliminaes

	int i,nvars;


	// Nmero de variveis
	nvars = sizeThreadTree(threadtree);

#ifdef NOTHREADS
	// Executa as threats uma a uma
	for (i=0;i<nvars;i++) {
		thread = getThreadVertex(&threadtree,i);
		thread->status = THREAD_READY;
		run(&thread);
		join(&thread);
	}
#else
	// Executa a rvore como um todo para fazer a inferncia
	// execute(&threadtree);
	restart(&threadtree);
	joinThreadTree(&threadtree);

#endif

	// Recupera os dados de execuo das threads
	eliminations = (Elimination**) getThreadVertex(&threadtree,0)->data;

	// Copia o resultado da ltima eliminao
	result = buildPotential(-1,eliminations[(nvars-1)]->result->nvars,eliminations[(nvars-1)]->result->vars,eliminations[(nvars-1)]->result->values);

	// Normaliza o resultado final se necessrio
	if (result!=NULL) normalize(result->nvalues,&result->values);

	return result;
}

Potential* doBNLinearVariableElimination(int nxq, int* xq, int nxe, int* xe, Potential** findings, BayesNet* bayesnet, int* elmorder, Graph* moral) {

	Variable* first = NULL;		// Primeira varivel a ser eliminada do potencial

	VertexNode* current = NULL;		// Varivel auxiliar para percorrer o Moral Graph

	Potential* result = NULL;		// Potencial resultante da sequencia de operaes
	Potential* potential = NULL;	// Varivel potencial auxiliar

	Elimination** eliminations = NULL;	// Array com operes de eliminaes

	int varelmorder[bnsize(bayesnet)];	// Mapa o id da varivel para a ordem de eliminao
	int used[bnsize(bayesnet)];			// Vetor de flag para indicar se um potencial foi utilizado

	SingleLinkedListNode* sllcurrent;

	int i,j,id;
	int nvars, ct_used;

	// Nmero de variveis
	nvars = graphsize(moral);

	// Aloca o espao necessrio para o array de operaes
	eliminations = malloc(nvars*sizeof(Elimination*));

	for (i=0;i<nvars;i++) {
		// Inicializa o mapa
		varelmorder[elmorder[i]] = i;

		// Inicializa a flag
		used[elmorder[i]] = 0;

		// Monta as eliminaes e popula varelmorder
		eliminations[i] = (Elimination*) buildElimination(i,bayesnet->variables[elmorder[i]],1);
	}

	// Tomando Xq, se no NULL, marca as operaes que no devem ser marginalizadas
	if (xq!=NULL) { // Apenas por segurana
		for (i=0;i<nxq;i++) eliminations[varelmorder[xq[i]]]->marginalize = 0;
	}

	for (i=0, ct_used=0;i<nvars && ct_used<nvars;i++) {

		// Atribui os potenciais da operao de eliminao
		// 1. Verifica o n da varivel "da vez"
		if (!used[elmorder[i]]) {
			id = elmorder[i];
			if (findings && findings[id]) {
				// Marginaliza a distribuio de probabilidade se necessrio
				potential =(bayesnet->potentials[id]->nvars>1)?marginalizeNotRequired(moral,bayesnet->potentials[id]):NULL;
				// Produto do finding pela distribuio de probabilidade
				sllappend(&eliminations[i]->potentials,multPotentials((potential?potential:bayesnet->potentials[id]),findings[id]));
				// Libera o espao alocado (se necessrio)
				if (potential) destroyPotential(&potential);
			} else {
				sllappend(&eliminations[i]->potentials,bayesnet->potentials[id]);
			}
			used[id] = 1;
			ct_used++;
		}

		// 2. Verifica os filhos da varivel "da vez"
		current = getVertexChildren(getVertexById(&bayesnet->graph,elmorder[i]))->first;
		while (current!=NULL) {
			if (getVertexById(&moral,current->vertex->id) && !used[current->vertex->id]) {
				id = current->vertex->id;
				if (findings && findings[id]) {
					// Marginaliza a distribuio de probabilidade se necessrio
					potential =(bayesnet->potentials[id]->nvars>1)?marginalizeNotRequired(moral,bayesnet->potentials[id]):NULL;
					// Produto do finding pela distribuio de probabilidade
					sllappend(&eliminations[i]->potentials,multPotentials((potential?potential:bayesnet->potentials[id]),findings[id]));
					// Libera o espao alocado (se necessrio)
					if (potential) destroyPotential(&potential);
				} else {
					sllappend(&eliminations[i]->potentials,bayesnet->potentials[id]);
				}
				used[id] = 1;
				ct_used++;
			}
			current = current->next;
		}
	}

	// Executa linearmente a eliminao de variveis
	for (i=0;i<nvars;i++) {

		// Executa a eliminao
		executeElimination(&eliminations[i]);

		// Adiciona o resultado a uma operao futura
		if (i<(nvars-1)) {
			// Identifica em cada potencial a primeira varivel a ser eliminada
			first = NULL;
			potential = eliminations[i]->result;
			for (j=0;j<potential->nvars;j++) {
				if (varelmorder[potential->vars[j]->id]>i) {
					if (first==NULL || varelmorder[potential->vars[j]->id] < varelmorder[first->id]) first = potential->vars[j];
				}
			}
			// Adiciona o potential  operao de eliminao correspondente  prxima varivel a ser eliminada
			if (first!=NULL) sllappend(&eliminations[varelmorder[first->id]]->potentials,potential);
		}
	}

	// Copia o resultado da ltima eliminao
	result = buildPotential(-1,eliminations[(nvars-1)]->result->nvars,eliminations[(nvars-1)]->result->vars,eliminations[(nvars-1)]->result->values);

	// Normaliza o resultado final se necessrio
	if (result!=NULL) normalize(result->nvalues,&result->values);

	// Desaloca a memria alocada para esta iterao preservando as demais estruturas

	// Elimina as operaes executadas
	for (i=0;i<nvars;i++) destroyElimination(&eliminations[i]);
	free(eliminations);
	eliminations = NULL;

	return result;
}

Potential* marginalizeNotRequired(Graph* moral, Potential* source) {

	Potential* potential = NULL;
	Potential* marginalized = NULL;
	Variable* var = NULL;

	int i;

	for (i=0;i<source->nvars;i++) {
		var = source->vars[i];
		if (!getVertexById(&moral,var->id)) {
			// Inicializa a varivel auxiliar
			if (potential==NULL) potential = buildPotential (source->id, source->nvars, source->vars, source->values);

			// Marginalizao
			marginalized = marginalizePotential(var,potential);

			// Destri o potencial temporrio
			destroyPotential(&potential);

			potential = marginalized;
			marginalized = NULL;
		}
	}
	// Libera memria
	var = NULL;

	return potential;
}


void dumpEliminationPotentials(Elimination* elimination) {

	SingleLinkedListNode* current;
	Potential* potential;

	int i,j;

	fprintf(DBGOUT,"elimination[%d]->potentials = ",elimination->id);	
	current = elimination->potentials->first;
	while (current!=NULL) {
		potential = (Potential*) current->data; 
		fprintf(DBGOUT,"(");	
		for (j=0;j<potential->nvars;j++) fprintf(DBGOUT,"%s%s",potential->vars[j]->name,(j<potential->nvars-1?",":")"));	
		current = current->next;
	}
	fprintf(DBGOUT,"\n");	
}

void *taskVariableElimination(void* arg) {

	ThreadVertex* thread  = (ThreadVertex*) arg;

	Elimination** eliminations;

	int i,n;

	// Aguarda a execuo dos descendentes
	joinDescendants(&thread);

	// Inicializa as variveis a serem utilizadas. Fiz desta forma apenas para ficar mais claro o cdigo
	eliminations = (Elimination**) thread->data;

#ifdef DUMPELM
	fprintf(DBGOUT,"\nDEBUG: thread->id %d\n",thread->id);
	dumpElimination(eliminations[thread->id]);
#endif

	// Executa a eliminao
	executeElimination(&eliminations[thread->id]);

#ifdef DUMPELM
	fprintf(DBGOUT,"Result of thread->id %d to eliminations[%d]\n",thread->id,(thread->parent!=NULL?thread->parent->id:-1));
	dumpPotential(eliminations[thread->id]->result);
#endif

	// Adiciona o resultado a uma operao futura de forma sincronizada
	if (thread->parent!=NULL) {
		sllsyncpush(&eliminations[thread->parent->id]->potentials,eliminations[thread->id]->result,&(eliminations[thread->parent->id]->mutex));
	}

	// Remove o n primeiros elementos do potencial
	n = slllength(eliminations[thread->id]->potentials) - eliminations[thread->id]->nconst;
	for (i=0;i<n;i++) {
		sllremove_first(&eliminations[thread->id]->potentials);
	}

	return NULL;
}

void executeElimination(Elimination** refElimination) {

	Elimination* elimination = *refElimination;	
	SingleLinkedListNode* current;

	Potential* marginalized = NULL;
	Potential* result = NULL;
	Potential* A;
	Potential* B;

	current = elimination->potentials->first;
	if (slllength(elimination->potentials)>1) {

		// Executa o produto dos potenciais 2 a 2

		// Executa o primeiro produto
		A = (Potential*)current->data;
		current = current->next;
		B = (Potential*)current->data;
		result = multPotentials(A,B);

#ifdef DUMPELM
	fprintf(DBGOUT,"\tResult:\n");
	dumpPotential(result);
#endif

		// Executa os demais produtos
		current = current->next;
		while (current!=NULL) {
			A = result;
			result = multPotentials(A,(Potential*) current->data);

#ifdef DUMPELM
	fprintf(DBGOUT,"\tResult:\n");
	dumpPotential(result);
#endif

			destroyPotential(&A);
			current = current->next;
		}

		// Faz a marginalizao
		if (elimination->marginalize==1) {
			// Faz a marginalizao
			marginalized = marginalizePotential(elimination->variable,result);

			// Cria ou atualiza o result da eliminao
			if (elimination->result==NULL) {
				// Cria o novo potencial
				elimination->result = buildPotential(-1,marginalized->nvars,marginalized->vars,marginalized->values);
			} else { 
				// Apenas copia os novos valores
				memcpy(elimination->result->values,marginalized->values,marginalized->nvalues*sizeof(double));
			}

		} else {
			// Cria ou atualiza o result da eliminao
			if (elimination->result==NULL) {
				// Cria o novo potencial
				elimination->result = buildPotential(-1,result->nvars,result->vars,result->values);
			} else { 
				// Apenas copia os novos valores
				memcpy(elimination->result->values,result->values,result->nvalues*sizeof(double));
			}
		}


	} else {
		// Faz a marginalizao
		if (elimination->marginalize==1) {
			marginalized = marginalizePotential(elimination->variable,(Potential*)current->data);

			// Cria ou atualiza o result da eliminao
			if (elimination->result==NULL) {
				// Cria o novo potencial
				elimination->result = buildPotential(-1,marginalized->nvars,marginalized->vars,marginalized->values);
			} else { 
				// Apenas copia os novos valores
				memcpy(elimination->result->values,marginalized->values,marginalized->nvalues*sizeof(double));
			}

		} else {
			// Cria ou atualiza o result da eliminao
			if (elimination->result==NULL) {
				// Cria o novo potencial
				elimination->result = buildPotential(-1,((Potential*)current->data)->nvars,((Potential*)current->data)->vars,((Potential*)current->data)->values);
			} else { 
				// Apenas copia os novos valores
				memcpy(elimination->result->values,((Potential*)current->data)->values,((Potential*)current->data)->nvalues*sizeof(double));
			}
		}
	}

	// Desloca final
	if (marginalized!=NULL) destroyPotential(&marginalized);

	// Desaloca o result
	if (result!=NULL) destroyPotential(&result);

	// Para garantir...
	current = NULL;
}

void normalize(int nvalues, double** refValues) {
	double* values = *refValues;
	double sum = 0;

	int i;
	
	for (i=0;i<nvalues;i++) sum = sum + values[i];
	
	if (sum!=1) for (i=0;i<nvalues;i++) values[i] = values[i]/sum;	
}


