/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	BNVarElimination.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Junho/2007 (2007-06-28)
 * @brief 	Biblioteca para fazer inferncia em Redes Bayesianas (BN: \e Bayesian \e Network) utilizando o mtodo de eliminao de variveis.
 *
 */

#ifndef DBGOUT
#define DBGOUT stdout
#endif

#ifndef BNVARELIMINATION_H
#define BNVARELIMINATION_H

/* Includes */
#include <math.h>
#include <pthread.h> 

/* Bibliotecas especficas */
#include "BNMain.h"
#include "graphtools.h"
#include "threadtree.h"
#include "singlelinkedlist.h"

/* Typedefs */

/*! Estrutura de dados para definir uma operao de eliminao */
typedef struct struct_Elimination {
	int id;							//!< \e id da operao de eliminao. Manipulado internamente pela biblioteca no deve ser alterado.
	int marginalize;				//!< Flag para indicar se deve fazer a operao de marginalizao: 1:marginaliza; 0:faz o produto mas no marginaliza.
	int nconst;						//!< Parte constante da lista de potenciais da eliminao. Correspondem aos potenciais originais da Rede Bayesiana
	Variable* variable;				//!< Varivel que ser eliminada na operao.
	SingleLinkedList* potentials;	//!< Potenciais envolvidos na operao.
	Potential* result;				//!< Potencial resultado da eliminao.
	pthread_mutex_t mutex;			//!< Multex para garantir a sincronia das operaes na eliminao	
} Elimination;


// METHODS

void scale(int nvalues, double** refValues);


/*!
 * @brief Constri a rvore de eliminao de \e threads. A rvore de eliminao uma vez definida pode ser reutilizada para executar
 * diversas inferncias.
 * @param[in] elmorder \e Array com os \e ids das variveis na ordem que devem ser eliminadas.
 * @param[in] elmtree Grafo que representa a rvore de eliminao para resultante da fatorao simblica.
 * @return \e ThreadTree construida a partir da ordem de eliminao e da rvore de eliminao.
 */

// TODO: Atualizar documentao
// ThreadTree* buildBNThreadTree(int nxq, int* xq, int nxe, int* xe, Potential** findings, BayesNet* bayesnet, Graph* moral, int* elmorder, Graph* elmtree);
// ThreadTree* buildBNThreadTree(BayesNet* bayesnet, int* elmorder,Graph* elmtree);
ThreadTree* buildBNThreadTree(int nxq, int* xq, int nxe, int* xe, Potential** findings,BayesNet* bayesnet, int* elmorder, Graph* moral, Graph* elmtree);

/*!
 * @brief Destri a rvore de threads criada para executar a inferncia.
 * @param[out] refThreadTree Referncia para a \e ThreadTree a ser destruida.
 */

void destroyBNThreadTree(ThreadTree** refThreadTree);

/*!
 * @brief Executa a inferncia com a eliminao de variveis da Rede Bayesiana explorando a possibilidade de executar a eliminao em paralelo.
 * @param[in] nxq Nmero de variveis questionadas (\e Query \e Variables).
 * @param[in] xq \e Array com os \e ids das variveis questionadas (\e Query \e Variables).
 * @param[in] nxe Nmero de variveis que compe a evidncia (\e Evidence), ou seja, as variveis observadas. O estado da varivel esta
 * em outra estrutura denominada \e finding.
 * @param[in] xe \e Array com os \e ids das variveis observadas.
 * @param[in] findings Potenciais que indicam os estados observados das variveis.
 * @param[in] bayesnet BN da qual se deseja obter as variveis requisitadas para a inferncia.
 * @param[in] elmorder \e Array com os \e ids das variveis na ordem que devem ser eliminadas.
 * @param[in] moral \e Moral \e Graph utilizado para a eliminao de variveis e construido apenas com variveis requisitadas.
 * @param[in] threadtree \e ThreadTree construida a partir da ordem de eliminao e da rvore de eliminao.
 * @return Potencial resultante do processo de eliminao de variveis do grafo e portanto o resultado da inferncia.
 */

// Potential* doBNVariableElimination(int nxq, int* xq, int nxe, int* xe, Potential** findings, BayesNet* bayesnet, int* elmorder, Graph* moral, ThreadTree* threadtree);
// Potential* doBNVariableElimination(Potential** findings, ThreadTree* threadtree);
Potential* doBNVariableElimination(ThreadTree* threadtree);

/*!
 * @brief Executa linearmente a inferncia com a eliminao de variveis da Rede Bayesiana, ou seja, apenas uma varivel  eliminada de cada vez
 * de forma serial, sem explorar a possibilidade de paralelizao.
 * @param[in] nxq Nmero de variveis questionadas (\e Query \e Variables).
 * @param[in] xq \e Array com os \e ids das variveis questionadas (\e Query \e Variables).
 * @param[in] nxe Nmero de variveis que compe a evidncia (\e Evidence), ou seja, as variveis observadas. O estado da varivel esta
 * em outra estrutura denominada \e finding.
 * @param[in] xe \e Array com os \e ids das variveis observadas.
 * @param[in] findings Potenciais que indicam os estados observados das variveis.
 * @param[in] bayesnet BN da qual se deseja obter as variveis requisitadas para a inferncia.
 * @param[in] elmorder \e Array com os \e ids das variveis na ordem que devem ser eliminadas.
 * @param[in] moral \e Moral \e Graph utilizado para a eliminao de variveis e construido apenas com variveis requisitadas.
 * @return Potencial resultante do processo de eliminao de variveis do grafo e portanto o resultado da inferncia.
 */

Potential* doBNLinearVariableElimination(int nxq, int* xq, int nxe, int* xe, Potential** findings, BayesNet* bayesnet, int* elmorder, Graph* moral);

/*!
 * @brief Funo genrica invocada pela \e thread para executar uma operao de eliminao.
 * @param[out] arg Argumento genrico utilizada para passar literamente qualquer estrutura de dados necessria para executar a tarefa.
 */

void *taskVariableElimination(void* arg);

/*!
 * @brief Executa a operao de eliminao de varivel e armazena o resultado na prpria estrutura de dados
 * @param[out] refElimination Referncia para a estrutura que define uma operao de eliminao. O resultado ser armazenado no
 * na prpria estrutura.
 */

void executeElimination(Elimination** refElimination);

/*!
 * @brief Normaliza um \e array de valores reais, ou seja, altera os valores de forma que a soma dos mesmos seja igual a 1.
 * @param[in] nvalues Nmero de elementos do \e array de nmeros reais que ser normalizado.
 * @param[out] refValues Referncia para o \e array de nmeros reais que ser normalizado.
 */

void normalize(int nvalues, double** refValues);

/*!
 * @brief Marginaliza as variveis no requeridas de um potencial.
 *
 * Ateno: Apesar do valor de retorno da funo de ser pouco intuitivo ela foi feita desta forma para evitar operaes desnecessrias.
 * @param[in] moral \e Moral \e Graph utilizado para a eliminao de variveis e construido apenas com variveis requisitadas.
 * @param[in] source Potencial a ser marginalizado.
 * @return Potencial marginalizado ou NULL se no for feita nenhuma marginalizao.
 */

Potential* marginalizeNotRequired(Graph* moral, Potential* source);

#endif
