/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifndef BNLOADER_H_INCLUDED
#include <BNLoader.h>
#endif

#ifndef BNUTILS_H_INCLUDED
#include <BNUtils.h>
#endif

#ifndef BNVARELIMINATION_H_INCLUDED
#include <BNVarElimination.h>
#endif


int main(int argc, char **argv) {

	if (argc<2) {
		fprintf(stderr,"Usage: %s -b <BayesNet filename> [-q <Query filename> -e <Evidences filename> -N <# Complete iterations> -L <# Inferece interations> -v <Verbose Level 1-5>]\n",argv[0]);
		exit(1);
	}

	// Nome dos arquivos
	char* bnfilename = NULL;
	char* queryfilename = NULL;
	char* evidencesfilename = NULL;

	// Rede Bayesiana
	BayesNet* bayesnet = NULL;

	// Potencial que armazena o resultado
	Potential* result = NULL;

	// Variveis questionadas (Xq)
	int* xq = NULL;
	int nxq = 0;

	// Variveis observadas
	int* xe = NULL;
	int nxe = 0;
	Potential** findings = NULL;

	// Grafos auxiliares
	Graph* moral = NULL;		// Moral Graph
	Graph* filled = NULL;		// Filled Graph

	// Ordem de eliminao
	int* elmorder = NULL;

	// Parmetros de entrada de linha de comando
	long int I,N = 1;
	long int l,L = 1;
	int verbose = 0;

	// Controle do tempo
	time_t Nt0,Nt1;
	time_t Lt0,Lt1;

	// Variveis auxiliares
	int i;

	// Verifica os parmetros da linha de comando
	for (i=1;i<argc;i++) {
		// -b <Arquivo de definio da rede bayesiana>
		if (strcmp(argv[i],"-b")==0) {
			bnfilename = argv[++i];
			continue;
		}

		// -q <Arquivo com a definio da query>
		if (strcmp(argv[i],"-q")==0) {
			queryfilename = argv[++i];
			continue;
		}

		// -e <Arquivo com as evidncias observadas>
		if (strcmp(argv[i],"-e")==0) {
			evidencesfilename = argv[++i];
			continue;
		}

		// -N <Nmero de iteraes completas> 
		if (strcmp(argv[i],"-N")==0) {
			N = atoi(argv[++i]);
			continue;
		}

		// -L <Nmero de execues da inferencia> 
		if (strcmp(argv[i],"-L")==0) {
			L = atoi(argv[++i]);
			continue;
		}

		// -v verbose
		if (strcmp(argv[i],"-v")==0) {
			verbose = atoi(argv[++i]);
			continue;
		}
	}

	// Arquivo de definio da rede  o nico parmetro obrigatrio
	if (bnfilename==NULL) {
		fprintf(stderr,"Bayesian Network definition file must be an input (bnfilename==NULL)\n");
		exit(1);
	}

	// Carrega a rede
	bayesnet = bnload(bnfilename);
	if (verbose>1) {
		printf("\nBayes Net:\n");
		dumpBayesNet(bayesnet);
	}

	// Carrega Xq do arquivo, se estiver definido
	if (queryfilename!=NULL) {
		xq = bnloadquery(&nxq,queryfilename,bayesnet);	
		if (verbose>1) {
			printf("\nQuery loaded from %s: [",queryfilename);
			for (i=0;i<nxq;i++) printf("%s%s",bayesnet->variables[xq[i]]->name,(i!=nxq-1)?",":""); printf("]\n");
		}
	}

	// Carrega Xe e findings do arquivo, se estiver definido
	if (evidencesfilename!=NULL) {
		findings = malloc(bnsize(bayesnet)*sizeof(Potential*));
		xe = bnloadevidences(&nxe,findings,evidencesfilename,bayesnet);	
		if (verbose>1) {
			printf("\n%d Findings loaded from %s\n",nxe,evidencesfilename);
			for (i=0;i<nxe;i++) dumpPotential(findings[xe[i]]);
		}
	}

	time(&Nt0);
	for (I=0;I<N;I++) {

		// Contador e controle do tempo
		time(&Nt1);
		fprintf(stderr,"#execues I=%5ld timer:%4.0fs\n",I,difftime(Nt1,Nt0));
		Nt0 = Nt1;

		// Gera o Moral Graph
		moral = buildBNMoralGraph(bayesnet);
		if (verbose>1) {
			printf("\nMoral Graph:\n");
			dumpGraph(moral);
		}

		// Gera a ordem de eliminao
		elmorder = defineBNEliminationOrder(moral);
		if (verbose>1) {
			printf("\nElimination order: [");
			for (i=0;i<bnsize(bayesnet);i++) printf("%3d%s",elmorder[i],(i!=bnsize(bayesnet)-1)?",":""); printf("]\n");
		}

		time(&Lt0);
		for (l=0;l<L;l++) {
			if ((l%100)==0) {
				time(&Lt1);
				fprintf(stderr,"#inferncias l=%5d timer:%4.0fs\n",l,difftime(Lt1,Lt0));
				Lt0 = Lt1;
			}

			// Executa a inferncia utilizando o mtodo paralelizado
			result = (Potential*) doBNLinearVariableElimination(nxq,xq,nxe,xe,findings,bayesnet,elmorder);
			
			if (verbose>0) {
				printf("\nResultado:\n");
				dumpPotential(result);
			}

			// Desaloca o resultado
			destroyPotential(&result);
			result = NULL;
		}

		// Desaloca as variveis

		// Ordem de eliminao
		free(elmorder);

		// Grafos auxiliares
		destroyGraph(&moral);

	}

	// Varivies questionadas
	if (xq!=NULL) {
		free(xq);
		xq = NULL;
	}

	// Variveis observadas
	if (xe!=NULL) {
		free(xe);
		xe = NULL;
		for (i=0;i<bnsize(bayesnet);i++) if (findings[i]!=NULL) destroyPotential(&findings[i]);
		free(findings);
		findings = NULL;
	}

	// Desaloca a rede
	destroyBayesNet(&bayesnet);
}
