/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


// Cdigo derivado de: test_BNVarEliminationSubThreadtree.c

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/times.h>

#include "BNLoader.h"
#include "BNUtils.h"
#include "BNVarElimination.h"

int main(int argc, char **argv) {

	if (argc<2) {
		fprintf(stderr,"Usage: %s -b <BayesNet filename> [-q <Query filename> -e <Evidences filename> -N <# Complete iterations>  -L <# Inferece interations> -v <Verbose Level 1-5>]\n",argv[0]);
		exit(1);
	}

	// Nome dos arquivos
	char* bnfilename = NULL;
	char* queryfilename = NULL;
	char* evidencesfilename = NULL;

	// Rede Bayesiana
	BayesNet* bayesnet = NULL;

	// Potencial que armazena o resultado
	Potential* result = NULL;

	// Variveis questionadas (Xq)
	int* xq = NULL;
	int nxq = 0;

	// Variveis observadas
	int* xe = NULL;
	int nxe = 0;
	Potential** findings = NULL;

	// Variveis requeridas
	int* xr = NULL;
	int nxr = 0;

	// Grafos auxiliares
	Graph* moral = NULL;		// Moral Graph
	Graph* filled = NULL;		// Filled Graph
	Graph* elmtree = NULL;		// Elimination Tree

	// Arvore de Threads
	ThreadTree* threadtree = NULL;

	// Ordem de eliminao
	int* elmorder = NULL;

	// Parmetros de entrada de linha de comando
	long int I,N = 1;
	long int l,L = 1;
	int verbose = 0;

	// Controle do tempo
	
	// Procedimentos completos
	clock_t startN, endN;
	struct tms tms_startN, tms_endN;

	// Execuo de uma inferncia
	clock_t startL, endL;
	struct tms tms_startL, tms_endL;

	// Variveis Auxiliares
	double timeN[2], timeL[2]; // timeX[0] = tms_cutime, timeX[1] = tms_cstime

	// Variveis auxiliares
	int i;

	// Verifica os parmetros da linha de comando
	for (i=1;i<argc;i++) {
		// -b <Arquivo de definio da rede bayesiana>
		if (strcmp(argv[i],"-b")==0) {
			bnfilename = argv[++i];
			continue;
		}

		// -q <Arquivo com a definio da query>
		if (strcmp(argv[i],"-q")==0) {
			queryfilename = argv[++i];
			continue;
		}

		// -e <Arquivo com as evidncias observadas>
		if (strcmp(argv[i],"-e")==0) {
			evidencesfilename = argv[++i];
			continue;
		}

		// -N <Nmero de iteraes completas> 
		if (strcmp(argv[i],"-N")==0) {
			N = atoi(argv[++i]);
			continue;
		}

		// -L <Nmero de execues da inferencia> 
		if (strcmp(argv[i],"-L")==0) {
			L = atoi(argv[++i]);
			continue;
		}

		// -v verbose
		if (strcmp(argv[i],"-v")==0) {
			verbose = atoi(argv[++i]);
			continue;
		}
	}

	// Arquivo de definio da rede  o nico parmetro obrigatrio
	if (bnfilename==NULL) {
		fprintf(stderr,"Bayesian Network definition file must be an input (bnfilename==NULL)\n");
		exit(1);
	}

	// Carrega a rede
	bayesnet = bnload(bnfilename);
	if (verbose>3) {
		printf("\nBayes Net:\n");
		dumpBayesNet(bayesnet);
	}

	// Carrega Xq do arquivo, se estiver definido
	if (queryfilename!=NULL) {
		xq = bnloadquery(&nxq,queryfilename,bayesnet);	
		if (verbose>1) {
			printf("\nQuery loaded from %s: [",queryfilename);
			for (i=0;i<nxq;i++) printf("%s%s",bayesnet->variables[xq[i]]->name,(i!=nxq-1)?",":""); printf("]\n");
		}
	}

	// Carrega Xe e findings do arquivo, se estiver definido
	if (evidencesfilename!=NULL) {
		findings = malloc(bnsize(bayesnet)*sizeof(Potential*));
		xe = bnloadevidences(&nxe,findings,evidencesfilename,bayesnet);	
		if (verbose>1) {
			printf("\n%d Findings loaded from %s\n",nxe,evidencesfilename);
			for (i=0;i<nxe;i++) dumpPotential(findings[xe[i]]);
		}
	}

	// Inicia a soma do tempo de inferncia
	timeN[0] = 0.0; timeN[1] = 0.0;
	timeL[0] = 0.0; timeL[1] = 0.0;

	// Loop de procedimentos completos
	for (I=0;I<N;I++) {

		// TIME: Start do tempo de execuo dos procedimentos completos
		startN = times(&tms_startN);

		// Contador e controle do tempo
		if (I%1==0 && verbose > 3) fprintf(stderr,"#execues I=%5ld\n",I);

		// Determina as variveis requeridas
		xr = (int*) defineBNRequiredVars(&nxr,nxq,xq,nxe,xe,bayesnet);
		if (verbose>1) {
			printf("\nRequired: [");
			for (i=0;i<nxr;i++) printf("%d%s",xr[i],(i!=nxr-1)?",":""); printf("]\n");
		}

		// Gera o Moral Graph
		moral = buildBNMoralGraph(nxr,xr,bayesnet);
		if (verbose>1) {
			printf("\nMoral Graph:\n");
			dumpGraph(moral);
		}

		// Gera a ordem de eliminao
		elmorder = defineBNEliminationOrder(moral);

		if (verbose>1) {
			printf("\nElimination order: [");
			for (i=0;i<graphsize(moral);i++) printf("%3d%s",elmorder[i],(i!=graphsize(moral)-1)?",":""); printf("]\n");
		}

		// Faz a fatorao simblica: rvore de Eliminao e Grafo Preenchido
		elmtree = newGraph(DIRECTED_GRAPH);
		filled = newGraph(NOT_DIRECTED_GRAPH);
		doBNSymbolicFactorization(&elmtree,&filled,moral,elmorder);
		
		if (verbose>1) {
			printf("\nElimination Tree:\n");
			dumpGraph(elmtree);

			printf("\nFilled Graph:\n");
			dumpGraph(filled);
		}

		// Monta a rvore de threads
		threadtree = (ThreadTree*) buildBNThreadTree(nxq,xq,nxe,xe,findings,bayesnet,elmorder,moral,elmtree);

		if (verbose>1) {
			printf("\nThread Tree:\n");
			dumpThreadTree(threadtree);
		}


		// Loop dos procedimentos de inferncia
		for (l=0;l<L;l++) {
			if ((l%100)==0  && verbose > 3) fprintf(stderr,"#inferncias l=%5d\n",l);

			// TIME: Start do tempo de execuo dos procedimentos de inferncia
			startL = times(&tms_startL);

			// Executa a inferncia utilizando o mtodo paralelizado
			result = (Potential*) doBNVariableElimination(threadtree);
			
			// TIME: End do tempo de execuo dos procedimentos de inferncia
			endL = times(&tms_endL);

			// TIME: Contabiliza o tempo dos procedimentos de inferncia
			timeL[0] = timeL[0] + (double)(tms_endL.tms_utime-tms_startL.tms_utime) + (double)(tms_endL.tms_cutime-tms_startL.tms_cutime);
			timeL[1] = timeL[1] + (double)(tms_endL.tms_stime-tms_startL.tms_stime) + (double)(tms_endL.tms_cstime-tms_startL.tms_cstime);

			if (verbose>0) {
				printf("\nResultado:\n");
				dumpPotential(result);
			}

			// Desaloca as variveis
		
			// Desaloca o resultado
			destroyPotential(&result);
			result = NULL;
		}

		// Destri a rvore de threads
		destroyBNThreadTree(&threadtree);

		// Grafos auxiliares
		destroyGraph(&filled);
		destroyGraph(&elmtree);

		// Ordem de eliminao
		free(elmorder);
		elmorder = NULL;

		// Desaloca o Moral Graph
		destroyGraph(&moral);

		// Variveis requeridas
		free(xr);
		xr = NULL;

		/*
		clock_t start,end;
		struct tms tms_start,tms_end;
		int j,k,m;

		start = times(&tms_start);
		for (j=0;j<10000;j++) {
			for (k=0;k<10000;k++) {
				m = (int) (sqrt(j)+sqrt(k));
			}
		}
		end = times(&tms_end);

		printf("delay: %.2f\n",(double)(end-start));
		printf("delay (tms): %.2f\n",(double)(tms_end.tms_cutime-tms_start.tms_cutime));
		
		*/

		// TIME: End do tempo de execuo dos procedimentos completos
		endN = times(&tms_endN);

		// TIME: Contabiliza o tempo dos procedimentos completos
		timeN[0] = timeN[0] + (double)(tms_endN.tms_utime-tms_startN.tms_utime) + (double)(tms_endN.tms_cutime-tms_startN.tms_cutime);
		timeN[1] = timeN[1] + (double)(tms_endN.tms_stime-tms_startN.tms_stime) + (double)(tms_endN.tms_cstime-tms_startN.tms_cstime);
	}

	// Varivies questionadas
	if (xq!=NULL) {
		free(xq);
		xq = NULL;
	}

	// Variveis observadas
	if (xe!=NULL) {
		free(xe);
		xe = NULL;
		for (i=0;i<bnsize(bayesnet);i++) if (findings[i]!=NULL) destroyPotential(&findings[i]);
		free(findings);
		findings = NULL;
	}

	// Desaloca a rede
	destroyBayesNet(&bayesnet);

	// Resumo dos tempos
	if (verbose == -1) {
		printf ("TIME: Total de iteraes N = %d tempo total: User = %.2f Sys = %.2f Tot = %.2f tempo mdio: User = %.3f Sys = %.3f Tot = %.2f\n",N,timeN[0],timeN[1],timeN[0]+timeN[1],timeN[0]/((double)N),timeN[1]/((double)N),(timeN[0]+timeN[1])/((double)N));
		printf ("TIME: Total de inferncias L = %d tempo total: User = %.2f Sys = %.2f Tot = %.2f tempo mdio: User = %.3f Sys = %.3f Tot = %.2f\n",N*L,timeL[0],timeL[1],timeL[0]+timeL[1],timeL[0]/((double)(N*L)),timeL[1]/((double)(N*L)),(timeL[0]+timeL[1])/((double)(L*N)));
		printf ("TIME: Resumo N = %d Ntotal: User = %.2f Sys = %.2f Tot = %.2f Nmedio: User = %.3f Sys = %.3f Tot = %.2f L = %d Ltotal: User = %.2f Sys = %.2f Tot = %.2f Lmedio: User = %.3f Sys = %.3f Tot = %.2f\n",N,timeN[0],timeN[1],timeN[0]+timeN[1],timeN[0]/((double)N),timeN[1]/((double)N),(timeN[0]+timeN[1])/((double)N),N*L,timeL[0],timeL[1],timeL[0]+timeL[1],timeL[0]/((double)(N*L)),timeL[1]/((double)(N*L)),(timeL[0]+timeL[1])/((double)(N*L)));
	}

}
