/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	gibbs.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Maro/2007
 * @brief 	Implementao da Heurstica BEL/Gibbs para definir a ordem de eliminao
 *
 */

#ifndef DBGOUT
#define DBGOUT stdout
#endif

#ifndef GIBBS_H
#define GIBBS_H

/* Includes */
#include "graphtools.h"

/* Defines */

#define nextSeparator nextSeparator_SEPARADOR_PONDERADO //!< Define a estratgia de clculo do separador com o qual foi compilado

#define visited lib[0].I  	//!< Redefine um campo da struct Graph utilizado para construir a rvore de disseco, ajuda a identificar os separadores.
#define list lib[1].L		//!< Redefine um campo da struct Graph para armazenar lista de vrtices que compe o vrtices da rvore de disseco.
#define order lib[2].I		//!< Redefine um campo da struct Graph utilizado para montar a ps-ordem da rvore de disseco.

/* Definio dos tipos de dados */

/*! rvore gerada pela Busca Em Largura (BEL) */
typedef struct struct_BELTree {
	int width;									//!< Largura da rvore BEL. 
	int depth;									//!< Profundidade da rvore BEL.
	int* sum;									//!< Soma dos vrtices da rvore BEL at o \e i-simo nvel. Utilizado para testar o balanceamento da rvore.
	struct struct_Graph* graph;					//!< Grafo para armazenar os vrtices e a estrutura da rvore.
	struct struct_VertexLinkedList** vertices;	//!< Array de Lista ligadas de vrtices correspondentes aos nveis da BEL.
} BELTree;

/* Funes */

/*!
 * @brief Determina a ordem de eliminao de um Grafo Moral.
 * @param[in] moral Grafo Moral do qual a ordem de eliminao deve ser determinada.
 * @return Array com os \e ids dos vrtices na ordem que devem ser eliminados.
 */

int* defineEliminationOrder(Graph* moral);

/*!
 * @brief Executa a heurstica de gibbs para encontrar o n quase perifrico e retorna a BEL para o n escolhido.
 * @param[in] moral Grafo Moral do qual se deseja obter a BEL.
 * @return BEL para o n escolhido quase perifrico escolhido.
 */

BELTree* doGibbsHeuristic(Graph* moral);

/*!
 * @brief Calcula o balanceamento da BEL, ou seja, a relao entre o nmero de vrtices  direita e  esquerda.
 * @param[in] beltree BEL tree da qual se deseja calcular o balanceamento.
 * @return Valor que indica o balanceamento.
 */

double BELBalance (BELTree* beltree);

/*!
 * @brief Calcula a varincia do nmero de vrtices nos nveis da BEL, quando mais uniforme (menor a varincia), melhor.
 * @param[in] beltree BEL tree da qual se deseja calcular a varincia.
 * @return Varincia do nmero de vrtices em cada nvel da BEL
 */

double BELVariance (BELTree* beltree);

/*!
 * @brief Constri a BEL associada a um grafo no deirecionado a partir de um vrtice dado.
 * @param[in] graph Grafo, necessariamente no direcionadao, que ser a base para se obter a BEL.
 * @param[in] vertex Vrtice de origem da BEL.
 * @return BEL para o vrtice e o grafo.
 */

// TODO: fazer o realloc para do nmero de nveis necessrios ser superior ao alocado

BELTree* buildBELTree(Graph* graph, Vertex* vertex);

/*!
 * @brief a BELTree e desaloca todos os recursos alocados.
 * @param[out] refBELTree Referencia para a BEL que deve ser destruida.
 */

void destroyBELTree (BELTree** refBELTree);

/*!
 * @brief Dump do contedo de uma BELTree.
 * @param[in] tree BEL da qual se deseja fazer o dump.
 */

void dumpBELTree (BELTree* tree);

/*!
 * @brief Estratgia bsica para determinao dos nveis que sero os separadores da BEL para o processo de dicotomia.
 *
 * Esta estratgia simplesmente calcula o nvel central de um range de nveis. A estratgia a
 * ser adotada  definida na compilao de acordo com o valor definido para \e nextSeparator.
 * @param[in] beltree BELTree que ser analisada.
 * @param[in] begin Indica o incio do trecho que deve ser dividido.
 * @param[in] end indica o final do trecho que deve ser dividido.
 * @return Nvel da BEL no qual deve ser feita a dicotomia.
 */

int nextSeparator_BASICO(BELTree* beltree, int begin, int end);

/*!
 * @brief Estratgia que pondera o nmero de vrtices em cada nvel para a determinao 
 * dos nveis que sero os separadores da BEL para o processo de dicotomia.
 *
 * Para determinar o separador considera-se o nmero de vrtices que ficar de cada lado aps a separao, 
 * procura-se divirdir igualmente, mas sem considerar a forma como os vrtices esto conectados.
 * A estratgia a ser adotada  definida na compilao de acordo com o valor definido para \e nextSeparator.
 * @param[in] beltree BELTree que ser analisada.
 * @param[in] begin Indica o incio do trecho que deve ser dividido.
 * @param[in] end indica o final do trecho que deve ser dividido.
 * @return Nvel da BEL no qual deve ser feita a dicotomia.
 */

int nextSeparator_SEPARADOR_PONDERADO(BELTree* beltree, int begin, int end);

/*!
 * @brief Estratgia que tenta obtar o separador com o menor nmero de vrtices para determinao 
 * dos nveis que sero os separadores da BEL para o processo de dicotomia.
 *
 * Esta estratgia tenta obter dentre os possveis candidatos para separador, aquele com o nenor nmero de vrtices.
 * A estratgia a ser adotada  definida na compilao de acordo com o valor definido para \e nextSeparator.
 * @param[in] beltree BELTree que ser analisada.
 * @param[in] begin Indica o incio do trecho que deve ser dividido.
 * @param[in] end indica o final do trecho que deve ser dividido.
 * @return Nvel da BEL no qual deve ser feita a dicotomia.
 */

int nextSeparator_MENOR_SEPARADOR(BELTree* beltree, int begin, int end);

/*!
 * @brief Dissecciona um range de vrtices.
 * @param[out] refDSTree Referncia para a rvore de disseco (\e dstree) que est sendo construda.
 * @param[out] refRoot Referncia para o vrtice que ser a raiz.
 * @param[in] beltree BELTree base para a rvore de disseco.
 * @param[in] begin Indica o incio do trecho que deve ser dividido.
 * @param[in] end Indica o final do trecho que deve ser dividido.
 */

void dissection(Graph** refDSTree, Vertex** refRoot, BELTree* beltree, int begin, int end);

/*!
 * @brief Contri a rvore de disseco (\e Dissection \e Tree)com os dados da BEL.
 * @param[in] moral Moral Graph do grafo de origem utilizado para se obter a BEL.
 * @param[in] beltree BELTree com base na qual ser construda a rvore de disseco.
 * @return rvore de disseco.
 */

Graph* buildDissectionTree(Graph* moral, BELTree* beltree);

/*!
 * @brief Destri a rvore de disseco (\e Dissection \e Tree).
 * @param[out] refDSTree referncia para a DSTree que deve ser destruida.
 */

void destroyDissectionTree (Graph** refDSTree);

/*!
 * @brief Dump do contedo de uma rvore de disseco (\e Dissection \e Tree).
 * @param[in] dstree rvore de disseco da qual se deseja fazer o dump.
 */

void dumpDissectionTree (Graph* dstree);

/*!
 * @brief Monta a ordem de eliminao com base na rvore de disseco percorrendo recursivamente os vrtices da rvore de disseco. 
 * @param[in] dstree rvore de disseco.
 * @param[in] vertex Vrtice do qual deve seguir a ordenao.
 * @return Array de inteiros com os \e ids dos vrtices e ps-ordem.
 */

int* buildEliminationOrder(Graph* dstree, Vertex* vertex);

/*!
 * @brief Monta a ps-ordem com base em uma rvore de disseco.
 * @param[in] elmorder Array que progressivamente armazena a ordem de eliminao.
 * @param[out] n Referncia para um nmero inteiro que corresponde  posio atual no array \e elmorder.
 * @param[in] refVertex Referncia para o vrtice a ser visitado.
 */

void searchEliminationOrder(int* elmorder, int *n, Vertex** refVertex);

#endif
