/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Fevereiro/2007
 *
 * Para documentao ver graph.h
 */


/* Lista o cabealho das funes e estrutura de dados */
#include "graph.h"

/* Mensagens de erros e warnings para lib de grafos */
#include "graph_messages.h"

/* Defines */

/* Funes */

Graph* newGraph (int type) {
	Graph* graph = malloc(sizeof(Graph));
	graph->vertices = newVertexLinkedList();
	graph->type = type;
	graph->name = NULL;
	graph->index = NULL;
	return graph;
}

void destroyGraph (Graph** refGraph) {
	Graph* graph = *refGraph;
	VertexNode* current;
	Vertex* vertex;

	/* Libera a memria da lista de index */
	destroyGraphIndex(&graph);

	/* Libera a memria dos vrtices */
	current = graph->vertices->first;
	while (current != NULL) {
		vertex = current->vertex;
		// printf("Vertex: %d %s\n",vertex->id,vertex->name);
		destroyVertex(&vertex);
		current = current->next;
	}

	/* Desaloca a lista de vtices */
	destroyVertexLinkedList(&(graph->vertices));

	/* Libera Graph */
	free(graph);
	*refGraph = NULL;
}

Vertex* newVertex () {
	Vertex* vertex = malloc(sizeof(Vertex));
	vertex->id = -1;
	vertex->name = NULL;
	vertex->graph = NULL;
	vertex->parents = newVertexLinkedList();
	vertex->children = newVertexLinkedList();

	return vertex;
}

Vertex* buildVertex (int id, char* name) {
	Vertex* vertex = newVertex();
	vertex->id = id;
	vertex->name = malloc(VERTEX_NAME_MAX_LENGTH*sizeof(char));
	strcpy(vertex->name,name);
	return vertex;
}

void destroyVertex (Vertex** refVertex) {
	Vertex* vertex = *refVertex;

	/* Libera o espao do nome, se estiver alocado */
	if (vertex->name!=NULL) {
		free(vertex->name);
		vertex->name = NULL;
	}

	/* Desaloca a lista de vtices */
	destroyVertexLinkedList(&(vertex->parents));
	destroyVertexLinkedList(&(vertex->children));

	/* Libera Vertex */
	free(vertex);
	*refVertex = NULL;
}

int addVertex(Graph** refGraph, Vertex** refVertex) {
	Graph* graph = *refGraph;
	Vertex* vertex = *refVertex;

	/* Verifica se esta indexado, se estiver elimina o index */
	if (graph->index != NULL) destroyGraphIndex(&graph);

	/* Determina o id do vrtice de acordo com a ordem de insero */
	if (vertex->id==-1) vertex->id = graph->vertices->length;

	/* Acrecenta o vrtice no final da lista ligada */
	appendVertexNode(&(graph->vertices),vertex);

	/* Aponta o vrtice para o grafo */
	vertex->graph = graph;

	return vertex->id;
}

int sizeGraph(Graph* graph) {
	return length(graph->vertices);
}

int graphsize(Graph* graph) {
	return length(graph->vertices);
}

void createGraphIndex(Graph** refGraph) {
	Graph* graph = *refGraph;
	VertexNode* current;
	int i,j;
	int maxid;

	/* Verifica se o index esta NULL (por segurana) */
	if (graph->index != NULL) destroyGraphIndex(&graph);

	/* Ordena a lista pelos IDs em ordem crescente */
	sortVertexLinkedList(&graph->vertices,VERTEX_ID,ASCENDANT);

	/* Obtem o mximo Id */
	maxid = graph->vertices->last->vertex->id;

	/* Aloca o espao para o index */
	graph->index = malloc((maxid+1)*sizeof(Vertex*));
	graph->order = malloc((maxid+1)*sizeof(int));

	/* Percorre a lista ligada dos vrtices */

	// IMPORTANTE: Os vrtices devem estar necessriamente ordenados
	
	i = 0;
	current = graph->vertices->first;
	for (j=0;j<graph->vertices->first->vertex->id;j++) graph->index[j] = NULL;
	while(current != NULL) {
		graph->index[current->vertex->id] = current->vertex;
		graph->order[current->vertex->id] = i++;
		if (current->next!=NULL) for (j=current->vertex->id+1;j<current->next->vertex->id;j++) graph->index[j] = NULL;
		current = current->next;
	}
}

Vertex* getVertexMaxId(Graph** refGraph) {
	Graph* graph = *refGraph; 
	if (graph->index == NULL) createGraphIndex(&graph);
	return graph->vertices->last->vertex;
}

Vertex* getVertexMinId(Graph** refGraph) {
	Graph* graph = *refGraph; 
	if (graph->index == NULL) createGraphIndex(&graph);
	return graph->vertices->first->vertex;
}

void destroyGraphIndex(Graph** refGraph) {
	Graph* graph = *refGraph; 

	if (graph->index != NULL) {
		free(graph->order);
		graph->order = NULL;

		free(graph->index);
		graph->index = NULL;
	}

}

int getVertexOrder(Graph** refGraph, Vertex* vertex) {
	Graph* graph = *refGraph;

	/* Indexa para otimizar buscas futuras, mas no seria necessrio faz-lo */
	if (graph->index == NULL) createGraphIndex(&graph);

	return graph->order[vertex->id];

}

Vertex* getVertex(Graph** refGraph, int n) {
	Graph* graph = *refGraph;
	VertexNode* current;
	int count = 0;

	/* Verifica o tamanho */
	if (sizeGraph(graph) < n) return NULL;

	current = graph->vertices->first;
	while (current!=NULL) {
		if (count == n) return current->vertex;
		count++;
		current = current->next;
	}

	return NULL;
}

Vertex* getVertexById(Graph** refGraph, int id) {
	Graph* graph = *refGraph;

	/* Indexa para otimizar buscas futuras, mas no seria necessrio faz-lo */
	if (graph->index == NULL) createGraphIndex(&graph);

	return (id<=graph->vertices->last->vertex->id)?(Vertex*)graph->index[id]:NULL;
}

Vertex* getVertexById_NAO_INDEXADA(Graph** refGraph, int id) {
	Graph* graph = *refGraph;
	VertexNode* current;

	current = graph->vertices->first;
	while (current!=NULL) {
		if(current->vertex->id==id) return current->vertex;
		current = current->next;
	}

	return NULL;
}

void addArc(Graph** refGraph, Vertex** refParent, Vertex** refChild) {
	Graph* graph = *refGraph;
	Vertex* parent = *refParent;
	Vertex* child = *refChild;

	/* Verifica se o grafo  direcionado */
	if (graph->type != DIRECTED_GRAPH) {
		fprintf(stderr,GRAPH_ERRORS[ERR_LINK_IN_DIRECTED_GRAPH]);
		exit(EXIT_FAILURE);
	}

	/* Adicina o vrtice child na lista de filhos do vrtice parent */
	appendVertexNode(&(parent->children),child);

	/* Adicina o vrtice parent na lista de pais do vrtice child */
	appendVertexNode(&(child->parents),parent);

}

void addArcByIds(Graph** refGraph, int idParent, int idChild) {
	Graph* graph = *refGraph;
	Vertex* parent = getVertexById(&graph,idParent);
	Vertex* child = getVertexById(&graph,idChild);

	/* Adicina o vrtice child na lista de filhos do vrtice parent */
	appendVertexNode(&(parent->children),child);

	/* Adicina o vrtice parent na lista de pais do vrtice child */
	appendVertexNode(&(child->parents),parent);

}

void addLink(Graph** refGraph, Vertex** refVertexA, Vertex** refVertexB) {
	Graph* graph = *refGraph;
	Vertex* vertexA = *refVertexA;
	Vertex* vertexB = *refVertexB;

	/* Verifica se o grafo  direcionado */
	if (graph->type == DIRECTED_GRAPH) {
		fprintf(stderr,GRAPH_ERRORS[ERR_ARC_IN_NOT_DIRECTED_GRAPH]);
		exit(EXIT_FAILURE);
	}

	/* Arco A->B */

	/* Adicina o vrtice child na lista de filhos do vrtice parent */
	appendVertexNode(&(vertexA->children),vertexB);

	/* Adicina o vrtice parent na lista de pais do vrtice child */
	appendVertexNode(&(vertexB->parents),vertexA);

	/* Arco B->A */

	/* Adicina o vrtice child na lista de filhos do vrtice parent */
	appendVertexNode(&(vertexB->children),vertexA);

	/* Adicina o vrtice parent na lista de pais do vrtice child */
	appendVertexNode(&(vertexA->parents),vertexB);
}

void addLinkByIds(Graph** refGraph, int idVertexA, int idVertexB) {
	Graph* graph = *refGraph;
	Vertex* vertexA = getVertexById(&graph,idVertexA);
	Vertex* vertexB = getVertexById(&graph,idVertexB);

	/* Arco A->B */

	/* Adicina o vrtice child na lista de filhos do vrtice parent */
	appendVertexNode(&(vertexA->children),vertexB);

	/* Adicina o vrtice parent na lista de pais do vrtice child */
	appendVertexNode(&(vertexB->parents),vertexA);

	/* Arco B->A */

	/* Adicina o vrtice child na lista de filhos do vrtice parent */
	appendVertexNode(&(vertexB->children),vertexA);

	/* Adicina o vrtice parent na lista de pais do vrtice child */
	appendVertexNode(&(vertexA->parents),vertexB);
}

int getVertexOutDegree(Vertex* vertex) {
	return length((vertex->children));
}

int getVertexInDegree(Vertex* vertex) {
	return length((vertex->parents));
}

int getVertexDegree(Vertex* vertex) {
	return (isDirected((Graph*)vertex->graph)?(length(vertex->parents)+length(vertex->children)):length((vertex->parents)));
}

int isDirected(Graph *graph) {
	return graph->type;
}

Graph* loadGraph(char* filename) {

	char str[1023];
	int i,c;
	int nVertices,nLinks,type;
	char** vertices;
	int** links;
	Graph* graph;
	FILE* IN;

	IN = fopen (filename, "r");

	if (IN == NULL) {
		fprintf (stderr, "%s: %s (%d)\n", filename,strerror(errno),errno);
		exit (EXIT_FAILURE);
	}

	/* Primeira linha no formato: vertice */
	fscanf(IN,"%s%*[ \t]%d%*[^\n]\n",str,&nVertices);

	/* Inicializa o vetor de nomes dos ns */
	vertices = malloc(nVertices*sizeof(char*));

	/* Carrega os vrtices */
	for (i=0;i<nVertices;i++){
		vertices[i] = malloc(100*sizeof(char));
		fscanf(IN,"%d%*[ \t]%s%*[^\n]\n",&c,vertices[i]);
	}

	/* Links/Arcs no formato: <arcs|links> <NmeroDeLinks> */
	fscanf(IN,"%s%*[ \t]%d%*[^\n]\n",str,&nLinks);

	/* Define o tipo de grafo */
	if (strcmp(str,ARCS)==0) {
		type = DIRECTED_GRAPH; 
	} else {
		if (strcasecmp(str,LINKS)==0) {
			type = NOT_DIRECTED_GRAPH; 
		} else {
			fprintf(stderr,GRAPH_ERRORS[ERR_GRAPH_TYPE_UNKNOW],str);
			exit(EXIT_FAILURE);
		}
	}

	/* Inicializa o vetor com a definio dos arcos */
	links = malloc(nLinks*sizeof(int[2]));

	/* Carrega os Arcs/Links */
	for (i=0;i<nLinks;i++){
		links[i] = malloc(2*sizeof(int));
		fscanf(IN,"%d%*[ \t]%d%*[^\n]\n",&(links[i][0]),&(links[i][1]));
	}

	/* Fecha o arquivo */
	fclose(IN);

	/* Constri o grafo com base nas informaes do arquivo */	
	graph = buildGraph(type,vertices,nVertices,links,nLinks);

	/* Desaloca a memria */
	for (i=0;i<nVertices;i++) free(vertices[i]);
	free(vertices);

	for (i=0;i<nLinks;i++) free(links[i]);
	free(links);

	/* Retorna o grafo construido */
	return graph;

}

Graph* buildGraph(int type, char** vertices, int nVertices, int** links, int nLinks) {

	Graph* graph;
	Vertex* vertex;
	int i;

	/* Cria-se o grafo */
	graph = newGraph(type);

	/* Inicializa os vrtices */
	for (i=0;i<nVertices;i++) {
		/* Cria-se o vrtice */
		vertex = newVertex();
		vertex->id = i;
		/* Aloca exatamente o tamanho necessrio de memria e copia o nome */
		// vertex->name = malloc(strlen(vertices[i])*sizeof(char)+1);
		vertex->name = malloc(VERTEX_NAME_MAX_LENGTH*sizeof(char)+1);
		strcpy(vertex->name,vertices[i]);
		/* Adiciona-se o vrtice criado */
		addVertex(&graph,&vertex);
	}

	/* Indexa os vrtices */
	createGraphIndex(&graph);

	/* Estabelece os arcs (ou os links) */
	for (i=0;i<nLinks;i++) {
		if (type==DIRECTED_GRAPH) {
		/* Adiciona os ARCs em um grafo direcionado */
			addArcByIds(&graph,links[i][0],links[i][1]);
		} else {
		/* Adiciona os LINKs em um grafo no-direcionado */
			addLinkByIds(&graph,links[i][0],links[i][1]);
		}
	}

	/* Retorna o grafo construido */
	return graph;

}

void dumpGraph(Graph* graph) {
	int i,j;
	int nVertices;
	int nLinks;
	Vertex* vertex;
	VertexNode* current;
	VertexNode* child;

	/* Dump dos vrtices */

	/* Nmero de vertices */
	nVertices = sizeGraph(graph);
	printf("vertices %d\n",nVertices);	

	/* Recria o index */
	createGraphIndex(&graph);

	/* id e nome dos vrtices */
	nLinks = 0;
	current = graph->vertices->first;	
	while (current!=NULL) {
		vertex = current->vertex;
		printf("%d %s\n",vertex->id,vertex->name);	
		nLinks = nLinks + length(vertex->children);
		current = current->next;
	}

	/* Dump dos arcs (ou links) */

	/* Nmero de arcs (ou links) */
	if (isDirected(graph)) {
		printf("%s %d\n",ARCS,nLinks);	
	} else {
		nLinks = nLinks/2;
		printf("%s %d\n",LINKS,nLinks);	
	}

	/* Conexes */
	current = graph->vertices->first;	
	while (current!=NULL) {
		vertex = current->vertex;
		child = vertex->children->first;
		while (child!=NULL) {
			if (isDirected(graph) || child->vertex->id > vertex->id) { 
				printf("%d %d\n",vertex->id,child->vertex->id);	
			}
			child = child->next;
		}
		current = current->next;
	}

}

VertexLinkedList* getVertices(Graph* graph) {
	return graph->vertices;
}

VertexLinkedList* getVertexParents(Vertex* vertex) {
	return vertex->parents;
}


VertexLinkedList* getVertexChildren(Vertex* vertex) {
	return vertex->children;
}



/* LINKED_LIST: Funes que lidam com a lista de vrtices */

VertexLinkedList* newVertexLinkedList () {
	VertexLinkedList* list = malloc(sizeof(VertexLinkedList));
	list->length = 0;
	list->first = NULL;
	list->last = NULL;
	return list;
}

VertexLinkedList* newSortedVertexLinkedList (VertexLinkedList* base, int field, int direction) {

	VertexLinkedList* list = newVertexLinkedList();
	VertexNode* current = base->first;

	while (current != NULL) {
		appendVertexNodeSorted(&list,current->vertex,field,direction);	
		current = current->next;
	}

	return list;
}

void destroyVertexLinkedList (VertexLinkedList** refList) {
	VertexLinkedList* list = *refList;
	VertexNode* current;
	VertexNode* next;

	/* Percorre a lista liberando os VertexNodes */
	current = list->first;
	while (current != NULL) {
		next = current->next;
		destroyVertexNode(&current);
		current = next;
	}
	
	/* Libera lista */
	free(list);
	*refList = NULL;
}

VertexNode* newVertexNode () {
	VertexNode* node = malloc(sizeof(VertexNode));
	node->vertex = NULL;
	node->next = NULL;
	return node;
}

void destroyVertexNode(VertexNode** refNode) {
	VertexNode* node = *refNode;
	free(node);
	*refNode = NULL;
}

int length(VertexLinkedList* list) {
	return list->length;
}

void push (VertexLinkedList** refList, Vertex* vertex) {
	VertexLinkedList* list = *refList;
	VertexNode* node;

	/* Gera o novo node */
	node = newVertexNode();
	node->vertex = vertex;
	node->next = NULL;

	/* Verifica se a lista foi inicializada, ou seja, primeiro elemento  null */
	if (list->first == NULL) {
		/* Inicializa o tamanho */
		list->length = 1;
		/* Inicializa o primeiro e o ltimo elemento com o node recm criado */
		list->first = node;
		list->last = node;
	} else {
		/* Incrementa o tamanho */
		list->length = list->length + 1;
		node->next = list->first;
		list->first = node;
	}
}


void appendVertexNode(VertexLinkedList** refList, Vertex* vertex) {
	VertexLinkedList* list = *refList;
	VertexNode* node;

	/* Gera o novo node */
	node = newVertexNode();
	node->vertex = vertex;
	node->next = NULL;

	/* Verifica se a lista foi inicializada, ou seja, primeiro elemento  null */
	if (list->first == NULL) {
		/* Inicializa o tamanho */
		list->length = 1;
		/* Inicializa o primeiro e o ltimo elemento com o node recm criado */
		list->first = node;
		list->last = node;
	} else {
		/* Incrementa o tamanho */
		list->length = list->length + 1;
		list->last->next = node;
		list->last = node;
	}
}

void appendVertexNodeSorted(VertexLinkedList** refList, Vertex* vertex, int field, int direction) {

	VertexLinkedList* list = *refList;
	VertexNode* A = NULL; // N "anterior"
	VertexNode* B = NULL; // N a ser inserido
	VertexNode* C = NULL; // N "posterior"
	int c,b;
	int done = 0;

	/* Verificao da entrada e dados */

	if (direction!=ASCENDANT && direction!=DESCENDANT) {
		fprintf(stderr,GRAPH_ERRORS[ERR_UNKNOW_SORT_DIRECTION]);
		exit(EXIT_FAILURE);
	}


	/* Gera o novo node */
	B = newVertexNode();
	B->vertex = vertex;
	B->next = NULL;

	if (list->first == NULL){
		push(&list,vertex);
		done = 1;
	} else {
		C = list->first;
	}

	while (!done && C != NULL) {

			switch (field) {

				case VERTEX_ID:
					b = B->vertex->id; 
					c = C->vertex->id; 
				break;

				case LINK_DEGREE:
					b = getVertexDegree(B->vertex); 
					c = getVertexDegree(C->vertex); 
				break;
	
				case IN_DEGREE:
					b = getVertexInDegree(B->vertex); 
					c = getVertexInDegree(C->vertex); 
				break;

				case OUT_DEGREE:
					b = getVertexOutDegree(B->vertex); 
					c = getVertexOutDegree(C->vertex); 
				break;

				default:
					fprintf(stderr,GRAPH_ERRORS[ERR_UNKNOW_SORT_FIELD]);
					exit(EXIT_FAILURE);
			}


			/* Insere o vrtice se necessrio */
			if ((direction==ASCENDANT && b<c) || (direction==DESCENDANT && b>c)) {
				if (A == NULL) {
					list->first = B;
				} else {
					A->next = B;
				}
				B->next = C;
				done = 1;
			} else {
				A = (A!=NULL?A->next:list->first);
				C = A->next;
			}
	}

	if (!done) appendVertexNode(&list,vertex);

}


Vertex* removeFirst(VertexLinkedList** refList) {
	VertexLinkedList* list = *refList;
	VertexNode* node;
	Vertex* vertex;

	/* Verifica se no  uma lista vazia */
	if (list->first == NULL) return (int) NULL;

	/* Decrementa o tamanho */
	list->length = list->length - 1;

	/* Atribui o contedo do primeiro node para uma varivel auxiliar */
	node = list->first;
	vertex = node->vertex;

	if (node->next == NULL) {
		list->first = NULL;
		list->last = NULL;
	} else {
		list->first = node->next;	
	}

	/* Libera o VertexNode */
	free(node);

	/* Retorna o contedo que estava armazenado no VertexNode */
	return vertex;

}

Vertex* removeLast(VertexLinkedList** refList) {
	VertexLinkedList* list = *refList;
	VertexNode* node;
	VertexNode* current;
	Vertex* vertex;

	/* Verifica se no  uma lista vazia */
	if (list->first == NULL) return (int) NULL;

	/* Decrementa o tamanho */
	list->length = list->length - 1;

	/* Atribui o contedo do ltimo node para uma varivel auxiliar */
	node = list->last;
	vertex = node->vertex;

	/* Percorre a lista para saber quem aponta para o 'last' */
	current = list->first;
	while (current->next != node) {
		current = current->next;	
	}
	current->next = NULL;
	list->last = current;

	/* Libera o VertexNode */
	free(node);

	/* Retorna o contedo que estava armazenado no VertexNode */
	return vertex;

}

void sortVertexLinkedList(VertexLinkedList** refList, int field, int direction) {

	VertexLinkedList* list = *refList;
	VertexNode* A;
	VertexNode* B;
	VertexNode* P;
	int a,b;
	int last = length(list)-1;
	int cursor;

	/* Verificao da entrada e dados */

	if (direction!=ASCENDANT && direction!=DESCENDANT) {
		fprintf(stderr,GRAPH_ERRORS[ERR_UNKNOW_SORT_DIRECTION]);
		exit(EXIT_FAILURE);
	}

	/* BubbleSort */
	while (last > 0) { 
	
		P = NULL;
		cursor = 0;

		while (cursor != last) {

			A = (P==NULL?list->first:P->next);
			B = A->next;

			switch (field) {
				case VERTEX_ID:
					a = A->vertex->id; 
					b = B->vertex->id; 
				break;

				case LINK_DEGREE:
					a = getVertexDegree(A->vertex); 
					b = getVertexDegree(B->vertex); 
				break;
	
				case IN_DEGREE:
					a = getVertexInDegree(A->vertex); 
					b = getVertexInDegree(B->vertex); 
				break;

				case OUT_DEGREE:
					a = getVertexOutDegree(A->vertex); 
					b = getVertexOutDegree(B->vertex); 
				break;

				default:
					fprintf(stderr,GRAPH_ERRORS[ERR_UNKNOW_SORT_FIELD]);
					exit(EXIT_FAILURE);
			}

			/* Faz o swap dos vrtice se necessrios */
			if ((direction==ASCENDANT && a>b) || (direction==DESCENDANT && a<b)) {
				if (P!=NULL) P->next = B;
				if (list->first == A) list->first = B;
				if (list->last == B) list->last = A;
				A->next = B->next;
				B->next = A;
			}
			
			P = (P==NULL?list->first:P->next);

			cursor++;
		}

		last--;
	}
}

void dumpVertexLinkedList(VertexLinkedList* list) {
	VertexNode* current = list->first;
	int i = 0;

	while (current!=NULL) {
		printf("#%3d vertex id: %3d name: %10s LinkDegree: %3d InDegree: %3d OutDegree: %3d\n",(i++),current->vertex->id,current->vertex->name,getVertexDegree(current->vertex),getVertexInDegree(current->vertex),getVertexOutDegree(current->vertex));
		current = current->next;
	}

}
