/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*
 * @file	gibbs.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Maro/2007
 *
 * Para documentao ver gibbs.h
 *
 */

/* Cabealhos das funes */
#include "gibbs.h"

/* Defines */

/* Funes */

int* defineEliminationOrder(Graph* moral) {

	Graph* dstree;
	BELTree* beltree;
	int* q = NULL;

	/* Encontra o n quase-perifrico e monta a BEL */
	beltree = doGibbsHeuristic(moral);

	/* Monta a rvore de disseco */
	dstree = buildDissectionTree(moral,beltree);

	/* Monta a ps-ordem e cria a ordem de eliminao */
	q = buildEliminationOrder(dstree,getVertexById(&dstree,0));

	// Desaloca a memria
	destroyDissectionTree(&dstree);
	destroyBELTree(&beltree);

	return q;

}

BELTree* doGibbsHeuristic(Graph* moral) {

	Vertex* vertex;

	VertexNode* current;
	VertexLinkedList* toTest;
	int countToTest;

	BELTree* currentBEL = NULL;
	BELTree* newBEL = NULL;

	double currentBalance;
	double newBalance;

	double currentVar;
	double newVar;

	int currentRootDegree;
	int newRootDegree;

	int min = sizeGraph(moral);
	int degree;

	int tested[getVertexMaxId(&moral)->id]; // Vetor para memria dos vrtices j testados
	int degrees[getVertexMaxId(&moral)->id]; // Cache para o grau dos ns

	int i,j;

	int done = 0; // Flag de controle para interromper as iteraes da heurstica
	int swap = 0; // Flag que indica se a BEL corrente deve ou no ser substituida
	int swaped = 0; // Flag que indica se a BEL corrente foi substituida

	// Identifica o n de menor grau
	i = 0;
	current = moral->vertices->first;
	while (current!=NULL) {

		// Escolhe o menor grau
		degree = getVertexDegree(current->vertex);
		if (min > degree) {
			min = degree;
			vertex = current->vertex;
		}

		// Memria dos vrtices j testados
		tested[i] = 0;

		// Cache do grau dos ns
		degrees[i] = degree;

		i = i + 1;
		current = current->next;	
	}

	// Obtem a primeira BEL
	currentBEL = buildBELTree (moral,vertex); 
	tested[vertex->id] = 1;
	swaped = 1;

	currentBalance = BELBalance(currentBEL);
	currentRootDegree = degrees[vertex->id];
	currentVar = BELVariance(currentBEL);

	while (!done) {

		if (swaped) {
			// Cria a lista de vrtices que devem ser verificados a partir dos vrtices no ltimo nvel da BEL corrente
			toTest = newVertexLinkedList();

			// Popula a lista de vrtices a serem verificados
			current = currentBEL->vertices[currentBEL->depth]->first;
			while (current!=NULL) {
				if (!tested[current->vertex->id]) appendVertexNode(&toTest,getVertexById(&moral,current->vertex->id));
				current = current->next;
			}
			// Atualiza o contador
			countToTest = toTest->length;

			// Atualiza a flag, indicando que que j foi considerada a informao da troca
			swaped = 0;
		}

		#ifdef DUMPGIBBS
			fprintf(DBGOUT,"\nDEBUG: Current BEL\n");
			dumpBELTree(currentBEL);
			fprintf(DBGOUT,"\nDEBUG: toTest length: %d\n",length(toTest));
			dumpVertexLinkedList(toTest);
		#endif

		// Se no houver ns a serem verificados encerra a procura
		if (countToTest==0) {
			done = 1;
		} else {
			// Percorre a lista dos vrtices a serem testados
			current = toTest->first;
			while (current!=NULL) {

				// Decrementa o contador
				countToTest--;

				// Obtem a nova BEL
				newBEL = buildBELTree (moral,current->vertex); 

				#ifdef DUMPGIBBS
					fprintf(DBGOUT,"\nDEBUG: New BEL\n");
					dumpBELTree(newBEL);
				#endif

				// Verifica se a nova BEL  superior a corrente
				newBalance = BELBalance(newBEL);
				newRootDegree = degrees[current->vertex->id];
				newVar = BELVariance(newBEL);

				// 1. Compara a profundidade das BELs
				if (newBEL->depth==currentBEL->depth) {
					// 2. Compara a largura das BELs
					if (newBEL->width==currentBEL->width) {
						// 3. Compara o balanceamento das BELs
						if (newBalance==currentBalance) {
							// 4. Compara o grau do n root
							if (newRootDegree==currentRootDegree) {
								// 5. Compara a varincia
								if (newVar < currentVar) {
									swap = 1;
								} else {
									swap = 0;
								}
							} else if (newRootDegree < currentRootDegree) {
								swap = 1;
							} else {
								swap = 0;
							}
						} else if (newBalance < currentBalance) {
							swap = 1;
						} else {
							swap = 0;
						}
					} else if (newBEL->width < currentBEL->width) {
						swap = 1;
					} else {
						swap = 0;
					}
				} else if (newBEL->depth > currentBEL->depth) {
					swap = 1;
				} else {
					swap = 0;
				}

				if (swap==1) {
					// Libera a BEL corrente
					destroyBELTree(&currentBEL);

					// A nova BEL passa aser a corrente
					currentBEL = newBEL;
					vertex = current->vertex;
					currentBalance = newBalance;
					currentRootDegree = newRootDegree;
					currentVar = newVar;

					// Indica que houve a troca
					swaped = 1;

				} else {
					destroyBELTree(&newBEL);
				}
				newBEL = NULL;

				// Atualiza o cache com o vrice verificado
				tested[current->vertex->id] = 1;
				current = current->next;

			}
		}

		// Libera o espao da lista de vrtices testada
		if (toTest) destroyVertexLinkedList(&toTest);
	}

	#ifdef DUMPGIBBS
		fprintf(DBGOUT,"\nDEBUG: Selected BEL\n");
		dumpBELTree(currentBEL);
	#endif
	return currentBEL;

}

double BELBalance (BELTree* beltree) {

	double q;
	double qsep;

	int* sum = beltree->sum;
	int sd,se,sep;

	int begin = 0;
	int end = beltree->depth;
	int range = end - begin;
	
	int i,j;

	qsep = 0.0;

	if ((range%2)==1) {
	/* Se for nmero impar possveis valores: (range+1)/2 e (range-1)/2 */
		for (j=-1;j<2;j+=2) {
			sep = (range+j)/2 + begin;
			sd = (begin>0?(sum[sep-1]-sum[begin-1]):sum[sep-1]);
			se = sum[end] - sum[sep];
			q = (se>sd?(float)((1.0*se)/sd):(float)((1.0*sd)/se));

			/* Valores de 'q' mais prximos de 1 de significa melhor distribuio */
			if (qsep==0 || (q-1)<(qsep-1)) {
				qsep = q;
			}
		}
	} else {
		/* Se for par tenta: i) central ii) central+1 iii) central-1 */
		for (j=-1;j<2;j+=1) {
			sep = range/2 + begin + j;
			sd = (begin>0?(sum[sep-1]-sum[begin-1]):sum[sep-1]);
			se = sum[end] - sum[sep];
			q = (se>sd?(float)((1.0*se)/sd):(float)((1.0*sd)/se));

			/* Valores de 'q' mais prximos de 1 de significa melhor distribuio */
			if (qsep==0 || (q-1)<(qsep-1)) {
				qsep = q;
			}
		}
	}

	return qsep;
}

double BELVariance (BELTree* beltree) {

	int sum = 0;
	int sum2 = 0;
	int i,n;

	for (i=0;i<=beltree->depth;i++) {
		n = beltree->vertices[i]->length;
		sum = sum + n;	
		sum2 = sum2 + n*n;	
	}

	return (sum2 - sum)/(beltree->depth + 1);
}


// TODO: fazer o realloc para do nmero de nveis necessrios ser superior ao alocado

BELTree* buildBELTree (Graph* graph, Vertex* root) {

	/* Primeira estimativa do tamanho da BEL no mximo = #ns */
	int depth = sizeGraph(graph); 

	/* rvore que ser construida */
	BELTree* tree;

	/* Variveis auxiliares */
	Vertex* belvertex;

	VertexNode* current;
	VertexNode* child;
	VertexLinkedList* children;

	int i;
	int counter;

	/* Variveis utilizadas na ordenao */
	VertexNode* A;
	VertexNode* B;
	VertexNode* P;
	int a,b;
	int last;
	int cursor;

	/* Aloca o espao para a BEL e inicializa os argumentos */
	tree = malloc(sizeof(BELTree)); 
	tree->width = 0;
	tree->depth = 0;
	tree->sum = NULL;
	tree->graph = newGraph(DIRECTED_GRAPH);
	tree->vertices = malloc(depth*sizeof(VertexLinkedList*));

	/* Inicializa os nveis com NULL */
	for (i=0;i<depth;i++) tree->vertices[i] = NULL;

	/* Insere os vrtices do grafo original na BELTree - visando garantir a mesma ordem do grafo original */
	current = graph->vertices->first;
	while(current != NULL) {
		/* Cria o vrtice e acrescenta  BEL */
		belvertex = buildVertex(current->vertex->id,current->vertex->name);
		belvertex->visited = 0;
		addVertex(&tree->graph,&belvertex);
		current = current->next;
	}

	/* Inicia o contador de vrtices */
	counter = 0;

	/* Primeiro elemento tem apenas 1 vrtice */

	/* Aloca espao para a lista de vrtices */
	tree->vertices[tree->depth] = newVertexLinkedList();

	/* Acrescenta o vrtice  lista */
	belvertex = getVertexById(&tree->graph,root->id);
	appendVertexNode(&(tree->vertices[tree->depth]),belvertex);

	/* Incrementa o contador de vrtices */
	counter++;

	/* Marca o vrtice como visitado */
	belvertex->visited = 1;

	/* Largura da rvore */
	tree->width = length(tree->vertices[tree->depth]);

	/* Looping principal. Condio de parada: counter = siza(graph) */
	while (counter < sizeGraph(graph)) {

		/* Incrementa o nvel corrente da BEL */
		tree->depth++;

		/* Aloca o espao */
		tree->vertices[tree->depth] = newVertexLinkedList();

		/* Percorre a lista de vrtice do nvel (i-1) para pegar os adjacentes */
		current = tree->vertices[tree->depth-1]->first; 

		while (current != NULL) {

			/* Obtem a lista dos ns adjacentes no grafo original */
			children = getVertexChildren(getVertexById(&graph,current->vertex->id));
			if (length(children)==0) continue;

			/* Ordena os vrtices */
			// sortVertexLinkedList(&children,LINK_DEGREE,ASCENDANT); 

			/* Acrescenta os ns no visitados ao nvel corrente da BEL */
			child = children->first;
			while (child != NULL) {

				/* Obtem o vrtice correspondente na BEL */
				belvertex = getVertexById(&tree->graph,child->vertex->id);

				/* Verifica se foi visitado */
				if (belvertex->visited != 1) {

					/* Acrescenta ao i-simo nvel da BEL */

					/* Acrescenta o vrtice da BEL  lista */
					appendVertexNode(&(tree->vertices[tree->depth]),belvertex);
				
					/* Acrescenta o vrtice do GRAFO original  lista  -> ERRADO !!! */
					// appendVertexNode(&(tree->vertices[tree->depth]),getVertexById(&graph,belvertex->id));

					/* Incrementa o contador de vrtices */
					counter++;

					/* Marca o vrtice como visitado */
					belvertex->visited = 1;

					/* Adiciona o arco */
					addArc(&(tree->graph),&current->vertex,&belvertex);
				}
				child = child->next;
			}
			current = current->next;
		}

		/* Ordena a lista de ns do i-simo */
		// sortVertexLinkedList(&(tree->vertices[tree->depth]),LINK_DEGREE,ASCENDANT); 
		
		/* Ordena os ns da BEL de forma acendente pelo grau do n no Moral Graph (BubbleSort) */
		last = length(tree->vertices[tree->depth])-1;
		while (last > 0) { 
			P = NULL;
			cursor = 0;
			while (cursor != last) {
				A = (P==NULL?tree->vertices[tree->depth]->first:P->next);
				B = A->next;

				/* Grau dos vrtices no Moral Graph */
				a = getVertexDegree(getVertexById(&graph,A->vertex->id)); 
				b = getVertexDegree(getVertexById(&graph,B->vertex->id)); 

				/* Faz o swap dos vrtice se necessrios */
				if (a>b) {
					if (P!=NULL) P->next = B;
					if (tree->vertices[tree->depth]->first == A) tree->vertices[tree->depth]->first = B;
					if (tree->vertices[tree->depth]->last == B) tree->vertices[tree->depth]->last = A;
					A->next = B->next;
					B->next = A;
				}
				P = (P==NULL?tree->vertices[tree->depth]->first:P->next);
				cursor++;
			}
			last--;
		}	
				
		/* Verifica a largura */
		tree->width = (length(tree->vertices[tree->depth])>tree->width?length(tree->vertices[tree->depth]):tree->width);
	}

	/* Volta para o estado inicial do vrtice */
	current = tree->graph->vertices->first;
	while(current != NULL) {
		current->vertex->visited = 0;
		current = current->next;
	}

	/* Inicializa o vetor com a soma acumulada do nmero de vrtices em cada um dos nveis */
	tree->sum = malloc((tree->depth+1)*sizeof(int));
	tree->sum[0] = length(tree->vertices[0]); 
	for (i=1;i<=tree->depth;i++) tree->sum[i] = tree->sum[i-1] + length(tree->vertices[i]); 

	return tree;
}

void destroyBELTree (BELTree** refBELTree) {
	BELTree* tree = *refBELTree;
	int i;

	/* Destri as listas de vrtices */
	for (i=0;i<=tree->depth;i++) {
		destroyVertexLinkedList(&(tree->vertices[i]));
	}
	/* Desaloca a memria do vetor com a lista de vrtices */
	free(tree->vertices);
	tree->vertices = NULL;

	/* Desaloca a memria do vetor com a soma acumulada */
	if (tree->sum!=NULL) {
		free(tree->sum);
		tree->sum = NULL;
	}

	/* Destri o grafo */
	destroyGraph(&tree->graph);

	/* Desaloca a memria da rvore */
	free(tree);
	refBELTree = NULL;
}

void dumpBELTree (BELTree* tree) {
	int i;
	VertexNode* current;
	VertexNode* child;
	Vertex* root = tree->vertices[0]->first->vertex;
	
	fprintf(DBGOUT,"BELTree root: %s degree: %d\n",root->name,getVertexDegree(root));
	fprintf(DBGOUT,"depth: %d width: %d balance: %f var: %f \n",tree->depth,tree->width,BELBalance(tree),BELVariance(tree));

	for (i=0;i<=tree->depth;i++) {
		fprintf(DBGOUT,"%3d (%3d):\n",i,tree->sum[i]);
		current = tree->vertices[i]->first;
		while (current != NULL) {
			fprintf(DBGOUT,"\t%s",current->vertex->name);
			if (length(current->vertex->children)>0) {
				fprintf(DBGOUT,"-( ");
				child = current->vertex->children->first;
				while (child != NULL) {
					fprintf(DBGOUT,"%s ",child->vertex->name);
					child = child->next;
				}
				fprintf(DBGOUT,")");
			} 
			fprintf(DBGOUT,"\n");
			current = current->next;
		}
	}
}

// Estratgica Bsica
int nextSeparator_BASICO(BELTree* beltree, int begin, int end) {
	int range = end - begin;
	return ((range%2)==1?(range+1)/2:range/2) + begin;

}

// Estratgia com ponderao do nmero de vrtices
int nextSeparator_SEPARADOR_PONDERADO(BELTree* beltree, int begin, int end) {

	int range = end - begin;
	int sd,se; 			/* Soma  direita, soma  esquerda */
	float qroot,q; 		/* Razo entre as somas SomaMaior/SomaMenor */
	int i,j,sep,root;
	int* sum = beltree->sum;

	/* incio: Estratgia considerando o vetor de soma dos ns */
	qroot = 0;
	
	if ((range%2)==1) {
	/* Se for nmero impar possveis valores: (range+1)/2 e (range-1)/2 */
		for (j=-1;j<2;j+=2) {
			sep = (range+j)/2 + begin;
			sd = (begin>0?(sum[sep-1]-sum[begin-1]):sum[sep-1]);
			se = sum[end] - sum[sep];
			q = (se>sd?(float)((1.0*se)/sd):(float)((1.0*sd)/se));

			/* Valores de 'q' mais prximos de significa melhor distribuio */
			if (qroot==0 || (q-1)<(qroot-1)) {
				root = sep;
				qroot = q;
			}
		}
	} else {
		if (range > 2) {
			/* Se for par tenta: i) central ii) central+1 iii) central-1 */
			for (j=-1;j<2;j+=1) {
				sep = range/2 + begin + j;
				sd = (begin>0?(sum[sep-1]-sum[begin-1]):sum[sep-1]);
				se = sum[end] - sum[sep];
				q = (se>sd?(float)((1.0*se)/sd):(float)((1.0*sd)/se));

				/* Valores de 'q' mais prximos de significa melhor distribuio */
				if (qroot==0 || (q-1)<qroot) {
					root = sep;
					qroot = q;
				}
			}
		} else {
			root = range/2 + begin;
		}
	}

	return root;

}

// Estratgia que procura o menor separador
int nextSeparator_MENOR_SEPARADOR(BELTree* beltree, int begin, int end) {
	int range = end - begin;
	int i,j,sep,root,rootLength;

	if ((range%2)==1) {
	/* Se for nmero impar possveis valores: (range+1)/2 e (range-1)/2 */
		root = (beltree->vertices[(range-1)/2+begin]->length > beltree->vertices[(range+1)/2+begin]->length)?((range+1)/2+begin):((range-1)/2+begin);
	} else {
		rootLength=0;
		if (range > 2) {
			/* Se for par tenta: i) central ii) central+1 iii) central-1 */
			for (j=-1;j<2;j+=1) {
				sep = range/2 + begin + j;
				if (rootLength==0 || beltree->vertices[sep]->length < rootLength) {
					root = sep;
					rootLength = beltree->vertices[root]->length;
				}
			}
		} else {
			root = range/2 + begin;
		}
	}

	return root;

}

void dissection(Graph** refDSTree, Vertex** refRoot, BELTree* beltree, int begin, int end) {
	Graph* dstree = *refDSTree;
	Vertex* root = (refRoot!=NULL?*refRoot:NULL);
	Vertex* vertex;
	int sep;

	if (begin>end) return;

	/* Define o prximo separador */
	if ((end - begin)<2) {
		if (refRoot==NULL) root = getVertexById(&dstree,begin);
		sep = ((root->id>end)?end:begin); 
	} else {
		sep = nextSeparator(beltree,begin,end);
	}

	/* Determina o vrtice que ser raiz ou folha */
	vertex = getVertexById(&dstree,sep);

	/* Vincula o novo vertex como filho da raiz */
	if (refRoot != NULL) addArc(&dstree,&root,&vertex);

	/* Para a esquerda */
	dissection(&dstree,&vertex,beltree,begin,sep-1);

	/* Para a direita */
	dissection(&dstree,&vertex,beltree,sep+1,end);

}

Graph* buildDissectionTree(Graph* moral, BELTree* beltree){
	Graph* dstree;
	Vertex* vertex;
	VertexNode* current;
	int i;

	/* Inicializa do grafo */
	dstree = newGraph(DIRECTED_GRAPH);

	/* Cria-se a lista de vrtices da rvore de disseco a partir dos nveis da BEL */
	for (i=0;i<=beltree->depth;i++) {

		/* Cria um novo vrtice */
		vertex = newVertex();
		vertex->id = i;
		vertex->name = malloc(VERTEX_NAME_MAX_LENGTH*sizeof(char));
		sprintf(vertex->name,"%d",i);

		/* Adiciona os vrtices do Moral Graph ao vrtice da rvore de disseco com base na BEL */
		vertex->list = newVertexLinkedList();
		current = beltree->vertices[i]->first;
		while (current!=NULL) {
			appendVertexNode(&(vertex->list),getVertexById(&moral,current->vertex->id));
			current = current->next;
		}

		/* Adiciona o novo vrtice ao grafo */
		addVertex(&dstree,&vertex);

	}

	/* Determina os separadores */
	dissection(&dstree,NULL,beltree,0,beltree->depth);


	#ifdef DUMPGIBBS
		fprintf(DBGOUT,"\nDEBUG DissectionTree:\n");
		dumpDissectionTree(dstree);
	#endif

	return dstree;
}

void destroyDissectionTree (Graph** refDSTree) {
	Graph* dstree = *refDSTree;
	VertexNode* current;

	/* Desaloca a lista de vrtices */
	current = dstree->vertices->first;
	while (current != NULL) {
		destroyVertexLinkedList(&(current->vertex->list));
		current->vertex->list = NULL;
		current = current->next;
	}

	/* Destri o grafo */
	destroyGraph(&dstree);

}

void dumpDissectionTree (Graph* dstree) {
	int i;
	VertexNode* current;
	VertexNode* child;
	VertexNode* node;

	current = dstree->vertices->first;

	while (current != NULL) {
		fprintf(DBGOUT,"node %s (%d/%d): ",current->vertex->name,current->vertex->id,length(dstree->vertices));

		/* Lista os vertices da BEL associados ao vertex da rvore de disseo */
		node = current->vertex->list->first;
		fprintf(DBGOUT,"BEL [ ");
		while (node != NULL) {
			fprintf(DBGOUT,"%s ",node->vertex->name);
			node = node->next;
		}
		fprintf(DBGOUT,"]\n");
		
		/* Lista os vertices "children" */
		if (length(current->vertex->children)>0) {
			fprintf(DBGOUT,"|_children( ");
			child = current->vertex->children->first;
			while (child != NULL) {
				fprintf(DBGOUT,"%s ",child->vertex->name);
				child = child->next;
			}
			fprintf(DBGOUT,")\n");
		} 
		fprintf(DBGOUT,"\n");
		current = current->next;
	}
}

int* buildEliminationOrder(Graph* dstree, Vertex* vertex) {
	VertexNode* current;
	int* elmorder;
	int n = 0;	 // Contador da posio do vetor
	int sum  = 0;

	current = dstree->vertices->first;
	while (current != NULL) {
		current->vertex->order = 0; // Por segurana, inicializa o campo de ordenao dos ns com zero
		sum = sum + length(current->vertex->list);
		current = current->next;
	}

	/* Aloca a memria para a lista */
	elmorder = malloc(sum*sizeof(int));

	/* Monta a ps-ordem recursivamente percorrendo os vrtices do grafo */
	searchEliminationOrder(elmorder,&n,&vertex);

	return elmorder;
}

void searchEliminationOrder(int* elmorder, int *n, Vertex** refVertex) {
	Vertex* vertex = *refVertex;
	VertexLinkedList* parents = getVertexParents(vertex);
	VertexLinkedList* children = getVertexChildren(vertex);
	VertexNode* current;

	/* Marca o n como visitado */
	vertex->order = 1;

	/* Percorre recursivamente os filhos do n acrescenta  ps-ordem e marca como visitado */
	if (length(children) > 0) {
		sortVertexLinkedList(&children,LINK_DEGREE,ASCENDANT); 
		current = children->first;	
		while (current != NULL) {
			if (current->vertex->order == 0) {
				searchEliminationOrder(elmorder,n,&(current->vertex));
			}
			current = current->next;
		}
	} 

	/* Marca como visitado acrescenta os vrtices da lista de vrtices  ps ordem */	
	current = vertex->list->first;
	while (current != NULL) {
		elmorder[*n] = current->vertex->id;
		*n = *n + 1;
		current = current->next;
	}

	/* Sobe um nvel acima */
	if (length(parents) > 0 && parents->first->vertex->order == 0) {
		searchEliminationOrder(elmorder,n,&(parents->first->vertex));
	}
}

