/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Abril/2007
 *
 * Para documentao ver potencial.h
 */


/* Lista o cabealho das funes e estrutura de dados */

#include "potential.h"

/* Variveis Globais */

char* VARIABLE_TYPES[] = {"DISCRETE","CONTINUOUS"}; // Tipos de variveis possveis. A verso atual suporta apenas variveis discretas

/* Funes */

Variable* buildVariable(int id, char* name, int type, int nstates, char** states) {
	Variable* var = malloc(sizeof(Variable));	
	int i;

	// Id
	var->id = id;

	// Nome da varivel
	var->name = malloc(VARIABLE_NAME_LENGTH*sizeof(char));
	strcpy(var->name,name);

	// Tipo
	var->type = type;

	// Inicializao dos estados
	var->nstates = nstates;
	var->states = malloc(nstates*sizeof(char*));
	for (i=0;i<nstates;i++) {
		var->states[i] = malloc(VARIABLE_STATE_NAME_LENGTH*sizeof(char*));
		strcpy(var->states[i],states[i]);
	}

	return var;
}

void dumpVariable(Variable* var) {
	int i;
	fprintf(DBGOUT,"Variable %s {\n",var->name);
	fprintf(DBGOUT,"\ttype %s [ %d ] { ",VARIABLE_TYPES[var->type],var->nstates);
	for (i=0;i<var->nstates-1;i++) fprintf(DBGOUT,"%s, ",var->states[i]);
	fprintf(DBGOUT,"%s };\n",var->states[var->nstates-1]);
	fprintf(DBGOUT,"}\n");
}

void destroyVariable(Variable** refVariable) {
	Variable* var = *refVariable;
	int i;

	// Desaloca o 'name'
	free(var->name);

	// Desaloca o vetor de estados
	for (i=0;i<var->nstates;i++) free(var->states[i]);
	free(var->states);

	// Desaloca a varivel
	free(var);
	*refVariable = NULL;
}

Potential* buildPotential (int id, int nvars, Variable** vars, double* values) {
	Potential* pot = malloc(sizeof(Potential));
	int i;

	// Id
	pot->id = id;

	// Nmero de variveis
	pot->nvars = nvars;

	// Inicializa os arrays de variveis
	pot->vars = malloc(nvars*sizeof(Variable*));
	pot->nvalues = 1; // Total de valores
	for (i=0;i<nvars;i++) {
		pot->vars[i] = vars[i];
		pot->nvalues = pot->nvalues*vars[i]->nstates;
	}	

	// Faz uma cpia dos valores
	pot->values = malloc(pot->nvalues*sizeof(double));
	// for (i=0;i<pot->nvalues;i++) pot->values[i] = values[i];
	memcpy(pot->values,values,pot->nvalues*sizeof(double));

	return pot;
}

void dumpPotential(Potential* pot) {
	int mult;
	int from;
	int i,j;

	// Caso especial
	if (pot->nvars==0) {
		fprintf(DBGOUT,"potential () { %10.8f }\n",pot->values[0]);
		return;
	}

	// Caso geral
	mult = pot->nvalues/pot->vars[pot->nvars-1]->nstates;
	fprintf(DBGOUT,"potential (");
	for (i=0;i<pot->nvars;i++) fprintf(DBGOUT," %s%s",pot->vars[i]->name,(i<pot->nvars-1?",":" ) {\n"));

	for (i=0;i<pot->vars[pot->nvars-1]->nstates;i++) {
		fprintf(DBGOUT,"\t");
		from = i*mult;
		for (j=0;j<mult;j++) fprintf(DBGOUT,"%10.8f%s",pot->values[from+j],(j<mult-1?",":"\n"));
	}

	fprintf(DBGOUT,"}\n");

}

void destroyPotential(Potential** refPotential) {
	Potential* pot = *refPotential;

	// Desaloca o vetor de variveis mas NO destri ou desaloca as variveis
	free(pot->vars);

	// Desloca a memria dos valores
	free(pot->values);

	// Libera a memria da estrutura
	free(pot);
	*refPotential = NULL;
}


Potential* multPotentials (Potential* pot_1, Potential* pot_2) {

	Potential* pot_a;
	Potential* pot_b;
	Potential* result;

	int *index_a = NULL;	// Array de index permutado
	int *index_b = NULL;	// Array de index permutado

	// Multiplicadores para calcular os estados de uma observao dado o index
	int *src_mult_a = NULL;
	int *dst_mult_a = NULL;
	int *src_mult_b = NULL;
	int *dst_mult_b = NULL;
	int value_index_a;
	int value_index_b;

	// Variveis auxiliares para converso de index do nvalues;
	int state;
	int rest;

	// Variveis auxiliares para permutar os indices
	int swap_a = -1;
	int swap_b = -1;
	int swap;

	int nequals;	// Nmero de variveis iguais nos dois potenciais
	int mult;		// Tamanho do bloco formado pela variveis iguais
	int nvalues;	// Nmero de valores no potencial resultante
	int nvars;		// Nmero de variveis no potencial resultante

	Variable** vars; // Variveis do potencial resultante

	int i,j,k,l,m;

	int fromto_b[18];

	// Define qual o potencial com menor nmero de variveis
	if ((pot_1)->nvars < (pot_2)->nvars) {
		pot_a = pot_1;
		pot_b = pot_2;
	} else {
		pot_a = pot_2;
		pot_b = pot_1;
	}

	// Inicializa o array de index permutados
	index_a = malloc(pot_a->nvars*sizeof(int));
	index_b = malloc(pot_b->nvars*sizeof(int));
	for (i=0;i<pot_b->nvars;i++) {
		if (i<pot_a->nvars) index_a[i] = i;
		index_b[i] = i;
	}

	// Sincroniza os potenciais virtualmente, ou seja, ontem apenas o novo vetor de index
	swap = 0;
	for (i=0;i<pot_a->nvars;i++) {
		for (j=swap;j<pot_b->nvars;j++) {
			if (pot_a->vars[index_a[i]]==pot_b->vars[index_b[j]]) {
				if (index_a[i]!=index_b[j] || index_a[i]!=swap) {
					// swap no index pot_a->vars
					if (index_a[i]!=swap) {
						swap_a = index_a[swap];
						index_a[swap] = index_a[i];
						index_a[i] = swap_a;
					}
					// swap no index pot_b->vars
					if (index_b[j]!=swap) {
						swap_b = index_b[swap];
						index_b[swap] = index_b[j];
						index_b[j] = swap_b;
					}
				}
				// Incrementa a posio de swap
				swap++;
				break;
			}
		}	
	}

	// Calcula os multiplicadores, se necessrio
	if (swap_a!=-1) {
		src_mult_a = malloc(pot_a->nvars*sizeof(int));
		dst_mult_a = malloc(pot_a->nvars*sizeof(int));
		src_mult_a[0] = 1;
		dst_mult_a[0] = 1;
		for (i=1;i<pot_a->nvars;i++) {
			src_mult_a[i] = src_mult_a[i-1]*pot_a->vars[i-1]->nstates;
			dst_mult_a[i] = dst_mult_a[i-1]*pot_a->vars[index_a[i-1]]->nstates;
		}
	}
	if (swap_b!=-1) {
		src_mult_b = malloc(pot_b->nvars*sizeof(int));
		dst_mult_b = malloc(pot_b->nvars*sizeof(int));
		src_mult_b[0] = 1;
		dst_mult_b[0] = 1;
		for (i=1;i<pot_b->nvars;i++) {
			src_mult_b[i] = src_mult_b[i-1]*pot_b->vars[i-1]->nstates;
			dst_mult_b[i] = dst_mult_b[i-1]*pot_b->vars[index_b[i-1]]->nstates;
		}
	}

	// Define as variveis do potentical resultante com base nos potenciais sincronizados
	// Incializa variveis para armazenar dados temporrios
	nequals = 0;
	mult = 1;
	nvalues = 1;

	// Cria um array (mesmo que superdimensionado) para determinar as variaveis do potencial resultante
	vars = malloc((pot_a->nvars+pot_b->nvars-1)*sizeof(Variable));

	j = 0;
	for (i=0;i<pot_a->nvars;i++) {

		// Dimenso da varivel de pot_a
		nvalues = nvalues*pot_a->vars[index_a[i]]->nstates;

		// Acrescenta a varivel do pot_a no conjunto de variveis do potencial resultante
		vars[j++] = pot_a->vars[index_a[i]];

		// Compara a sequncia de variveis dos potenciais
		if (pot_a->vars[index_a[i]] == pot_b->vars[index_b[i]]) {
			nequals++;	
			mult = mult*pot_a->vars[index_a[i]]->nstates;
		}
	}

	for (i=nequals;i<pot_b->nvars;i++) {

		// Dimenso da varivel de pot_b
		nvalues = nvalues*pot_b->vars[index_b[i]]->nstates;

		// Acrescenta a varivel do pot_b no conjunto de variveis do potencial resultante
		vars[j++] = pot_b->vars[index_b[i]];
	}

	/*
	j = 0;
	for (i=0;i<pot_b->nvars;i++) {
		if (i<pot_a->nvars) {

			// dimenso da varivel de pot_a
			nvalues = nvalues*pot_a->vars[index_a[i]]->nstates;

			// Acrescenta a varivel do pot_a no conjunto de variveis do potencial resultante
			vars[j++] = pot_a->vars[index_a[i]];

			// Compara a sequncia de variveis dos potenciais
			if (pot_a->vars[index_a[i]] == pot_b->vars[index_b[i]]) {
				nequals++;	
				mult = mult*pot_a->vars[index_a[i]]->nstates;
			} else {
				// dimenso da varivel de pot_b
				nvalues = nvalues*pot_b->vars[index_b[i]]->nstates;

				// Acrescenta a varivel do pot_a no conjunto de variveis do potencial resultante
				vars[j++] = pot_b->vars[index_b[i]];
			}
			continue;

		} else {
			// Dimenso da varivel de pot_b
			nvalues = nvalues*pot_b->vars[index_b[i]]->nstates;

			// Acrescenta a varivel do pot_a no conjunto de variveis do potencial resultante
			vars[j++] = pot_b->vars[index_b[i]];
		}
	}
	*/

	// Nmero de variveis do potencial resultado
	nvars = nequals + (pot_a->nvars - nequals) + (pot_b->nvars - nequals);

	// Inicializa o potentical resultante

	// Aloca a memria para o potencial resultante
	result = malloc(sizeof(Potential));

	// Valor arbitrrio (=-1) o valor final dever ser a posio no vetor de referncias a potenciais
	result->id = -1;

	// Variveis do potencial
	result->nvars = nvars;
	result->vars = malloc(nvars*sizeof(Variable*));

	// Variveis do vetor de resultados
	for (i=0;i<nvars;i++) {
		result->vars[i] = vars[i];
		vars[i] = NULL;
	}

	// Aloca o espao para os valores
	result->nvalues = nvalues;
	result->values = malloc(nvalues*sizeof(double));

	// Faz efetivamente o produto
	l = 0;
	for (j=0;j<pot_b->nvalues;j=j+mult){
		for (i=0;i<pot_a->nvalues;i=i+mult){
			for (k=0;k<mult;k++) {

				// Calcula o index permutado dos valores do potencial a
				if (swap_a!=-1) {
					value_index_a = 0;
					rest = i + k;
					for (m=pot_a->nvars-1;m>=0;m--) {
						state = (int)(rest/dst_mult_a[m]); 
						rest = rest - state*dst_mult_a[m];
						value_index_a = value_index_a + state*src_mult_a[index_a[m]];
					}
				} else {
					value_index_a = i + k;
				}

				// Calcula o index permutado dos valores do potencial b
				if (swap_b!=-1) {
					value_index_b = 0;
					rest = j + k;
					for (m=pot_b->nvars-1;m>=0;m--) {
						state = (int)(rest/dst_mult_b[m]); 
						rest = rest - state*dst_mult_b[m];
						value_index_b = value_index_b + state*src_mult_b[index_b[m]];
					}
				} else {
					value_index_b = j + k;
				}

				// Executa o produto
				result->values[l] = pot_a->values[value_index_a]*pot_b->values[value_index_b];
				l++;
			}
		}
	}

	// Libera a memria
	free(index_a);
	if (swap_a!=-1) {
		free(src_mult_a);
		free(dst_mult_a);
	}

	free(index_b);
	if (swap_b!=-1) {
		free(src_mult_b);
		free(dst_mult_b);
	}

	// Varivel auxiliar com as variveis
	free(vars);

	return result;
}

Potential* marginalizePotential (Variable* variable, Potential* pot) {

	Potential* result;

	// Variveis auxiliares para permutar os indces se necessrio
	int index[pot->nvars];
	int src_mult[pot->nvars];
	int dst_mult[pot->nvars];
	int value_index;

	// Variveis auxiliares para converso de index do nvalues;
	int state;
	int rest;

	// Soma da marginalizao
	double sum;

	int i,j,k,m;

	// Garante que a varivel a ser marginalizada  a primeira
	
	// Inicializa o vetor de indces
	for (i=0;i<pot->nvars;i++) {
		index[i] = i;
		if(variable==pot->vars[i]) {
			index[0] = i;
			index[i] = 0;
		}
	}

	// Verifica se a primeira varivel  a varivel que ser marginalizada
	if (pot->vars[index[0]]!=variable) {
		fprintf(stderr,"Potential id: %d does not contain variable '%s'\n",pot->id,variable->name);
		exit(1);
	}


	// Inicializa os multimplicadores apenas se necessrio para permutar os valores
	if (variable!=pot->vars[0]) {
		src_mult[0] = 1;
		dst_mult[0] = 1;
		for (i=1;i<pot->nvars;i++) {
			src_mult[i] = src_mult[i-1]*pot->vars[i-1]->nstates;
			dst_mult[i] = dst_mult[i-1]*pot->vars[index[i-1]]->nstates;
		}
	}


	// Aloca a memria para o potencial resultante
	result = malloc(sizeof(Potential));

	// Valor arbitrrio (=-1) o valor final dever ser a posio no vetor de referncias a potenciais
	result->id = -1;

	// Variveis do potencial
	result->nvars = pot->nvars - 1;
	result->vars = malloc(result->nvars*sizeof(Variable*));
	for (i=1;i<pot->nvars;i++) result->vars[i-1] = pot->vars[index[i]];

	// Aloca o espao para os valores
	result->nvalues = pot->nvalues/variable->nstates;
	result->values = malloc(result->nvalues*sizeof(double));

	// Marginaliza a varivel
	k = 0;
	for (i=0;i<pot->nvalues;i=i+pot->vars[index[0]]->nstates){
		sum = 0;
		for (j=0;j<pot->vars[index[0]]->nstates;j++) {

			// Calcula o valor do index pois pode ter ocorrido permutao
			if (variable!=pot->vars[0]) {
				value_index = 0;
				rest = i + j;
				for (m=pot->nvars-1;m>=0;m--) {
					state = (int)(rest/dst_mult[m]); 
					rest = rest - state*dst_mult[m];
					value_index = value_index + state*src_mult[index[m]];
				}
			} else {
				value_index = i + j;
			}

			// Soma os valores da varivel a ser marginalizada
			sum = sum + pot->values[value_index];
		}

		result->values[k] = sum;
		k++;
	}

	return result;
}




/* Executa a multiplicao efetuando a permutao "fisica" dos potenciais */

Potential* permuteMultPotentials (Variable* first, Potential** refPotential_1, Potential** refPotential_2) {

	Potential* pot_a;
	Potential* pot_b;
	Potential* result;

	int nequals;	// Nmero de variveis iguais nos dois potenciais
	int mult;		// Tamanho do bloco formado pela variveis iguais
	int nvalues;	// Nmero de valores no potencial resultante
	int nvars;		// Nmero de variveis no potencial resultante

	Variable** vars; // Variveis do potencial resultante

	int i,j,k,l;

	// Define qual o potencial com menor nmero de variveis
	if ((*refPotential_1)->nvars < (*refPotential_2)->nvars) {
		pot_a = *refPotential_1;
		pot_b = *refPotential_2;
	} else {
		pot_a = *refPotential_2;
		pot_b = *refPotential_1;
	}

	// Sincroniza os potenciais
	syncPotentials(first,&pot_a,&pot_b);

	printf("Potenciais sincronizados:\n");
	printf("Potencial a:");
	dumpPotential(pot_a);
	printf("Potencial b:");
	dumpPotential(pot_b);

	// Define as variveis do potentical resultante com base nos potenciais sincronizados
	
	// Incializa variveis para armazenar dados temporrios
	nequals = 0;
	mult = 1;
	nvalues = 1;

	// Cria um array (mesmo que superdimensionado) para determinar as variaveis do potencial resultante
	vars = malloc((pot_a->nvars+pot_b->nvars-1)*sizeof(Variable));


	// Importante: pot_a (o menor) e pot_b (o maior) devem necessariamente estar sincronizados
	j = 0;
	for (i=0;i<pot_a->nvars;i++) {

		// dimenso da varivel de pot_a
		nvalues = nvalues*pot_a->vars[i]->nstates;

		// Acrescenta a varivel do pot_a no conjunto de variveis do potencial resultante
		vars[j++] = pot_a->vars[i];

		// Compara a sequncia de variveis dos potenciais
		if (pot_a->vars[i] == pot_b->vars[i]) {
			nequals++;	
			mult = mult*pot_a->vars[i]->nstates;
		}
	}


	for (i=nequals;i<pot_b->nvars;i++) {

		// dimenso da varivel de pot_a
		nvalues = nvalues*pot_b->vars[i]->nstates;

		// Acrescenta a varivel do pot_a no conjunto de variveis do potencial resultante
		vars[j++] = pot_b->vars[i];

	}

	// Nmero de variveis do potencial resultado
	nvars = nequals + (pot_a->nvars - nequals) + (pot_b->nvars - nequals);

	/* DEBUG
	printf("nequals: %d mult: %d nvalues: %d nvars: %d\n",nequals,mult,nvalues,nvars); 
	printf("{"); for (i=0;i<nvars;i++) printf("%s,",vars[i]->name); printf("}\n");
	*/
	
	// Inicializa o potentical resultante

	// Aloca a memria para o potencial resultante
	result = malloc(sizeof(Potential));

	// Valor arbitrrio (=-1) o valor final dever ser a posio no vetor de referncias a potenciais
	result->id = -1;

	// Variveis do potencial
	result->nvars = nvars;
	result->vars = vars;

	// Aloca o espao para os valores
	result->nvalues = nvalues;
	result->values = malloc(nvalues*sizeof(double));

	// Faz efetivamente o produto
	l = 0;
	for (j=0;j<pot_b->nvalues;j=j+mult){
		for (i=0;i<pot_a->nvalues;i=i+mult){
			for (k=0;k<mult;k++) {
				result->values[l] = pot_a->values[i+k]*pot_b->values[j+k];
				l++;
			}
		}
	}

	return result;
}


/* Executa a marginalizao efetuando a permutao "fisica" dos potenciais */

Potential* permuteMarginalizePotential (Variable* variable, Potential** refPotential) {

	Potential* pot = *refPotential;
	Potential* result;
	
	double sum;

	int fromto[pot->nvars];	// varivel auxiliar eventualmente necessria para permutar o potencial

	int i,j,k;

	// Garante que a varivel a ser marginalizada  a primeira
	if (variable!=pot->vars[0]) {
		for (i=0;i<pot->nvars;i++) fromto[i] = i;
		for (i=0;i<pot->nvars;i++) {
			if(variable==pot->vars[i]) {
				fromto[0] = i;
				fromto[i] = 0;
				break;
			}
		}
		// Permuta o potencial
		permutePotential(fromto,&pot);
	}

	// Aloca a memria para o potencial resultante
	result = malloc(sizeof(Potential));

	// Valor arbitrrio (=-1) o valor final dever ser a posio no vetor de referncias a potenciais
	result->id = -1;

	// Variveis do potencial
	result->nvars = pot->nvars - 1;
	result->vars = malloc(result->nvars*sizeof(Variable*));
	for (i=1;i<pot->nvars;i++) result->vars[i-1] = pot->vars[i];

	// Aloca o espao para os valores
	result->nvalues = pot->nvalues/pot->vars[0]->nstates;
	result->values = malloc(result->nvalues*sizeof(double));

	// Marginaliza a varivel
	k = 0;
	for (i=0;i<pot->nvalues;i=i+pot->vars[0]->nstates){
		sum = 0;
		for (j=0;j<pot->vars[0]->nstates;j++) {
			sum = sum + pot->values[i+j];
		}
		result->values[k] = sum;
		k++;
	}

	return result;
}


void permutePotential (int* from_to, Potential** refPotential) {

		Potential* pot = *refPotential;
		Variable** dst_vars;
		double* dst_values;
		int* states; 
		int src_mult[pot->nvars];
		int dst_mult[pot->nvars];
		int dst_nstates[pot->nvars];
		int i,j;
		int src,dst;
		int rest;

		// Aloca memria para o array de valores permutado
		dst_values = malloc(pot->nvalues*sizeof(double));

		// Aloca memria para o array na nova ordem das variveis
		dst_vars = malloc(pot->nvars*sizeof(Variable*));

		// Aloca memria para o vetor temporrio de estados
		states = malloc(pot->nvars*sizeof(int));

		// Monta o array com a nova ordem das variveis e o vetor auxiliar com o nmero de estados das variveis
		for (i=0;i<pot->nvars;i++) {
			dst_nstates[from_to[i]] = pot->vars[i]->nstates;
			dst_vars[from_to[i]] = pot->vars[i];
		}

		// Monta os vetores de multiplicadores
		src_mult[0] = 1;
		dst_mult[0] = 1;
		for (i=1;i<pot->nvars;i++) {
			src_mult[i] = src_mult[i-1]*pot->vars[i-1]->nstates;
			dst_mult[i] = dst_mult[i-1]*dst_nstates[i-1];
		}
		
		// Percorre o vetor de valores fazendo permutao
		for (src=0;src<pot->nvalues;src++) {
			dst = 0;	
			rest = src;
			for (i=pot->nvars-1;i>=0;i--) {
				// "Descasca" o ndice do vetor de origem para obter os valores dos estados
				states[i] = (int)(rest/src_mult[i]); 
				rest = rest - states[i]*src_mult[i];

				// Com base nos estados compe o ndice do vetor de destimo
				dst = dst + states[i]*dst_mult[from_to[i]];
			}

			// Atribui o valor ao vetor de destino
			dst_values[dst] = pot->values[src];
		}
	
		// Desaloca a memria
		free(states);

		// Desaloca o array corrente de valores do potencial
		free(pot->values);

		// Atribui o novo conjunto de valores
		pot->values = dst_values;

		// Desaloca o array corrente de variveis
		free(pot->vars);

		// Atribui o novo array de variveis
		pot->vars = dst_vars;

}

void syncPotentials (Variable* first, Potential** refPotential_1, Potential** refPotential_2) {

	// Variveis auxiliares "a": menor array; "b": maior array
	Potential* pot_a;
	Potential* pot_b;
	int *index_a, *index_b;
	int *fromto_a, *fromto_b;

	int i,j;
	int swap;
	int swap_a = -1;
	int swap_b = -1;


	// "pot_a->vars" deve ser o menor array
	if ((*refPotential_1)->nvars <= (*refPotential_2)->nvars) {
		pot_a = *refPotential_1;
		pot_b = *refPotential_2;
	} else {
		pot_a = *refPotential_2;
		pot_b = *refPotential_1;
	}

	// Inicializa os arrays de->para
	index_a = malloc(pot_a->nvars*sizeof(int));
	fromto_a = malloc(pot_a->nvars*sizeof(int));
	index_b = malloc(pot_b->nvars*sizeof(int));
	fromto_b = malloc(pot_b->nvars*sizeof(int));
	for (i=0;i<pot_b->nvars;i++) {
		if (i<pot_a->nvars) index_a[i] = i;
		index_b[i] = i;
	}

	// Sincroniza os potenciais comeando pela primeira posio

	// Fora o primeiro elemento ser o elemento inicial
	swap = 0;
	if (first!=pot_a->vars[index_a[0]] || first!=pot_b->vars[index_b[0]]) {
		for (i=0;i<pot_b->nvars;i++) {
			if (first!=pot_a->vars[index_a[0]] && i<pot_a->nvars) {
				if(first==pot_a->vars[index_a[i]]) {
					// swap no index
					swap_a = index_a[0];
					index_a[0] = index_a[i];
					index_a[i] = swap_a;
				}
			}
			if (first!=pot_b->vars[index_b[0]]) {
				if(first==pot_b->vars[index_b[i]]) {
					// swap no index
					swap_b = index_b[0];
					index_b[0] = index_b[i];
					index_b[i] = swap_b;
				}
			}
			if (first==pot_a->vars[index_a[0]] && first==pot_b->vars[index_b[0]]) break;
		}
	}

	// Verifica se em ambos vetores o primeiro elemento  o elemento correto
	if (!(first==pot_a->vars[index_a[0]] && first==pot_b->vars[index_b[0]])) {
		fprintf(stderr,"first variable not found ipot_a->vars[index_a[0]]=%s pot_b->vars[index_b[0]]=%s\n",pot_a->vars[index_a[0]]->name,pot_b->vars[index_b[0]]->name);
		exit(1);
	}
	
	fromto_a[index_a[swap]] = swap;
	fromto_b[index_b[swap]] = swap;

	// Sincroniza o restante dos potenciais
	swap = 1;
	for (i=1;i<pot_a->nvars;i++) {
		for (j=swap;j<pot_b->nvars;j++) {
			if (pot_a->vars[index_a[i]]==pot_b->vars[index_b[j]]) {
				if (index_a[i]!=index_b[j] || index_a[i]!=swap) {
					// swap no index pot_a->vars
					if (index_a[i]!=swap) {
						swap_a = index_a[swap];
						index_a[swap] = index_a[i];
						index_a[i] = swap_a;
					}
					// swap no index pot_b->vars
					if (index_b[j]!=swap) {
						swap_b = index_b[swap];
						index_b[swap] = index_b[j];
						index_b[j] = swap_b;
					}
				}
				// Incrementa a posio de swap
				fromto_a[index_a[swap]] = swap;
				fromto_b[index_b[swap]] = swap;
				swap++;
				break;
			}
		}	
	}

	/* DEBUG
	for (i=0;i<pot_a->nvars;i++) printf("pot_a->vars[%d] = %s\n",i,pot_a->vars[i]->name); printf("\n");
	for (i=0;i<pot_a->nvars;i++) printf("index_a[%d] = %d pot_a->vars[index_a[%d]] = %s\n",i,index_a[i],i,pot_a->vars[index_a[i]]->name); printf("\n");
	for (i=0;i<pot_b->nvars;i++) printf("pot_b->vars[%d] = %s\n",i,pot_b->vars[i]->name); printf("\n");
	for (i=0;i<pot_b->nvars;i++) printf("index_b[%d] = %d pot_b->vars[index_b[%d]] = %s\n",i,index_b[i],i,pot_b->vars[index_b[i]]->name); printf("\n");
	*/

	// Se necessrio permunta o potencial A
	if (swap_a!=-1)  {
		for (i=swap;i<pot_a->nvars;i++) fromto_a[index_a[i]] = i;
		permutePotential(fromto_a,&pot_a);
	}

	// Se necessrio permunta o potencial B
	if (swap_b!=-1) {
		for (i=swap;i<pot_b->nvars;i++) fromto_b[index_b[i]] = i;
		permutePotential(fromto_b,&pot_b);
	}

	// Libera a memria alocada
	free(index_a);
	free(fromto_a);
	free(index_b);
	free(fromto_b);

}

Variable** fakeVariables (int nvars, int* nstates) {

	char* varNames[] = {"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","X","W","Y","Z"};
	char* stateNames[] = {"STATE_1","STATE_2","STATE_3","STATE_4","STATE_5","STATE_6","STATE_7","STATE_8","STATE_9","STATE_10"};

	Variable** vars;

	int i;

	// Aloca o espao para o vetor de variveis
	vars = malloc(nvars*sizeof(Variable*));

	// Cria as variveis e determina a quantidade de valores do potential
	for (i=0;i<nvars;i++) {
		vars[i] = buildVariable(i,varNames[i],VARIABLE_TYPE_DISCRETE,nstates[i],stateNames);

		if (nstates[i]>10 || nvars>26) {
			fprintf(stderr,"potential.c fakeVariables (int nvars, int* nstates): 'nvars' (max: 26) and 'nstates[i]' (max: 10)\n");
			exit(1);
		}
	}

	return vars;
}

Potential* fakePotential (int id, int first, int offset, Variable** vars) {

	Potential* pot;

	double* values;
	double base = 100;

	int nvalues;
	int i;

	int nvars = offset;
	Variable** potvars = malloc(nvars*sizeof(Variable*));

	// Cria as variveis e determina a quantidade de valores do potential
	nvalues = 1;
	for (i=first;i<(first+offset);i++) {
		potvars[i-first] = vars[i];
		nvalues = nvalues * vars[i]->nstates;
	}

	// Cria os valores com nmero randomicos
	values = malloc(nvalues*sizeof(double));
	for (i=0;i<nvalues;i++) values[i] = (fmod(rand(),base))/base;

	// Cria o potencial
	pot = buildPotential(id,nvars,potvars,values); 

	// Verificar se  necessrio - aparentemente no  necessrio
	// for (i=0;i<nvars;i++) potvars[i] = NULL;

	// Desaloca a memria
	free(potvars);
	potvars = NULL;

	free(values);
	values = NULL;

	return pot;
}
