/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Fevereiro/2007
 *
 * Para documentao ver singlelinkedlist.h
 */

/* Includes */

/* Lista o cabealho das funes e estrutura de dados */

#include "singlelinkedlist.h"

/* Funes */

SingleLinkedList* newSingleLinkedList () {
	SingleLinkedList* list = malloc(sizeof(SingleLinkedList));
	list->length = 0;
	list->first = NULL;
	list->last = NULL;
	return list;
}

void destroySingleLinkedList (SingleLinkedList** refList) {
	SingleLinkedList* list = *refList;
	SingleLinkedListNode* current;
	SingleLinkedListNode* next;

	/* Percorre a lista liberando os SingleLinkedListNodes */
	current = list->first;
	while (current != NULL) {
		next = current->next;
		destroySingleLinkedListNode(&current);
		current = next;
	}
	
	/* Libera lista */
	free(list);
	*refList = NULL;
}

SingleLinkedListNode* newSingleLinkedListNode () {
	SingleLinkedListNode* node = malloc(sizeof(SingleLinkedListNode));
	node->data = NULL;
	node->next = NULL;
	return node;
}

SingleLinkedListNode* buildSingleLinkedListNode (SLLDATA data) {
	SingleLinkedListNode* node = newSingleLinkedListNode();
	node->data = data;
	node->next = NULL;
	return node;
}

void destroySingleLinkedListNode(SingleLinkedListNode** refNode) {
	SingleLinkedListNode* node = *refNode;
	free(node);
	*refNode = NULL;
}

int slllength(SingleLinkedList* list) {
	return list->length;
}

int sllpush (SingleLinkedList** refList, SLLDATA data) {
	return sllsyncpush (refList,data,NULL);
}

int sllsyncpush (SingleLinkedList** refList, SLLDATA data, pthread_mutex_t* refMutex) {
	SingleLinkedList* list = *refList;
	SingleLinkedListNode* node = newSingleLinkedListNode();

	/* Inicializa o novo node */
	node->data = data;
	node->next = NULL;

	/* Verifia se o mutex est definido e tenta trav-lo */
	if (refMutex && pthread_mutex_lock(refMutex)) {
		/* Ocorreu um erro */
		perror("pthread_mutex_lock");
		pthread_exit(NULL);
	}

	/* Verifica se a lista foi inicializada, ou seja, primeiro elemento  null */
	if (list->first == NULL) {
		/* Inicializa o tamanho */
		list->length = 1;
		/* Inicializa o primeiro e o ltimo elemento com o node recm criado */
		list->first = node;
		list->last = node;
	} else {
		/* Incrementa o tamanho */
		list->length = list->length + 1;
		node->next = list->first;
		list->first = node;
	}

	/* Verifia se o mutex est definido e tenta destrav-lo */
	if (refMutex && pthread_mutex_unlock(refMutex)) {
		/* Ocorreu um erro */
		perror("pthread_mutex_unlock");
		pthread_exit(NULL);
	}

	return 1;
}

int sllappend(SingleLinkedList** refList, SLLDATA data) {
	return sllsyncappend (refList,data,NULL);
}

int sllsyncappend (SingleLinkedList** refList, SLLDATA data, pthread_mutex_t* refMutex) {
	SingleLinkedList* list = *refList;
	SingleLinkedListNode* node = newSingleLinkedListNode();

	/* Inicializa o novo node */
	node->data = data;
	node->next = NULL;

	/* Verifia se o mutex est definido e tenta trav-lo */
	if (refMutex && pthread_mutex_lock(refMutex)) {
		// Ocorreu um erro
		perror("pthread_mutex_lock");
		pthread_exit(NULL);
	}

	/* Verifica se a lista foi inicializada, ou seja, primeiro elemento  null */
	if (list->first == NULL) {
		/* Inicializa o tamanho */
		list->length = 1;
		/* Inicializa o primeiro e o ltimo elemento com o node recm criado */
		list->first = node;
		list->last = node;
	} else {
		/* Incrementa o tamanho */
		list->length = list->length + 1;
		list->last->next = node;
		list->last = node;
	}

	/* Verifia se o mutex est definido e tenta destrav-lo */
	if (refMutex && pthread_mutex_unlock(refMutex)) {
		/* Ocorreu um erro */
		perror("pthread_mutex_unlock");
		pthread_exit(NULL);
	}

	return 1;
}

SLLDATA sllremove_first(SingleLinkedList** refList) {
	SingleLinkedList* list = *refList;
	SingleLinkedListNode* node;
	SLLDATA data;

	/* Verifica se no  uma lista vazia */
	if (list->first == NULL) return (int) NULL;

	/* Decrementa o tamanho */
	list->length = list->length - 1;

	/* Atribui o contedo do primeiro node para uma varivel auxiliar */
	node = list->first;
	data = node->data;

	if (node->next == NULL) {
		list->first = NULL;
		list->last = NULL;
	} else {
		list->first = node->next;	
	}

	/* Libera o SingleLinkedListNode */
	free(node);

	/* Retorna o contedo que estava armazenado no SingleLinkedListNode */
	return data;

}

SLLDATA sllremove_last(SingleLinkedList** refList) {
	SingleLinkedList* list = *refList;
	SingleLinkedListNode* node;
	SingleLinkedListNode* current;
	SLLDATA data;

	/* Verifica se no  uma lista vazia */
	if (list->first == NULL) return (int) NULL;

	/* Decrementa o tamanho */
	list->length = list->length - 1;

	/* Atribui o contedo do ltimo node para uma varivel auxiliar */
	node = list->last;
	data = node->data;

	/* Percorre a lista para saber quem aponta para o 'last' */
	current = list->first;
	while (current->next != node) {
		current = current->next;	
	}
	current->next = NULL;
	list->last = current;

	/* Libera o SingleLinkedListNode */
	free(node);

	/* Retorna o contedo que estava armazenado no SingleLinkedListNode */
	return data;

}

void dumpSingleLinkedList(SingleLinkedList* list) {
	SingleLinkedListNode* current = list->first;
	int i = 0;

	while (current!=NULL) {
		printf("#%d data: %s next: ",(i++),(current->data==NULL?"NULL":"SLLDATA"));
		if (current->next!=NULL) printf("#%d",i); else printf("NULL");
		printf("\n");
		current = current->next;
	}
}
