/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*!
 * @file	singlelinkedlist.h
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	2007-04-30
 * @brief 	Implementao uma biblioteca de lista simplesmente ligada genrica (Single Linked List)
 *
 * Implementao de uma biblioteca de lista ligada simples genrica
 * com ponteiros para o primeiro e o ltimo elementos para que a
 * a insero de um novo elemento no incio e no final seja em tempo
 * constante.
 * A implentao foi feita para ser \e threadsafe, ou seja, pode ser 
 * utilizada para aplicao multithreads.
 * 
 */

#ifndef SINGLELINKEDLIST_H
#define SINGLELINKEDLIST_H

/* Includes */
#include <math.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

/* Defines */

/* Definio dos tipos de dados */

/*! Alias para ponteiro void para armazenar dados nos nodes da lista ligada */
typedef void* SLLDATA;

/*! Node de uma Single Linked Liste de uma lista ligada */
typedef struct struct_SingleLinkedListNode {
	SLLDATA data;									//!< Varivel que armazena o contedo do node
	struct struct_SingleLinkedListNode* next;		//!< Ponteiro para o prximo elemento da lista ligada ou \e NULL se for o ltimo
} SingleLinkedListNode;

/*! Single Linked List */
typedef struct struct_SingleLinkedList {
	int length;										//!< Tamanho da lista, ou seja, nmero de ns atual
	struct struct_SingleLinkedListNode* first;		//!< Ponteiro para o primeiro n da lista
	struct struct_SingleLinkedListNode* last;		//!< Ponteiro para o ltimo elemento da lista
} SingleLinkedList;


/* Funes */

/*!
 * @brief Cria uma nova lista ligada simples.
 *
 * Para desalocar a memria utilizar destroySingleLinkedList
 * @return Ponteiro para uma lista ligada
 */

SingleLinkedList* newSingleLinkedList ();

/*!
 * @brief Destri uma lista ligada simples, mas NO destri o contedo dos ns da lista ligada. 
 * @param[out] refList Referncia  lista que deve ser desalocada, atribui NULL para refrencia
 */

void destroySingleLinkedList (SingleLinkedList** refList);

/*!
 * @brief Cria um novo node da lista ligada com contedo NULL
 * @return Ponteiro para um novo node da lista ligada de vrtices
 */

SingleLinkedListNode* newSingleLinkedListNode ();

/*!
 * @brief Cria e inicializa o contedo de um novo node da lista ligada
 * @param[in] data	Dado a ser armazenado no Node
 * @return Ponteiro para um novo node da lista ligada de vrtices
 */

SingleLinkedListNode* buildSingleLinkedListNode (SLLDATA data);

/*!
 * @brief Destri um node da lista ligada simples, mas NO destri o contedo dos ns da lista ligada 
 * @param[out] refNode Referncia ao Node da lista que deve ser destruido, atribui NULL para refrencia
 */

void destroySingleLinkedListNode(SingleLinkedListNode** refNode);

/*!
 * @brief Retorna o tamanho da lista
 * @param[in] list Lista da qual se deseja saber o tamanho
 * @return Tamanho da lista
 */

int slllength(SingleLinkedList* list);

/*!
 * @brief Acrescenta um SingleLinkedListNode no incio da lista ligada.
 * 
 * O tempo de insero de um novo elemento no incio ou no fim da lista  linear.
 * O \e node inserido passa a ser o \e first.
 * @param[out] refList Referncia para lista na qual deve ser acrescentado o node
 * @param[in] data Contedo do SingleLinkedListNode que ser criado e que dever ser acrescentado no incio da lista
 * @return 1 se inseriu com sucesso; 0 caso contrrio
 */

int sllpush (SingleLinkedList** refList, SLLDATA data);

/*!
 * @brief Acrescenta um SingleLinkedListNode no incio da lista ligada garantindo a atomicidade da operao verso que deve ser utilizada em aplicaes multi-thread que compartilham listas ligadas
 * 
 * O tempo de insero de um novo elemento no incio ou no fim da lista  linear.
 * O \e node inserido passa a ser o \e first.
 * @param[out] refList Referncia para lista na qual deve ser acrescentado o node
 * @param[in] data Contedo do SingleLinkedListNode que ser criado e que dever ser acrescentado no incio da lista
 * @param[in] mutex Mutex utilizado para sincronizar a operao
 * @return 1 se inseriu com sucesso; 0 caso contrrio
 */

int sllsyncpush (SingleLinkedList** refList, SLLDATA data, pthread_mutex_t* refMutex);

/*!
 * @brief Acrescenta um SingleLinkedListNode no final da lista ligada. 
 *
 * O tempo de insero de um novo elemento no incio ou no fim da lista  linear.
 * O \e node inserido passa a ser o \e last.
 * @param[out] refList Referncia para lista na qual deve ser acrescentado o node
 * @param[in] data Contedo do SingleLinkedListNode que ser criado e que dever 
 * ser acrescentado no final da lista
 * @return 1 se inseriu com sucesso; 0 caso contrrio
 */

int sllappend(SingleLinkedList** refList, SLLDATA data);

/*!
 * @brief Acrescenta um SingleLinkedListNode no final da lista ligada garantindo a atomicidade da operao verso que deve ser utilizada em aplicaes multi-thread que compartilham listas ligadas
 *
 * O tempo de insero de um novo elemento no incio ou no fim da lista  linear.
 * O \e node inserido passa a ser o \e last.
 * @param[out] refList Referncia para lista na qual deve ser acrescentado o node
 * @param[in] data Contedo do SingleLinkedListNode que ser criado e que dever 
 * ser acrescentado no final da lista
 * @param[in] mutex Mutex utilizado para sincronizar a operao
 * @return 1 se inseriu com sucesso; 0 caso contrrio
 */

int sllsyncappend (SingleLinkedList** refList, SLLDATA data, pthread_mutex_t* refMutex);

/*! 
 * @brief Remove o primeiro elemento da lista
 * @param[out] refList Referncia para lista da qual deve ser removido o node
 * @return O contedo que estava armazenado no SingleLinkedListNode
 */

SLLDATA sllremove_first(SingleLinkedList** refList);

/*!
 * @brief Remove o ltimo elemento da lista
 * @param[out] refList Referncia para lista da qual deve ser removido o node
 * @return O contedo que estava armazenado no SingleLinkedListNode
 */

SLLDATA sllremove_last(SingleLinkedList** refList);

/*! 
 * @brief Dump do contedo da lista ligada
 * @param[in] list Lista ligada
 */

void dumpSingleLinkedList(SingleLinkedList* list);

#endif // SINGLELINKEDLIST_H
