/*
* Copyright 2007, 2008 Ernesto Coutinho Colla
*
* "BN Parallel Package"  um nome que identifica o conjunto de programas 
* que forma uma biblioteca de rotinas que foram desenvolvidas como parte da 
* dissertao de mestrado do autor em Cincias da Computao.
*
* O conjunto de programas foi integralmente desenvolvido pelo autor e est
* disponvel sob a licena GPL (GNU General Public License). O entendimento
* integral dos termos da licena GPL  condio necessria para a utilizao
* parcial ou integral de qualquer parte deste conjunto de programas.
*
* This file is part of BN Parallel Package.
* 
* BN Parallel Package is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License.
*
* BN Parallel Package is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with BN Parallel Package. If not, see <http://www.gnu.org/licenses/>.
*/


/*
 * @author	Ernesto Colla (ernesto@gmail.com)
 * @version	0.0.1
 * @date	Fevereiro/2007 (2007-02-18)
 *
 * Para documentao ver ThreadTree.h
 */

/* Includes */

#include <stdio.h>
#include <stdlib.h>

/* Lista o cabealho das funes e estrutura de dados */
#include "threadtree.h"

/* Funes */

ThreadTree* buildThreadTree() {
	ThreadTree* threadtree = malloc(sizeof(ThreadTree));
	threadtree->vertices = buildThreadVertexLinkedList();
	threadtree->root = NULL;
	threadtree->index = NULL;
	return threadtree;
}

void destroyThreadTree(ThreadTree** refThreadTree) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertexNode* current;
	ThreadVertex* vertex;

	// Libera a memria da lista de index
	destroyThreadTreeIndex(&threadtree);

	// Libera a memria dos vrtices
	
	current = threadtree->vertices->first;
	while (current != NULL) {
		vertex = current->vertex;
		destroyThreadVertex(&vertex);
		current = current->next;
	}
	
	// Desaloca a lista de vtices
	destroyThreadVertexLinkedList(&(threadtree->vertices));

	// Libera ThreadTree
	free(threadtree);
	*refThreadTree = NULL;
}

ThreadVertex* buildThreadVertex (void* task, void* data) {
	ThreadVertex* vertex = malloc(sizeof(ThreadVertex));
	vertex->id = -1;
	vertex->status = THREAD_READY;
	vertex->task = task;
	vertex->data = data;
	vertex->parent = NULL;
	vertex->children = buildThreadVertexLinkedList();

	return vertex;
}

void destroyThreadVertex (ThreadVertex** refThreadVertex) {
	ThreadVertex* vertex = *refThreadVertex;

	// Desaloca a lista de vtices
	destroyThreadVertexLinkedList(&(vertex->children));

	vertex->task = NULL;
	vertex->data = NULL;
	vertex->parent = NULL;
	vertex->children = NULL;

	// vertex->tid = NULL;					// Comentada na verso linux

	// Libera ThreadVertex
	free(vertex);
	*refThreadVertex = NULL;
}

int addThreadVertex(ThreadTree** refThreadTree, ThreadVertex** refThreadVertex) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertex* vertex = *refThreadVertex;

	// Determina o Id da vrtice de acordo com a ordem de insero
	vertex->id = threadtree->vertices->length;

	// Verifica se esta indexado, se estiver elimina o index
	if (threadtree->index != NULL) destroyThreadTreeIndex(&threadtree);

	// Acrecenta o vrtice no final da lista ligada
	return appendThreadVertexNode(&(threadtree->vertices),vertex);
}

int sizeThreadTree(ThreadTree* threadtree) {
	return threadtree->vertices->length;
}

void createThreadTreeIndex(ThreadTree** refThreadTree) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertexNode* current;
	int i;

	// fprintf(stderr,"\n\n ThreadTreeIndex created !!!!\n\n");

	// Verifica se o index esta NULL (por segurana)
	if (threadtree->index != NULL) destroyThreadTreeIndex(&threadtree);

	// Aloca o espao para o index
	threadtree->index = malloc(sizeThreadTree(threadtree)*sizeof(ThreadVertex*));

	// Percorre a lista ligada dos vrtices
	i = 0;
	current = threadtree->vertices->first;
	while(current != NULL) {
		threadtree->index[i++] = current->vertex;
		current = current->next;
	}
}

void destroyThreadTreeIndex(ThreadTree** refThreadTree) {
	ThreadTree* threadtree = *refThreadTree; 

	if (threadtree->index != NULL) {
		free(threadtree->index);
		threadtree->index = NULL;
	}
}

ThreadVertex* getThreadVertex(ThreadTree** refThreadTree, int n) {
	ThreadTree* threadtree = *refThreadTree;

	// Verifica o tamanho
	if (sizeThreadTree(threadtree) < n) return NULL;

	// Indexa para otimizar buscas futuras, mas no seria necessrio faz-lo
	if (threadtree->index == NULL) createThreadTreeIndex(&threadtree);

	return (ThreadVertex*) threadtree->index[n];
}

ThreadVertex* getThreadVertex_NO_MEMORY_LEAK(ThreadTree** refThreadTree, int n) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertexNode* current;
	int i;

	if (sizeThreadTree(threadtree) < n) return NULL;

	// Percorre a lista ligada dos vrtices
	i = 0;
	current = threadtree->vertices->first;
	while(current != NULL && i!=n) {
		i++;
		current = current->next;
	}
}

int addTreeArc(ThreadTree** refThreadTree, ThreadVertex** refParent, ThreadVertex** refChild) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertex* parent = *refParent;
	ThreadVertex* child = *refChild;

	// Define o parent do vrtice
	child->parent = parent;

	// Adicina o vrtice child na lista de filhos do vrtice parent
	return appendThreadVertexNode(&(parent->children),child);

}

int addTreeArcByIds(ThreadTree** refThreadTree, int idParent, int idChild) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertex* parent = getThreadVertex(&threadtree,idParent);
	ThreadVertex* child = getThreadVertex(&threadtree,idChild);

	return addTreeArc(&threadtree,&parent,&child);

}

void dumpThreadTree (ThreadTree* threadtree) {

	ThreadVertexNode* child;
	ThreadVertexNode* current;
	int i;

	current = threadtree->vertices->first;
	while (current != NULL) {
		printf("(%d) %d",current->vertex->status,current->vertex->id);

		child = current->vertex->children->first;
		if (child!=NULL) printf("->"); 
		while (child != NULL) {
			printf("%d ",child->vertex->id);
			child = child->next;
		}

		printf("\n");

		current = current->next;
	}
}

// LINKED LIST: Funes que lidam com a lista de vrtices

ThreadVertexLinkedList* buildThreadVertexLinkedList () {
	ThreadVertexLinkedList* list = malloc(sizeof(ThreadVertexLinkedList));
	list->length = 0;
	list->first = NULL;
	list->last = NULL;
	return list;
}

void destroyThreadVertexLinkedList (ThreadVertexLinkedList** refList) {
	ThreadVertexLinkedList* list = *refList;
	ThreadVertexNode* current;
	ThreadVertexNode* next;

	// Percorre a lista liberando os ThreadVertexNodes
	current = list->first;
	while (current != NULL) {
		next = current->next;
		destroyThreadVertexNode(&current);
		current = next;
	}
	
	// Libera lista
	free(list);
	*refList = NULL;
}

ThreadVertexNode* buildThreadVertexNode (ThreadVertex* vertex, ThreadVertexNode* next) {
	ThreadVertexNode* node = malloc(sizeof(ThreadVertexNode));
	node->vertex = vertex;
	node->next = next;
	return node;
}

void destroyThreadVertexNode(ThreadVertexNode** refNode) {
	ThreadVertexNode* node = *refNode;
	node->vertex = NULL;
	node->next = NULL;
	free(node);
	*refNode = NULL;
}

int appendThreadVertexNode(ThreadVertexLinkedList** refList, ThreadVertex* vertex) {
	ThreadVertexLinkedList* list = *refList;
	ThreadVertexNode* node;

	// Gera o novo node
	node = buildThreadVertexNode(vertex,NULL);

	// Verifica se a lista foi inicializada, ou seja, primeiro elemento  null
	if (list->first == NULL) {
		// Inicializa o tamanho
		list->length = 1;
		// Inicializa o primeiro e o ltimo elemento com o node recm criado
		list->first = node;
		list->last = node;
	} else {
		// Incrementa o tamanho
		list->length = list->length + 1;
		list->last->next = node;
		list->last = node;
	}

	return (list->length - 1);
}

void dumpThreadVertexLinkedList(ThreadVertexLinkedList* list) {
	ThreadVertexNode* current = list->first;
	int i = 0;

	while (current!=NULL) {

		// TODO

		current = current->next;
	}
}

// THREADS

int run(ThreadVertex** refVertex) {
	ThreadVertex* vertex = *refVertex;
	int result = -1;

	// Executa apenas as threads que estejam prontas
	if (vertex->status != THREAD_READY) return result;

	// Altera o status da thread
	vertex->status = THREAD_RUNNING;

	// Cria e executa a query
	result = pthread_create(&vertex->tid, NULL, vertex->task, vertex);

	return result;

}

void setStatus (ThreadVertex** refVertex, int status) {
	ThreadVertex* vertex = *refVertex;

	// Altera o status da thread
	vertex->status = status;
}

int getStatus (ThreadVertex* vertex) {
	return vertex->status;
}

void join (ThreadVertex** refVertex) {
	ThreadVertex* vertex = *refVertex;

	// Aguarda a finalizao da thread
	pthread_join(vertex->tid,NULL);

	// Altera o status da thread
	vertex->status = THREAD_FINISHED;
}

void joinDescendants (ThreadVertex** refVertex) {
	ThreadVertex* vertex = *refVertex;
	ThreadVertexNode* current;

	current = vertex->children->first;
	while (current != NULL) {
		join(&current->vertex);		
		current = current->next;
	}
}

void joinThreadTree (ThreadTree** refThreadTree) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertexNode* current;

	//Verifica se a ROOT da rvore esta determinado
	if (threadtree->root == NULL) {
		current = threadtree->vertices->first;
		while (current != NULL) {
			// Determina o root
			if (current->vertex->parent == NULL) {
				threadtree->root = current->vertex;
				break;
			}
			current = current->next;
		}

		if (threadtree->root==NULL) {
			fprintf(stderr,"NO ROOT !");
			exit(1);
		}
	}

	// Join na ROOT da rvore
	join(&threadtree->root);		

}

void execute(ThreadTree** refThreadTree) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertexNode* current;

	current = threadtree->vertices->first;
	while (current != NULL) {

		// Se necessario determina o root
		if (threadtree->root==NULL && current->vertex->parent==NULL) threadtree->root = current->vertex;

		// Dispara a thread
		if (current->vertex->status == THREAD_READY) run(&current->vertex);

		current = current->next;
	}

	if (threadtree->root==NULL) {
		fprintf(stderr,"NO ROOT !");
		exit(1);
	}

}

void restart(ThreadTree** refThreadTree) {
	ThreadTree* threadtree = *refThreadTree;
	ThreadVertexNode* current;

	current = threadtree->vertices->first;
	while (current != NULL) {

		// Se necessario determina o root
		if (threadtree->root==NULL && current->vertex->parent==NULL) threadtree->root = current->vertex;

		// Dispara a thread
		current->vertex->status = THREAD_READY; // Fora status THREAD_READY
		run(&current->vertex);

		current = current->next;
	}

	if (threadtree->root==NULL) {
		fprintf(stderr,"NO ROOT !");
		exit(1);
	}

}

void setRootById(ThreadTree** refThreadTree, int rootId) {
	ThreadTree* threadtree = *refThreadTree;
	setRoot(&threadtree,getThreadVertex(&threadtree,rootId));
}

void setRoot(ThreadTree** refThreadTree, ThreadVertex* root) {
	ThreadTree* threadtree = *refThreadTree;
	threadtree->root = root;
}
